#ifndef BASIC_LEARN_CPP
#define BASIC_LEARN_CPP
#include "basic_learn.h"

//乱数 rand, srand用
#include <stdlib.h>
//ファイル読み書き用
#include <iostream>
#include <fstream>
#include <boost/tokenizer.hpp>


bool
BasicLearn::init( rcsc::PlayerAgent * agent )
{
  //current_cycle の更新
  current_cycle = agent->world().time().cycle();
  if( !init_flag )
    //一度だけ初期化する
    {
      unum = agent->world().self().unum();		//背番号を記憶する
      std::string fdebug_name = "csv/debug" + tostring(unum) + ".txt";
      fdebug.open( fdebug_name.c_str(), std::ios::out | std::ios::app );
      if( initLearn( agent ) && initWriteFile( agent ) )
	{
	  init_flag = true;	//初期化済みフラグをtrueにする
	  debug( "初期化しました" );
	}
      else
	{
	  init_flag = true;
	  debug( "初期化に失敗しました" );
	  debug( init_flag );
	  return false;
	}
      if( read_flag && init_flag )
	{
	  if( !ReadFile( agent ) )
	    {
	      //読み込む
	      debug( "重みの読み込みに失敗しました" );
	      return false;
	    }
	  debug( "重みの読み込むに成功にしました" );
	}
      else
	{
	  debug("学習結果を読み込みません");
	}
      //初期化と同時には更新しないのでここでtrueを返す
      return true;
    }
  
  if( learn_flag && save_flag  )
    //学習中ならば更新する
    //セーブモードであれば書き込む
    {
      if( isEndEpisode( agent ) )
	{
	  //報酬rをエピソード長から更新する
	  if( !setReward( agent, base ) )
	    {
	      //報酬の更新に失敗したらfalse
	      debug( "報酬rの更新に失敗しました" );
	      return true;
	    }
	  if( Updata( agent, action ) )
	    //更新する
	    {
	      if( !SaveFile( agent ) )
		{
		  //保存する
		  debug("重みの書き込みに失敗しました");
		  return false;
		}
	      debug("重みの書き込みに成功しました");
	      return true;		//更新に成功すればtrue
	    }
	  else
	    {
	      debug( "更新に失敗しました" );
	      return false;
	    }
	}
    }
  else if( learn_flag )
    {
      if( isEndEpisode( agent ) )
	debug("実証実験:レシーブ終了");
    }
  
  //初期化と更新をしない場合もtrueを返す
  return true;
};

bool
BasicLearn::set_learn_param( rcsc::PlayerAgent * agent )
{
  set_boltz_param( agent );	//boltz_paramを設定
  if( !learn_flag )
    //エピソード初めの処理
    {
      learn_flag =  true ;	//学習フラグをオンにします
      first_action_cycle = agent->world().time().cycle();
      //エピソード開始時のボール位置を保持
    }
  last_action_cycle = agent->world().time().cycle();
  episode++;					//エピソード長を増やします
  return true;
};

int BasicLearn::Boltzmann( rcsc::PlayerAgent * agent )
//ボルツマン分布を計算
//@return 行動番号
{
  double energy[MAX_ACTION];	//各行動のエネルギーを保存
  //double tEnergy, tPolicy;		//計算用の変数
  double sum_Policy;      	   	//乱数用に方策の合計を格納
  double probability;				//確率を保存
  double rand_num;					//乱数を保存
  double sumExpEnergy;  //exp(-E/T)の総和//追加
  
  //初期化
  probability = 0;
  sum_Policy = 0;
  //tPolicy = 0;
  sumExpEnergy = 0;  //追加
  
  for(int i=0; i<MAX_ACTION; i++)
    {
      energy[i] = calcEnergy( agent, i );
      //debug("Energy["+tostring(i)+"]="+tostring(Energy[i]));
      //エネルギーを計算
      sumExpEnergy += pow(M_E,-energy[i]/T);//追加
    }

  //ボルツマン分布を計算
  //計算式を変形したもの
  for(int i=0; i<MAX_ACTION; i++){
    /*
      for(int j=0; j<MAX_ACTION; j++){
      tEnergy = (Energy[i] - Energy[j]) / T;
      tPolicy += pow(M_E, tEnergy);
      }
      Policy[i] = 1 / tPolicy;
    */
    policy[i] = pow(M_E,-energy[i]/T)/sumExpEnergy;//追加
    sum_Policy += policy[i];
    //tPolicy = 0.0;
  }
  
  //乱数を発生
  rand_num = sum_Policy * rand() / RAND_MAX;	//乱数の公式　0~sum_Policyの乱数を取得
  //各行動の確率を計算
  for(int i=0; i<MAX_ACTION; i++)
    {
      probability += policy[i];
      if( rand_num <= probability )
	{
	  //実行する行動を返す
	  return i;
	}
    };
  
  return MAX_ACTION - 1;
};

bool
BasicLearn::initLearn( rcsc::PlayerAgent * agent )
{
  std::ifstream infile( learn_file_name.c_str(), std::ios::in );
  //読み込み用ストリーム
  debug("設定ファイルの読み込みを開始します");
  
  if( !infile )
    {
      debug(tostring(learn_file_name)+"が見つかりません");
      return false;
    }
  else
    //トークンに分けて読み出し
    {
      std::string mes;
      //読み込み用バッファ
      typedef boost::char_separator< char > character_separator;
      typedef boost::tokenizer< character_separator > tokenize;
      character_separator c_separator( "=", "", boost::drop_empty_tokens );
      //"="でトークンを作る
      while( getline( infile, mes ) )
	{
	  if( ( mes.substr(0)=="#" ) || ( mes.substr(0)=="\n" ) || mes.empty() )
	    continue;
	  //#（コメント)　空行なら飛ばす
	  tokenize token_mes( mes, c_separator );	//tokenizeインスタンスを作成し、設定ファイルの行を読み込み
	  tokenize::iterator ifirst = token_mes.begin();
	  for(tokenize::iterator ifirst = token_mes.begin();
	      ifirst != token_mes.end();
	      ifirst++ )
	    {
	      //flagを読み出す
	      if( (*ifirst) == "save" )
		{
		  ifirst++;
		  if( (*ifirst) == "true" && ( ifirst != token_mes.end() ) )
		    save_flag = true;
		  else
		    save_flag = false;
		}
	      if( (*ifirst) == "read" )
		{
		  ifirst++;
		  if( (*ifirst) == "true" && ( ifirst != token_mes.end() ) )
		    read_flag = true;
		  else
		    read_flag = false;
		}
	      //ファイルパスを読み出す
	      if( (*ifirst) == "path" )
		{
		  ifirst++;
		  if( ifirst != token_mes.end() )
		    path = (*ifirst);
		}
	      //学習係数を読み出す
	      if( (*ifirst) == "epsilon" )
		{
		  ifirst++;
		  if( ifirst != token_mes.end() )
		    epsilon = boost::lexical_cast< double >( (*ifirst) );
		}
	      //温度を読み出す
	      if( (*ifirst) == "T" )
		{
		  ifirst++;
		  if( ifirst != token_mes.end() )
		    T = boost::lexical_cast< double >( (*ifirst) );
		}
	    }
	}
      
    }
  infile.close();
  return true;
};

bool
BasicLearn::initWriteFile( rcsc::PlayerAgent * agent )
{
  //背番号の値からファイルを作成
  file_name = "learn" + boost::lexical_cast< std::string >(agent->world().self().unum()) + ".csv";
  //pathと結合する
  path = path + file_name;
  return true;
};

bool
BasicLearn::Updata( rcsc::PlayerAgent * agent, int action )
{
  //重みを更新
  if( ( 0 <= action) && (action < MAX_ACTION) )
    //actionが範囲外ならエラー
    {
      double heuristics_sum = 0.0;     // エピソード毎のヒューリスティクスごとの合計
      double all_heuristics_sum[MAX_HEURISTICS]; // 全エピソードのヒューリスティクスの合計
      for(int i=0; i<MAX_HEURISTICS; i++)
	all_heuristics_sum[i] = 0.0; // 初期化
      //最後に行動したほうの更新
      for(int i=0; i<episode; i++)
	{    // エピソードを調べていく
	  for(int j=0; j<MAX_HEURISTICS; j++)
	    {
	      heuristics_sum = 0.0;   //hue_sumを初期化
	      for(int k=0; k<MAX_ACTION; k++)
		{
		  heuristics_sum 
		    += boltz_Heuristics[k][j][i] * boltz_Policy[k][i];
		  //iエピソード時の背番号mのドリブル角度番号nの時のヒューリスティクスの合計
		}
	      all_heuristics_sum[j] 
		+= boltz_Heuristics[action][j][i] - heuristics_sum;
	      //ヒューリスティクス合計に加えるとき現在の行動番号actionのヒューリスティクスを引くこと
	    }
	}
      for(int i=0; i<MAX_HEURISTICS; i++){
	omega[i] += epsilon * r * all_heuristics_sum[i] / T;
	//debug("Hue["+tostring(i)+"]="+tostring(all_hue_sum[i]));
	//重みを更新
	if(omega[i] < 0) 
	  omega[i] = 0.0;
	//重みが0以下のとき0にする
	//重みがマイナスになるのを防ぐため
      }
      //エピソード中に保存したヒューリスティクスの値と方策の値、各ヒューリスティクスの合計を初期化
      //終了処理を行う
      finish_learn( agent );
      return true;
    }
  else
    {
      debug("更新に失敗しました");
      return false;
    }
};

/**** Updataメソッドに必要なメソッド ***/
bool
BasicLearn::set_boltz_param( rcsc::PlayerAgent * agent )
{
  for(int i=0; i<MAX_ACTION; i++)
    {
      //現エピソード中の方策の値を保存
      boltz_Policy[i].push_back(policy[i]);
      for(int j=0; j<MAX_HEURISTICS; j++)
	{
	  //現エピソード中のヒューリスティクスを追加
	  boltz_Heuristics[i][j].push_back( getHeuristics(agent, i, j) );
	}
    }
  return true;
};

bool
BasicLearn::finish_learn( rcsc::PlayerAgent * agent )
{
  for(int i=0; i<MAX_ACTION; i++)
    {
      for(int j=0; j<MAX_HEURISTICS; j++)
	boltz_Heuristics[i][j].clear();
      boltz_Policy[i].clear();
    }
  //学習中フラグをオフにする
  learn_flag = false;
  //エピソード長を初期化する
  episode = 0;
  return true;
};

/**** ボルツマン分布計算に必要な関数 ****/
double BasicLearn::calcEnergy( rcsc::PlayerAgent * agent, int action)
{
  double energy = 0;	//計算用変数

  for(int i=0; i<MAX_HEURISTICS; i++)
    {
      energy += omega[i] * getHeuristics( agent, action, i);
    }
  return energy;
};

bool
BasicLearn::ReadFile( rcsc::PlayerAgent * agent )
{
  //ファイルには 試合数count,報酬r,行動action,重みw[0],w[1],w[2]・・・とする
  std::ifstream infile( path.c_str(), std::ios::in );
  //読み込み用ストリーム
  int game_count=0;
  //試合数を記憶
  
  if( !infile )
    {
      debug( file_name + "が見つかりません" );
      read_game_count_flag = true;	//ファイルが見つからない=0回目の試合と判断
      return false;
    }
  else
    {
      std::string mes;		//infileから取り込むのに用いる
      std::string mes_buf;//実際に読み込むのに用いる
      //読み込み用バッファ
      typedef boost::char_separator< char > char_separator;
      typedef boost::tokenizer<char_separator> tokenize;
      char_separator c_separator( ",", "", boost::drop_empty_tokens );
      //"="でトークンを作る
      while( getline( infile, mes ) )
	{
	  if( (mes.substr(0)=="#") || (mes.substr(0)=="\n") || mes.empty() )
	    continue;
	  //#（コメント)　空行なら飛ばす
	  tokenize token_mes( mes, c_separator );	//tokenizeインスタンスを作成し、設定ファイルの行を読み込み
	  for( tokenize::iterator ifirst = token_mes.begin();
	       ifirst != token_mes.end();
	       ifirst++ )
	    {
	      game_count = boost::lexical_cast< int >( (*ifirst) );
	      //最後の行の試合数を記憶
	      ifirst++;		//試合数skip
	      ifirst++;		//報酬skip
	      ifirst++;		//行動skip
				//重みを読み出し
	      for(int i=0; i<MAX_HEURISTICS; i++)
		{
		  omega[i] = boost::lexical_cast< double >( (*ifirst) );
		  ifirst++;		//次の重さにイテレーターをずらす
		  //debug( "read:w[" + tostring(i) + "]=" + tostring(w[i]) );
		}
	      if( ifirst == token_mes.end() ) 
		break;	//イテレーターがラストに来ていれば抜ける
	    }
	}
      debug(file_name+"の読み込みに成功しました");
    }
  if( !read_game_count_flag )
    //一度のみゲーム数を読み込む
    {
      game_count++;
      this->game_count = game_count;
      read_game_count_flag = true;
    }
  infile.close();
  return true;
};

bool
BasicLearn::SaveFile( rcsc::PlayerAgent * agent )
{
  std::string mesbuff;		//出力用バッファ
  std::ofstream outfile( path.c_str(), std::ios::out | std::ios::app );
  //追記で開く
  if( !outfile )
    {
      debug( file_name + "が見つかりません" );
      return false;
    }
  else
    {
      //ファイルには
      //試合数count,報酬r,行動action,重みomega[0],omega[1],omega[2]・・・とする
      mesbuff = 
	boost::lexical_cast< std::string >( game_count ) + "," 
	+ boost::lexical_cast< std::string >( r ) + "," 
	+ boost::lexical_cast< std::string >( action ) + ",";
      for(int i=0; i<MAX_HEURISTICS; i++)
	{
	  mesbuff = mesbuff + boost::lexical_cast< std::string >( omega[i] );
	  if( (i+1) < MAX_HEURISTICS )
	    mesbuff = mesbuff + ",";
	  debug( "write:w[" 
		 + boost::lexical_cast<std::string>(i) 
		 + "]=" 
		 + boost::lexical_cast<std::string>(omega[i]) );
	  
	}
      outfile << mesbuff << std::endl;
      debug(file_name+"の書き込みに成功しました");
    }
  outfile.close();
  return true;
};

/*** AbstractPlayerObject獲得用のメソッド ***/
const rcsc::AbstractPlayerObject *
BasicLearn::getTeam( rcsc::PlayerAgent * agent , int num )
//@return 背番号numの味方プレイヤーを返す
{
  /*
    rcsc::AbstractPlayerCont team = agent->world().allTeammates();
    for(	rcsc::AbstractPlayerCont::const_iterator target_it = team.begin();
    target_it != team.end();
    target_it++
    )
    if( (*target_it)->unum() == num ) return (*target_it);
    return ( *team.begin() );
  */
  return agent->world().teammate( num );
};

const rcsc::AbstractPlayerObject *
BasicLearn::getOpp( rcsc::PlayerAgent * agent , int num )
//@return 背番号numの敵プレイヤーを返す
{
  /*
    rcsc::AbstractPlayerCont opp = agent->world().allOpponents();
    for(	rcsc::AbstractPlayerCont::const_iterator target_it = opp.begin();
    target_it != opp.end();
    target_it++
    )
    if( (*target_it)->unum() == num ) return (*target_it);
    return ( *opp.begin() );
  */
  return agent->world().opponent( num );
};

double
BasicLearn::TeamCont( const rcsc::AbstractPlayerObject * team )
//@return 信頼性を[0,1]で返す
{
  //自作敵位置評価値
  double Conf = 1 - team->posCount() / 10;
  if( Conf < 0 )
    return 0;
  //Conf[0,1]
  return Conf;
};

double
BasicLearn::EnemyCont( const rcsc::AbstractPlayerObject * opp )
//@return 信頼性を[0,1]で返す
{
  //自作敵位置評価値
  double Conf = 1 - opp->posCount() / 10;
  if( Conf < 0 )
    return 0;
  //Conf[0,1]
  return Conf;
};

double
BasicLearn::BallCont( const rcsc::BallObject * ball )
//@return 信頼性を[0,1]で返す
{
  //自作敵位置評価値
  double Conf = 1 - ball->posCount() / 10;
  if( Conf < 0 )
    return 0;
  //Conf[0,1]
  return Conf;
};

void
BasicLearn::set_learn_flag( bool flag )
{
  learn_flag = flag;
  return;
}
bool
BasicLearn::get_learn_flag( void )
{
  return learn_flag;
}

#endif
