/*** 学習に必要なメソッドを定義 ***/

#ifndef BASIC_LEARN_H
#define BASIC_LEARN_H

#define DRIBBLE_DEBUG	//デバッグモードを使用する

/*** include ***/
//サッカーライブラリ読み込み
#include <rcsc/player/player_agent.h>
//文字列用
#include <string>
#include <iostream>
#include <boost/lexical_cast.hpp>

class BasicLearn
{
 protected:
  //継承を考えてprotectedにします
  /*** 変数 ***/
  //15<x<45、-30<x<30を12×6セルに分割する
  //1セル=5×5m平方
  static const int WIDTH = 12;
  static const int HEIGHT = 6;
  static const int MAX_HEURISTICS = 6;	       //ヒューリスティクスの最大数
  static const int MAX_ACTION = WIDTH * HEIGHT;//行動の最大値
  static const double MAX_VALUE = 10.0;	       //評価値最大値
  static const int MAX_TEAMMATES = 11;	       //味方チームプレイヤーの数
  static const int MAX_OPPONENTS = 11;	       //相手チームプレイヤーの数
  
  /*** 学習フラグ ***/
  //ファイルから読み込む
  bool read_flag;		//学習結果を読み込むするフラグ
  bool save_flag;		//学習結果を保存するフラグ
  bool read_game_count_flag;
  //ゲーム数を読み出しかを保存
  //学習中に使用する
  bool init_flag;		//初期化したかを確認
  bool learn_flag;	        //行動をして学習しているかを確認
  
  /*** ファイル用変数 ***/
  std::string learn_file_name;	//学習設定ファイルのファイル名
  std::string file_name;	//学習を保存するファイルの名前
  std::string path;		//学習を保存するファイルのパス
  
  // 学習関係
  //ファイル読み出し
  int game_count;	      	//現在の試合数
  int action;   		//学習中、最後に行った行動
  double omega[MAX_HEURISTICS];	//重み配列
  double epsilon;		//学習係数
  double r;			//行動に対する報酬
  double T;			//温度
  //報酬計算に使用
  int current_cycle;		//現在のサイクル
  int first_action_cycle;	//最初に行動したサイクル
  int last_action_cycle;	//最後に行動したサイクル
  double base;			//報酬を規定する基礎値
  //計算に使用
  double policy[MAX_ACTION];	//行動番号のエネルギーを格納する
  std::vector< double > boltz_Heuristics[MAX_ACTION][MAX_HEURISTICS];		//ヒューリスティックの一時的保存
  std::vector< double > boltz_Policy[MAX_ACTION]; //方策（確率)の一時的保存
  //エピソード分の保存はvectorを用いる
  int episode;			//学習のエピソード長
  double heuristics_Conf;	//ヒューリスティクスに置ける信頼性
  
  /*** デバッグ用 ***/
  //背番号
  int unum;
  //デバッグ用ファイルストリーム
  std::ofstream fdebug;
 public:
 BasicLearn(int unum_ = 0
	    ,bool read_flag_ = false ,bool save_flag_ = false 
	    ,bool init_flag_ =false ,bool learn_flag_ = false 
	    ,bool read_game_count_flag_ = false
	    ,std::string learn_file_name_ = "csv/learn.ini"
	    ,std::string file_name_ = "learn.csv"
	    ,std::string path_ = "$HOME/csv/"
	    ,int game_count_ = 0 ,int action_ = 0 
	    ,double epsilon_ = 0.0 ,double r_ = 0.0 ,double T_ = 0.0
	    ,int episode_ = 0 ,double heuristics_Conf_ = 0.0
	    ,int first_action_cycle_ = 0 ,int last_action_cycle_ = 0
	    ,double base_ = 1.0)
   //コンストラクタ
   :learn_file_name(learn_file_name_)
	  ,file_name(file_name_) ,path(path_)
	{
	  unum = unum_;
	  read_flag = read_flag_; save_flag = save_flag_;
	  init_flag = init_flag; learn_flag = learn_flag_;
	  read_game_count_flag = read_game_count_flag_;
	  game_count = game_count_; action = action_;
	  epsilon = epsilon_; r = r_; T = T_;
	  episode = episode_; heuristics_Conf = heuristics_Conf_;
	  first_action_cycle = first_action_cycle_;
	  last_action_cycle = last_action_cycle_;
	  base = base_;
	  //乱数seedを設定
	  srand( (unsigned)time(NULL) );
	  //初期化
	  for(int i=0; i<MAX_HEURISTICS; i++)
	    {
	      omega[i] = 0;
	    }
	  for(int i=0; i<MAX_ACTION; i++)
	    {
	      policy[i] = 0;
	    }
	};
	~BasicLearn()
	  //デストラクト
	  {
	    fdebug.close();
	  };
	
	//BasicLearn();
	//~BasicLearn();

	bool init( rcsc::PlayerAgent * agent );
	//初期化メソッド

	bool set_learn_param( rcsc::PlayerAgent * agent );
	//set_boltz_param, set_learn_flag( true )を実行
	//ドリブル実行時に行うメソッドをまとめる
	//ボルツマン分布の計算と学習実行を分割する

protected:
	/*** Botzmann分布 ***/
	virtual int Boltzmann( rcsc::PlayerAgent * agent );
	//ボルツマン分布を用いて行動を決定する
	//@param PlayerAgent
	//@return 選択された行動番号

	bool initLearn( rcsc::PlayerAgent * agent );
	//学習初期化
	//learn.iniファイルからフラグを読み出す

	bool initWriteFile( rcsc::PlayerAgent * agent );
	//書き込むファイルの設定
	//各背番号ごとにファイルを作成

	bool Updata( rcsc::PlayerAgent * agent, int action );
	//重みを更新する
	//報酬を与える
	//@param action 分割番号 0~MAX_ACTION-1

	/*** チェックメソッド、設定メソッド ***/
	virtual bool isEndEpisode( rcsc::PlayerAgent * agent ) = 0;
	//エピソードが終了したかを確認
	//エピソードが終了していればtrue そうでなければfalse

	virtual bool setReward( rcsc::PlayerAgent * agent, double base = 1.0 ) = 0;
	//報酬rの値を設定

	bool set_boltz_param( rcsc::PlayerAgent * agent );
	//boltz_Heuristics,boltz_Policyに値を追加する
	//動作上、execute実行後に更新しなければならないので分離する

	bool finish_learn( rcsc::PlayerAgent * agent);
	//学習終了処理
	//まとめて関数にすると便利なので作成

	/*** Botzmann計算用関数 ***/
	double calcEnergy(rcsc::PlayerAgent * agent, int action);
	//divisionドリブル番号のエネルギー(ヒューリスティクス*重みの合計)を計算
	//@param division ドリブル番号
	//@return ヒューリスティクスのエネルギー

	virtual double getHeuristics(rcsc::PlayerAgent * agent
				     , int action
				     , int heuristics_num) = 0;
	//行動actionのhulis_num番目のヒューリスティクスの値を返す
	//純粋仮想関数として定義
	//実際の定義は継承して行う

	//学習結果読み書き用メソッド
	bool ReadFile( rcsc::PlayerAgent * agent );
	//ファイルから重みを読み出す
	bool SaveFile( rcsc::PlayerAgent * agent );
	//ファイルに重みを書き込む

public:
	/*** AbstractPlayerObject獲得用のメソッド ***/
	const rcsc::AbstractPlayerObject 
	  * getTeam( rcsc::PlayerAgent * agent , int num );
	//@return 背番号numの味方プレイヤーを返す

	const rcsc::AbstractPlayerObject 
	  * getOpp( rcsc::PlayerAgent * agent , int num );
	//@return 背番号numの敵プレイヤーを返す

protected:
	double TeamCont( const rcsc::AbstractPlayerObject * team );
	//@return 背番号team_numの味方の信頼性[0,1]を返す

	double EnemyCont( const rcsc::AbstractPlayerObject * opp );
	//@return 背番号enemy_numの敵の信頼性[0,1]を返す

	double BallCont( const rcsc::BallObject * ball );
	//@return ボールの信頼性[0,1]を返す

public:
	//セッター、ゲッター
	void set_learn_flag( bool flag );
	//learn_flagの値を設定
	//外から設定できると便利なので作成
	//@param flag セットするbool値

	bool get_learn_flag( void );

protected:
	/*** デバッグ用メソッド ***/
	template <class T> void debug( const T string )
	{
		#ifdef DRIBBLE_DEBUG
			std::cout << unum << ": " << string << std::endl;
			std::cout.flush();
		#endif
		//DEBUGモードにするにはDRIBBLE_DEBUGを定義する
	};
	template <class T> std::string tostring( const T base)
	{
		return boost::lexical_cast< std::string >( base );
	};
};

#endif
