#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_pass_receiver_move.h"

#include "bhv_basic_tackle.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/intercept_table.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

static PassReceiver passreceiver;
static rcsc::Vector2D goToPoint(0.0,0.0);	//移動先座標
static bool moving_flag=false;   	//レシーブ移動中フラグ
//init処理
bool
Bhv_PassReceiverMove::init( rcsc::PlayerAgent * agent )
{
  return passreceiver.init( agent );
}

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_PassReceiverMove::execute( rcsc::PlayerAgent * agent )
{
  rcsc::dlog.addText( rcsc::Logger::TEAM,
		      "%s:%d: Bhv_PassReceiverMove"
		      ,__FILE__, __LINE__ );
  const rcsc::WorldModel & wm = agent->world();
  
  if( passreceiver.isPassing()
      && wm.self().pos().dist( goToPoint ) > 1.0
      //&& wm.self().pos().dist( wm.ball().pos() ) > 5.0
      && moving_flag )
    {
      //HELIOS処理
      const double dash_power = rcsc::ServerParam::i().maxDashPower();
      //試験的に最大パワーで目標座標に移動する
      double dist_thr = 1.0;
      if ( ! rcsc::Body_GoToPoint( goToPoint, dist_thr, dash_power
				   ).execute( agent ) )
	{
	  rcsc::Body_TurnToPoint(goToPoint).execute( agent );
	}
      agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
      //printf("go to pass point(%2f,%2f)\n",goToPoint.x,goToPoint.y);
      return true;
    }
  else
    {
      moving_flag = false;
    }
  
  //味方がボールを持っている場合、パスレシーバークラスを使う
  //条件:ボール保持者との距離が声の届く範囲内であること
  //１０サイクル以内の信頼性を持つ人物がボールを保持していること
  //条件1：オフサイドライン手前である
  //条件2：敵陣地奥でないでないこと
  bool existKickableTeammate = false;
  if( wm.getTeammateNearestToBall(10) )
    existKickableTeammate 
      = wm.getTeammateNearestToBall(10)->isKickable();
  if( existKickableTeammate
      && wm.self().pos().x < wm.offsideLineX()
      && wm.getTeammateNearestToBall(10)->pos().x < 20.0
      && !passreceiver.isPassing()
      ){
    const double pass_to_receiver_dist
      //パーサーとレシーバーとの距離を測定
      =wm.getTeammateNearestToBall(10)->pos().dist( wm.self().pos() );
    if( pass_to_receiver_dist <= 50 )
      //声の聞こえる範囲にパーサーがいる場合実行
      {
	goToPoint = passreceiver.execute( agent );
	moving_flag = true;
	//const double dash_power = getDashPower( agent, goToPoint );
	const double dash_power = rcsc::ServerParam::i().maxDashPower();
	//試験的に最大パワーで目標座標に移動する
	double dist_thr = dist_thr = 1.0;
	if ( ! rcsc::Body_GoToPoint( goToPoint, dist_thr, dash_power
				     ).execute( agent ) )
	  //状況が変わるに連れて移動先を変えるため移動継続は20サイクル
	  {
	    rcsc::Body_TurnToPoint(goToPoint).execute( agent );
	  }
	agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	return true;
      }
  }

  if( passreceiver.isPassing()
      && wm.self().pos().dist( goToPoint ) < 1.0
      && wm.self().pos().dist( wm.ball().pos() ) > 15.0
      && !moving_flag ){
    rcsc::Body_TurnToBall().execute(agent);
    agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan() );
    return true;
  }

  return false;
}

/*-------------------------------------------------------------------*/
/*!
  
 */
double
Bhv_PassReceiverMove::getDashPower( const rcsc::PlayerAgent * agent,
				    const rcsc::Vector2D & target_point )
{
  static bool s_recover_mode = false;
  
  const rcsc::WorldModel & wm = agent->world();
  /*--------------------------------------------------------*/
  // check recover
  
  if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.5 )
    {
      s_recover_mode = true;
    }
  else if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.7 )
    {
      s_recover_mode = false;
    }
  
  /*--------------------------------------------------------*/
  double dash_power = 100.0;
  const double my_inc
    = wm.self().playerType().staminaIncMax()
    * wm.self().recovery();
  
  if ( wm.defenseLineX() > wm.self().pos().x
       && wm.ball().pos().x < wm.defenseLineX() + 20.0 )
    {
      rcsc::dlog.addText( rcsc::Logger::TEAM,
			  "%s:%d: correct DF line. keep max power"
			  ,__FILE__, __LINE__ );
      // keep max power
      dash_power = rcsc::ServerParam::i().maxPower();
    }
  else if ( s_recover_mode )
    {
      dash_power = my_inc - 25.0; // preffered recover value
      if ( dash_power < 0.0 ) dash_power = 0.0;
      
      rcsc::dlog.addText( rcsc::Logger::TEAM,
			  "%s:%d: recovering"
			  ,__FILE__, __LINE__ );
    }
  // exist kickable teammate
  else if ( wm.existKickableTeammate()
	    && wm.ball().distFromSelf() < 20.0 )
    {
      dash_power = std::min( my_inc * 1.1,
			     rcsc::ServerParam::i().maxDashPower() );
      rcsc::dlog.addText( rcsc::Logger::TEAM,
			  "%s:%d: exist kickable teammate. dash_power= %f"
			  ,__FILE__, __LINE__,
			  dash_power );
    }
  // in offside area
  else if ( wm.self().pos().x > wm.offsideLineX() )
    {
      dash_power = rcsc::ServerParam::i().maxDashPower();
      rcsc::dlog.addText( rcsc::Logger::TEAM,
			  "%s:%d: in offside area. dash_power= %f"
			  ,__FILE__, __LINE__,
			  dash_power );
    }
  // normal
  else
    {
      dash_power = std::min( my_inc * 1.7,
			     rcsc::ServerParam::i().maxDashPower() );
      rcsc::dlog.addText( rcsc::Logger::TEAM,
			  "%s:%d: normal mode dash_power= %f"
			  ,__FILE__, __LINE__,
			  dash_power );
    }
  
  return dash_power;
}

