// -*-c++-*-

/*
 *Copyright:

 Copyright (C) SIT RoboCup Soccer Project

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_pass_to_receiver.h"

#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>


//追加ヘッダ
//乱数 rand, srand用
#include <stdlib.h>
#include <rcsc/common/audio_memory.h>
//messageを送るのに必要
#include<rcsc/player/free_message.h>
//パス計算用
#include <rcsc/math_util.h>

#include <rcsc/action/body_smart_kick.h>


//パス用変数
static int get_msg_cycle = -1;		//最新のメッセジを受け取ったサイクル
static int receiver_unum = -1;		//パス先の番号
static rcsc::Vector2D pass_pos( 0, 0);	//パス先の座標
static bool wait_pass_flag = false;	//パス待機中であるかの判断
static int start_wait_cycle = -1;	//パス待機時間のために記憶
static bool passing_flag = false;	//パス中であるかの判断
static int start_pass_cycle = -1;	//パスを開始したサイクル
                                        //3サイクルパスを行う

bool
Bhv_PassToReceiver::execute( rcsc::PlayerAgent * agent )
{
  /* 追加 */
  //パスメッセジを受け取る
  if( !agent->world().audioMemory().freeMessage().empty()
      && ( get_msg_cycle 
	   != agent->world().audioMemory().freeMessageTime().cycle() )
      && !wait_pass_flag
      && !passing_flag)
    {
      //プレイヤーから受け取ったレシーバー情報を取得
      receiver_unum = -1;
      int receiver_action = -1;
      
      std::string msg;	//メッセジバッファ
      for(unsigned int i=0; 
	  i < agent->world().audioMemory().freeMessage().size();
	  i++)
	{
	  msg = agent->world().audioMemory().freeMessage()[i].message_;
	  if(msg.size() != 2)
	    continue;
	  //展開します
	  if( msg.substr(2,2) == "OK")
	    continue;
	  receiver_unum 
	    = agent->world().audioMemory().freeMessage()[i].sender_;
	  receiver_action 
	    = boost::lexical_cast<int>( msg.substr(0,2) );
	  get_msg_cycle 
	    = agent->world().audioMemory().freeMessageTime().cycle();
	}
      
      if( agent->world().time().cycle() - get_msg_cycle <= 10 )
	//5サイクル以上前の情報は見ない
	{
	  //std::cout << "get msg:" << msg << std::endl;
	  if( receiver_unum != -1 )
	    {
	      //座標を計算
	      int action_x = receiver_action % 6;	//x座標を調べる
	      int action_y = receiver_action / 6;	//y座標を調べる
	      const double offsetX = 15.0 + 2.5; //レシーブ位置は15<x<45
	      const double offsetY = -30.0 + 2.5; //,-30<y<30
	      pass_pos.x = action_x * 5.0 + offsetX;
	      pass_pos.y = action_y * 5.0 + offsetY;
	      wait_pass_flag = true;	//パス待機中フラグをONにする
	      start_wait_cycle = agent->world().time().cycle();	//サイクル記憶
	      //まずパスを行うことを決めたメッセージを送る(番号つき)
	      char buff[254];
	      sprintf(buff,"%02dOK", receiver_unum);
	      std::string msg(buff);
	      agent->addSayMessage( new rcsc::FreeMessage<4>( msg ) );
	      agent->debugClient().addMessage( "receiver_pass" );
	      agent->debugClient().setTarget( pass_pos );
	    }
	}
    }
  if( wait_pass_flag )
    //パス待機中であれば、味方の座標との距離を見てボールを保持する
    {
      const rcsc::AbstractPlayerObject 
	* receiver = agent->world().teammate( receiver_unum );
      double diff_dist = 0.0;
      if( receiver )
	diff_dist = 
	  agent->world().ball().pos().dist( pass_pos ) 
	  - receiver->pos().dist( pass_pos );
      if( receiver
	  && diff_dist < 0.0
	  && agent->world().time().cycle() - start_wait_cycle < 10
	  && receiver->pos().x < agent->world().offsideLineX() - 2 )
	{
	  rcsc::dlog.addText( rcsc::Logger::TEAM,
			      "%s:%d: hold"
			      ,__FILE__, __LINE__ );
	  agent->debugClient().addMessage( "OffKickHold" );
	  rcsc::Body_HoldBall().execute( agent );
	  agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
	  printf("%2d:waiting to pass to %2d\n",agent->world().self().unum(),receiver_unum);
	  return true;
	}
      else
	{
	  wait_pass_flag = false;
	  passing_flag = true;
	  start_pass_cycle = agent->world().time().cycle();
	}
    }
  //パスを確実にするために3サイクルパスを行う
  if( agent->world().time().cycle() - start_pass_cycle < 3
      && passing_flag )
    {
      //速度生成
      const rcsc::Vector2D 
	receiver_rel = pass_pos - agent->world().ball().pos();
      const double receiver_dist = receiver_rel.r();
      const rcsc::AngleDeg receiver_angle = receiver_rel.th();
      double end_speed = 1.5;
      double first_speed = 100.0;
      do
	{
	  first_speed
	    = rcsc::calc_first_term_geom_series_last
	    ( end_speed,
	      receiver_dist,
	      rcsc::ServerParam::i().ballDecay() );
	  if ( first_speed < rcsc::ServerParam::i().ballSpeedMax() )
	    {
	      break;
	    }
	  end_speed -= 0.1;
	}
      while ( end_speed > 0.8 );
      if(
	 rcsc::Body_SmartKick( pass_pos,
			       first_speed,
			       first_speed * 0.96,
			       3
			       ).execute( agent )
	 )
	{
	  //パーサーとレシーバーとの距離を測定
	  //const double pass_to_receiver_dist
	  //=agent->world().teammate( receiver_unum )->pos().dist( agent->world().self().pos() );
	  //std::cout << agent->world().time().cycle() << ":dist=" << pass_to_receiver_dist << std::endl;
	  
	  agent->debugClient().addMessage( "receiver_pass" );
	  agent->debugClient().setTarget( pass_pos );
	  if ( agent->config().useCommunication() )
	    {
	      rcsc::dlog.addText( rcsc::Logger::ACTION,
				  "%s:%d: execute() set pass communication."
				  ,__FILE__, __LINE__ );
	      rcsc::Vector2D target_buf = receiver_rel - agent->world().self().pos();
	      target_buf.setLength( 1.0 );
	      
	      agent->addSayMessage( new rcsc::PassMessage( receiver_unum,
							   receiver_rel 
							   + target_buf,
							   agent->effector().queuedNextBallPos(),
							   agent->effector().queuedNextBallVel() ) );
	    }
	  agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
	  printf("%2d:passing to %2d\n",agent->world().self().unum(),receiver_unum);	  
	  return true;
	}
    }
  else
    {
      passing_flag = false;
    }
  return false;
  //パスをしなかった場合falseを返す
  /* 追加終了 */
}
