#include "heuristics_method.h"
#include<rcsc/math_util.h>
#include<rcsc/player/free_message.h>
#include<cmath>
/*
 * heulistics_method.cpp
 *
 *  Created on: 2008/12/14
 *      Author: strman
 */

double
HeuristicsMethod::getHeuristics(rcsc::PlayerAgent * agent
				, int action
				, int heuristics_num)
{
  double heuristics_eval = 0.0;
  heuristics_Conf = 1.0;
  //ヒューリスティクスで使用している味方、敵の信頼性を格納
  //評価値heuristics_evalにかけることで信頼性を考慮する
  
  //現在ボールを持っているプレイヤーを取得
  //信頼値はヒューリスティクスで判断するのでposCount=99で取得
  const rcsc::PlayerObject * passer 
    = agent->world().getTeammateNearestToBall(10);
  
  rcsc::Vector2D goToPoint( 0.0, 0.0 );
  int action_x = action % 6;	//x座標を調べる
  int action_y = action / 6;	//y座標を調べる
  
  const double offsetX = 15.0 + 2.5; //レシーブ位置は15<x<45
  const double offsetY = -30.0 + 2.5; //,-30<y<30
  goToPoint.x = action_x * 5.0 + offsetX;
  goToPoint.y = action_y * 5.0 + offsetY; 

  switch( heuristics_num )
    {
      /*** パスの安全性を判断 ***/
    case 0:
      {
	//パスコースに敵がいるかを評価
	heuristics_eval = EnemyInPassCourse( agent, passer, goToPoint );
	break;
      }
    case 1:
      {
	//もっとも近い敵プレイヤーとの距離を評価
	heuristics_eval = DistToNearestEnemy( agent, passer, goToPoint );
	break;
      }
      /*** パスを受け取れるかを判断 ***/
    case 2:
      {
	//目標ポイントからのボールとレシーバーとの距離差を評価
	heuristics_eval 
	  = DiffTakeTimeToBallAndTakeTimeToReceiver( agent, passer, goToPoint );
	break;
      }
    case 3:
      {
	//レシーバーがボールをパーサーの下に蹴るのにかかるパワーの評価
	heuristics_eval = CheckBallSpeed( agent, passer, goToPoint );
	break;
      }
      /*** 攻撃性を判断 ***/
    case 4:
      {
	heuristics_eval = DistToOffsideLineX( agent, passer, goToPoint );
	//目標ポイントがどれだけオフサイドラインより後ろであるかを調べる
	break;
      }
    case 5:
      {
	//目標地点とゴールとの距離差を評価
	heuristics_eval = DistToGoal( agent, passer, goToPoint );
	break;
      }
    default:
      {
	debug("ヒューリスティクスの範囲外です:"+tostring(heuristics_num));
	break;
      }
    }
  heuristics_eval *= heuristics_Conf;
  return heuristics_eval;
};

/*** ヒューリスティクスの定義 ***/
double
HeuristicsMethod::EnemyInPassCourse( rcsc::PlayerAgent * agent
				     , const rcsc::PlayerObject * passer
				     , rcsc::Vector2D goToPoint )
{
  double value = 10.0;
  
  /*** ヒューリスティクスで使用されるプレイヤーの信頼性を考慮するための変数 ***/
  double tmp_heuristics_conf = 1.0;		//一時的に信頼性を保存
  bool get_player_flag = false;	//プレイヤー情報を利用したかのフラグ
  
  const rcsc::Vector2D passer_pos = passer->pos();
  //passerの位置情報を取得
  
  //レシーバーとパーサー間の計算
  const double passer_to_receive_point_dist = passer_pos.dist( goToPoint );
  //レシーバーとパーサーとの間の距離を計算
  const double passer_to_receive_point_angle = ( passer_pos - goToPoint ).th().degree();
  //レシーバとパーサとの間の角度を計算
  
  //敵プレイヤーについて計算
  for(int i=1; i<=MAX_OPPONENTS; i++)
    {
      const rcsc::AbstractPlayerObject * opp = getOpp( agent, i );
      //敵プレイヤーを取得
      if( !opp )
	//敵情報が取得できなければ次の選手へ
	continue;
      get_player_flag = true;
      //プレイヤー情報を使ったのでフラグをonにする
      const rcsc::Vector2D opp_pos = opp->pos();
      //敵の位置情報を取得
      const double passer_to_opp_dist = passer_pos.dist( opp_pos );
      //敵とパーサーとの間の距離を計算
      const double passer_to_opp_angle = ( passer_pos - opp_pos ).th().degree();
      //敵とパーサーとの間の角度を計算
      const double diff_angle = std::abs( passer_to_receive_point_angle - passer_to_opp_angle );
      //パーサーから見てレシーバーと敵との角度の差を計算
      
      //今敵プレイヤーに対する評価値を格納
      //敵が斜線上にいるか、かつ斜線城野殿角度にいるかによって評価を変化
      //敵が斜線上にいない場合
      if( passer_to_opp_dist >= passer_to_receive_point_dist || diff_angle >= 30.0 )
	value = std::min(10.0,value);
      //敵が斜線上にいる場合
      if( passer_to_opp_dist < passer_to_receive_point_dist && diff_angle < 30.0 )
	value = std::min(5.0,value);;
      if( passer_to_opp_dist < passer_to_receive_point_dist && diff_angle < 20.0 )
	value = std::min(3.3,value);;
      if( passer_to_opp_dist < passer_to_receive_point_dist && diff_angle < 10.0 )
	value = std::min(0.0,value);;
      //敵プレイヤーすべての中でもっとも低い評価値を評価とする

      if( tmp_heuristics_conf > EnemyCont( opp ) )
	//全プレイヤーの中で最小値を評価
	tmp_heuristics_conf = EnemyCont( opp );
    }

  if( get_player_flag )
    heuristics_Conf = tmp_heuristics_conf;

  return value;	//差を細かくするために大きめの値で評価し、後で[0,10]にする
};

double
HeuristicsMethod::DistToGoal(  rcsc::PlayerAgent * agent
			       , const rcsc::PlayerObject * passer
			       , rcsc::Vector2D goToPoint )
{
  double value = 0.0;
  
  const rcsc::Vector2D goal_pos(52.5, 0.0);
  //ゴールの位置座標
  const double target_dist = goToPoint.dist( goal_pos );
  //レシーバーとゴールとの距離を取得
  
  if( target_dist < 1.0 )
    //距離差が1m以下(小数)の場合、計算のため分離する
    {
      value = 10.0;
    }else
    {
      value = 500 / ( target_dist * target_dist ); // 10>value>1/5 要見当
      //評価値を計算
    }
  
  return value;	//差を細かくするために大きめの値で評価し、後で[0,10]にする
};

double
HeuristicsMethod::DistToNearestEnemy( rcsc::PlayerAgent * agent
				      , const rcsc::PlayerObject * passer
				      , rcsc::Vector2D goToPoint )
{
  double value = 0.0;
  
  /*** ヒューリスティクスで使用されるプレイヤーの信頼性を考慮するための変数 ***/
  double tmp_heuristics_conf = 1.0;		//一時的に信頼性を保存
  bool get_player_flag = false;	//プレイヤー情報を利用したかのフラグ
  
  double min_dist = 1000.0;
  //敵とレシーバーとの最小距離を格納
  for(int i=1; i<=MAX_OPPONENTS; i++)
    {
      const rcsc::AbstractPlayerObject * opp = getOpp( agent, i );
      //敵プレイヤーを取得
      if( !opp )
	//敵情報が取得できなければ次の選手へ
	continue;
      get_player_flag = true;
      //プレイヤー情報を使ったのでフラグをonにする
      const rcsc::Vector2D opp_pos = opp->pos();
      //敵の位置情報を取得
      const double opp_dist = goToPoint.dist( opp_pos );
      //敵とレシーバーとの間の距離を計算
      
      if( opp_dist < min_dist )
	//最小値を更新
	min_dist = opp_dist;
      if( tmp_heuristics_conf > EnemyCont( opp ) )
	//全プレイヤーの中で最小値を評価
	tmp_heuristics_conf = EnemyCont( opp );
    }
  if( min_dist == 1000.0 )
    //敵チームの情報が取得できなければ評価値を0とする
    {
      value = 0;
    }else if( min_dist > 30.0 )
    //敵がレシーバーの周りにいない場合
    {
      value = 30.0;
    }else
    {
      if( min_dist < 1.0 )
	//min_distが1以下の場合、計算のため1.0に調整
	min_dist = 1.0;
      value = 30.0 - 30.0 / min_dist;
      //min_distの距離が短ければ評価を小さくする
    }
  
  if( get_player_flag )
    heuristics_Conf = tmp_heuristics_conf;
  
  return value / 3.0;
};

double
HeuristicsMethod::DistToNearestTeammate( rcsc::PlayerAgent * agent
					, const rcsc::PlayerObject * passer
					, rcsc::Vector2D goToPoint )
{
  double value = 0.0;
  
  /*** ヒューリスティクスで使用されるプレイヤーの信頼性を考慮するための変数 ***/
  double tmp_heuristics_conf = 1.0;	//一時的に信頼性を保存
  bool get_player_flag = false;	//プレイヤー情報を利用したかのフラグ
  
  double min_dist = 1000.0;
  //味方とレシーバーとの最小距離を格納
  for(int i=1; i<=MAX_TEAMMATES; i++)
    {
      const rcsc::AbstractPlayerObject * team = getTeam( agent, i );
      //味方プレイヤーを取得
      if( !team )
	//味方情報が取得できなければ次の選手へ
	continue;
      get_player_flag = true;
      //プレイヤー情報を使ったのでフラグをonにする
      const rcsc::Vector2D team_pos = team->pos();
      //味方の位置情報を取得
      const double team_dist = goToPoint.dist( team_pos );
      //味方とレシーバーとの間の距離を計算
      
      if( team_dist < min_dist )
	//最小値を更新
	min_dist = team_dist;
      if( tmp_heuristics_conf > EnemyCont( team ) )
	//全プレイヤーの中で最小値を評価
	tmp_heuristics_conf = EnemyCont( team );
    }
  if( min_dist == 1000.0 )
    //味方チームの情報が取得できなければ評価値を0とする
    {
      value = 0;
    }else if( min_dist > 30.0 )
    //味方がレシーバーの周りにいない場合
    {
      value = 30.0;
    }else
    {
      if( min_dist < 1.0 )
	//min_distが1以下の場合、計算のため1.0に調整
	min_dist = 1.0;
      else
	value = 30.0 - 30.0 / min_dist;
      //min_distの距離が短ければ評価を小さくする
    }
  
  if( get_player_flag )
    heuristics_Conf = tmp_heuristics_conf;
  
  return value / 3.0;
};

double
HeuristicsMethod::EnemyNum( rcsc::PlayerAgent * agent
			   , const rcsc::PlayerObject * passer
			   , rcsc::Vector2D goToPoint )
{
  double value = 0.0;

  /*** ヒューリスティクスで使用されるプレイヤーの信頼性を考慮するための変数 ***/
  double tmp_heuristics_conf = 1.0;	//一時的に信頼性を保存
  bool get_player_flag = false;	//プレイヤー情報を利用したかのフラグ
  
	const double limit_dist = 10.0;
	//半径10m以内にいる敵を数える

	int opp_num = 0;	//数えた敵の数を格納
	bool skip_flag = false;	//continue命令を実行したかを判断
	//opp_num=0の時、敵がいなかったのか、敵チーム情報が取得できなかったかで評価を変えるため
	for(int i=1; i<=MAX_OPPONENTS; i++)
	{
		const rcsc::AbstractPlayerObject * opp = getOpp( agent, i );
		//敵プレイヤーを取得
		if( !opp )
		  //敵情報が取得できなければ次の選手へ
		  continue;
		get_player_flag = true;
		//プレイヤー情報を使ったのでフラグをonにする
		skip_flag = true;
		//敵情報を取得したのでtrue
		const rcsc::Vector2D opp_pos = opp->pos();
		//敵の位置情報を取得
		const double opp_dist = goToPoint.dist( opp_pos );
		//敵とレシーバーとの間の距離を計算

		if( opp_dist <= limit_dist )
		  //範囲内にいれば数を増やす
		  opp_num++;
		if( tmp_heuristics_conf > EnemyCont( opp ) )
		  //全プレイヤーの中で最小値を評価
		  tmp_heuristics_conf = EnemyCont( opp );
	}
	if( opp_num == 0 )
	{
		if( !skip_flag )
		  //敵情方が取得できなかった場合
		  value = 0.0;
		else
		  //敵が範囲内にいない場合
		  value = 10.0;
	}else
	//敵が範囲内に存在する場合
	{
	  value = 10 - 10 * opp_num / 11 ;
	  //評価値を計算
	}

	if( get_player_flag )
	  heuristics_Conf = tmp_heuristics_conf;

	return value;
};

double
HeuristicsMethod::PasserPosCount( rcsc::PlayerAgent * agent
				 , const rcsc::PlayerObject * passer
				 , rcsc::Vector2D goToPoint )
{
	double value = 0.0;
	const double PLAYER_RELIA = 0.9;	//信頼性の閾値、これ以下は信頼に値しないとする
	//ようは10カウント以上の信頼性は信頼しない

	double passer_posReli = 1 - passer->posCount() / 100;
	//位置情報のカウントサイクルから信頼性を取得
	if( passer_posReli <= PLAYER_RELIA )
	  //閾値以下ならば評価は0とする
	{
	  value = 0.0;
	}else
	{
	  value = 10 * passer_posReli;	//[0.9, 10]
	  //評価値を計算
	}

	return value;
};

double
HeuristicsMethod::CheckBallSpeed( rcsc::PlayerAgent * agent
				 , const rcsc::PlayerObject * passer
				 , rcsc::Vector2D goToPoint )
{
	double value = 0.0;

	//実際にパスを行うときの速度を計算します
	double dist = ( goToPoint - agent->world().ball().pos() ).r();	//ボールを蹴る距離
	double end_speed = 1.5;	//終了速度
	double first_speed = 100.0; //開始速度
	do
	{
	  first_speed
	    = rcsc::calc_first_term_geom_series_last
	    ( end_speed,
	      dist,
	      rcsc::ServerParam::i().ballDecay() );
	  if ( first_speed < rcsc::ServerParam::i().ballSpeedMax() )
	    //開始速度がボールの最大速度以下ならば終了
	    {
	      break;
	    }
	  end_speed -= 0.1;
	  //終了速度を減らして再トライ
	}
	while ( end_speed > 0.8 );
	if ( first_speed > rcsc::ServerParam::i().ballSpeedMax() )
	  //初速度がボールの最大速度以上の場合、(届かない）は評価を0にする
	  value = 0.0;
	else
	{
	  double speed_diff = rcsc::ServerParam::i().ballSpeedMax() - first_speed;
	  //速度差から計算小さいほど楽に蹴ることが出きると判断
	  //現サーバではServerParam::i().ballSpeedMax()=3.0
	  value = 10.0 - 5.0 * speed_diff / rcsc::ServerParam::i().ballSpeedMax();
	  //first_speed=0.0の場合でも半分は評価する
	}

	//信頼性を計算
	heuristics_Conf = BallCont( &agent->world().ball() );

	return value;
};

double
HeuristicsMethod::DistToPasser( rcsc::PlayerAgent * agent
			       , const rcsc::PlayerObject * passer
			       , rcsc::Vector2D goToPoint )
{
	double value = 0.0;

	double target_dist = passer->pos().dist( goToPoint );	//距離を測る
	if( target_dist == 0.0 )
	  //距離差が0である場合
	{
	  value = 40.0;
	}else if( target_dist < 1.0 )
	  //距離差が1m以下(小数)の場合、計算のため分離する
	{
	  value = 35.0;
	}else
	{
	  value = 35.0 / target_dist;
	  //評価値を計算
	}

	return value / 4.0;	//差を細かくするために大きめの値で評価し、後で[0,10]にする
};


double
HeuristicsMethod::DistToReceiver( rcsc::PlayerAgent * agent
				 , const rcsc::PlayerObject * passer
				 , rcsc::Vector2D goToPoint )
{
	double value = 0.0;

	double target_dist = agent->world().self().pos().dist( goToPoint );	//距離を測る
	if( target_dist == 0.0 )
	  //距離差が0である場合
	{
	  value = 40.0;
	}else if( target_dist < 1.0 )
	  //距離差が1m以下(小数)の場合、計算のため分離する
	{
	  value = 35.0;
	}else
	{
	  value = 35.0 / target_dist;
	  //評価値を計算
	}

	return value / 4.0;	//差を細かくするために大きめの値で評価し、後で[0,10]にする
};

double
HeuristicsMethod::DiffDistToBallAndDistToReceiver( rcsc::PlayerAgent * agent
						  , const rcsc::PlayerObject * passer
						  , rcsc::Vector2D goToPoint )
{
	double value = 0.0;

	double dist_receiver = goToPoint.dist( agent->world().self().pos() );
	double dist_ball = goToPoint.dist( agent->world().ball().pos() );
	//パス地点からのボールとレシーバーの距離を計算
	double diff_dist = dist_receiver - dist_ball;
	//距離差を計算

	if( diff_dist < 0.0 )
	//レシーバーがボールよりも遠ければ評価を0とする
	{
		value = 0;
	}else
	{
		//距離差によって評価を変える
		if( diff_dist >= 0 )
			value = 10.0;
		if( diff_dist >= 10.0 )
			value = 20.0;
		if( diff_dist >= 20.0 )
			value = 30.0;
		if( diff_dist >= 30.0 )
			value = 40.0;
	}

	return value / 4.0;
};

double
HeuristicsMethod::DiffTakeTimeToBallAndTakeTimeToReceiver( rcsc::PlayerAgent * agent
							  , const rcsc::PlayerObject * passer
							  , rcsc::Vector2D goToPoint )
{
	double value = 0.0;

	double dist_receiver = goToPoint.dist( agent->world().self().pos() );
	double dist_ball = goToPoint.dist( agent->world().ball().pos() );
	//パス地点からのボールとレシーバーの距離を計算
	int time_receiver = dist_receiver / rcsc::ServerParam::i().defaultPlayerSpeedMax();
	int time_ball = dist_ball / rcsc::ServerParam::i().ballSpeedMax();
	//最高速度での移動時間を計算
	int diff_time = time_ball - time_receiver;

	if( diff_time < 0 )
	//レシーバーがボールよりも遠ければ評価を0とする
	{
		value = 0;
	}else if( diff_time == 0 )
	{
		value = 10.0;
	}else
	{
		value = 10.0 / diff_time;
	}

	return value;
};

double
HeuristicsMethod::DistToOffsideLineXWithSelf( rcsc::PlayerAgent * agent
					     , const rcsc::PlayerObject * passer
					     , rcsc::Vector2D goToPoint )
{
	double value = 0.0;

	double dist_offsideLineX = agent->world().offsideLineX() - agent->world().self().pos().x;
	//パス地点とオフサイドラインとの距離を調べる

	if( dist_offsideLineX < 0.0 )
	//レシーバーがオフサイドラインよりも後ろならば評価を0とする
	{
		value = 0;
	}else
	{
		//距離差によって評価を変える
		if( dist_offsideLineX >= 0 )
			value = 10.0;
		if( dist_offsideLineX > 10.0 )
			value = 20.0;
		if( dist_offsideLineX > 20.0 )
			value = 30.0;
		if( dist_offsideLineX > 30.0 )
			value = 40.0;
	}

	return value / 4.0;
};

double
HeuristicsMethod::DistToOffsideLineX( rcsc::PlayerAgent * agent
				     , const rcsc::PlayerObject * passer
				     , rcsc::Vector2D goToPoint )
{
	double value = 0.0;

	double dist_offsideline = goToPoint.x - agent->world().offsideLineX();
	//パス地点とオフサイドラインとの距離を調べる

	if( agent->world().isOurRight() )
	//サイドが逆ならマイナスをかける
		dist_offsideline *= -1.0;

	if( dist_offsideline < 0.0 )
	//パス地点がオフサイドラインよりも前であれば評価を0とする
	{
		value = 0;
	}else
	{
		//距離差によって評価を変える
		if( dist_offsideline >= 0 )
			value = 10.0;
		if( dist_offsideline > 10.0 )
			value = 20.0;
		if( dist_offsideline > 20.0 )
			value = 30.0;
		if( dist_offsideline > 30.0 )
			value = 40.0;
	}

	return value / 4.0;
};
