#ifndef HEURISTICS_METHOD_H
#define HEURISTICS_METHOD_H
#include "basic_learn.h"

class HeuristicsMethod : public BasicLearn
{
 public:
  double getHeuristics(rcsc::PlayerAgent * agent
		       , int action
		       , int heuristics_num);
  //ヒューリスティクスheuristics_numの行動actionの値を返す
  
  /*** ヒューリスティクス ***/
  //移動場所番号actionと現在ボールを持っている味方について判断する
  //@return 評価値
  //@param passer : 現在ボールを持っているプレイヤークラス
  //@param goToPoint :行動actionの指すレシーブポイントの座標

 protected:
  double EnemyInPassCourse( rcsc::PlayerAgent * agent
			    ,const rcsc::PlayerObject * passer
			    ,rcsc::Vector2D goToPoint );
  //パスコースに敵がいるかを評価

  double DistToGoal(  rcsc::PlayerAgent * agent
			  ,const rcsc::PlayerObject * passer
			  ,rcsc::Vector2D goToPoint );
  //レシーバーとゴールとの距離差を評価

  double DistToNearestEnemy( rcsc::PlayerAgent * agent
			,const rcsc::PlayerObject * passer
			,rcsc::Vector2D goToPoint );
  //もっとも近い敵プレイヤーとの距離を評価
  
  double DistToNearestTeammate( rcsc::PlayerAgent * agent
			   ,const rcsc::PlayerObject * passer
			   ,rcsc::Vector2D goToPoint );
  //もっとも近い味方プレイヤーとの距離を評価
  
  double EnemyNum( rcsc::PlayerAgent * agent
		   ,const rcsc::PlayerObject * passer
		   ,rcsc::Vector2D goToPoint );
  //レシーバーの周りにいる敵の数を評価
  
  double PasserPosCount( rcsc::PlayerAgent * agent
			 ,const rcsc::PlayerObject * passer
			 ,rcsc::Vector2D goToPoint );
  //パーサーの位置情報の信頼性を評価

  double CheckBallSpeed( rcsc::PlayerAgent * agent
			 ,const rcsc::PlayerObject * passer
			 ,rcsc::Vector2D goToPoint );
  //レシーバーがボールをパーサーの下に蹴るのにかかるパワーの評価

  double DistToPasser( rcsc::PlayerAgent * agent
			       ,const rcsc::PlayerObject * passer
			       ,rcsc::Vector2D goToPoint );
  //パッサーと目標ポイントの距離を評価
  
  double DistToReceiver( rcsc::PlayerAgent * agent
				 ,const rcsc::PlayerObject * passer
				 ,rcsc::Vector2D goToPoint );
  //レシーバーと目標ポイントの距離を評価
  
  double DiffDistToBallAndDistToReceiver( rcsc::PlayerAgent * agent
					 ,const rcsc::PlayerObject * passer
					 ,rcsc::Vector2D goToPoint );
  //目標ポイントからのボールとレシーバーとの距離差を評価

  double DiffTakeTimeToBallAndTakeTimeToReceiver( rcsc::PlayerAgent * agent
					     ,const rcsc::PlayerObject * passer
					     ,rcsc::Vector2D goToPoint );
  //目標ポイントからのボールとレシーバーとの移動時間差を評価
  
  double DistToOffsideLineXWithSelf( rcsc::PlayerAgent * agent
				    ,const rcsc::PlayerObject * passer
				    ,rcsc::Vector2D goToPoint );
  //レシーバーがどれだけオフサイドラインより前であるかを調べる
  //オフサイド防止用

  double DistToOffsideLineX( rcsc::PlayerAgent * agent
				 ,const rcsc::PlayerObject * passer
				 ,rcsc::Vector2D goToPoint );
  //目標ポイントがどれだけオフサイドラインより後ろであるかを調べる
};
#endif
