/*** インクルードファイル ***/
#include "pass_receiver.h"
#include<rcsc/math_util.h>
#include<rcsc/player/free_message.h>
#include<cmath>
//数学計算用

PassReceiver::PassReceiver()
//コンストラクタ
  :goToPoint( 0.0 , 0.0 )
  , before_kicker_pos( 0.0, 0.0 ), after_kicker_pos( 0.0, 0.0 )
  , before_ball_pos( 0.0, 0.0 ) ,after_ball_pos( 0.0, 0.0 )
{
  SettingPoint_flag=false;
  say_cycle=-1;
  passer_unum=-1;
  get_msg_cycle=-1;
  passing_flag=false;
  catch_ball_flag=false;
  before_score=0;
  before_offside=0.0;
  after_offside=0.0;
  last_opp_hold_cycle=0;
  receive_count=0;
  catch_count=0;
};

PassReceiver::~PassReceiver()
//デストラクタ
{
  fdebug << receive_count << "," << catch_count << std::endl;
};

rcsc::Vector2D PassReceiver::execute( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel & wm = agent->world();
  
  if( ( LIMIT_CYCLE < ( wm.time().cycle() - say_cycle ) ) 
      || !SettingPoint_flag )
    //移動最大数を越えるか、移動座標を決定していなければ行う
    {
      //スルーパスなので敵陣に向かってのパスのみについて考えることにしました
      
      SettingPoint_flag = true;		//移動開始
      rcsc::Vector2D goToPoint( 0.0, 0.0 );   //移動すべき座標
      int action = Boltzmann( agent );
      BasicLearn::action = action;	//行動actionを記録
      
      int action_x = action % 6;	//x座標を調べる(x方向にセルが6個)
      int action_y = action / 6;	//y座標を調べる

      const double offsetX = 15.0 + 2.5;  //レシーブ位置は15<x<45
      const double offsetY = -30.0 + 2.5; //,-30<y<30
      goToPoint.x = action_x * 5.0 + offsetX;
      goToPoint.y = action_y * 5.0 + offsetY;
      this->goToPoint = goToPoint;	//座標を記録
      say_cycle = wm.time().cycle();    //最初にメッセジを送ったサイクルを記憶
      //学習値を保存する
      PassReceiver::set_learn_param( agent );
    }
  if( SettingPoint_flag 
      && wm.time().cycle() - say_cycle <= 10 )
    {
      //5サイクルほどパスメッセジを送ってみる
      //passerにメッセジを送る
      //メッセジ送信
      sayMessage( agent, action );
    }
  return this->goToPoint;	//座標を返す
};

bool
PassReceiver::init( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel & wm = agent->world();

  if( !wm.audioMemory().freeMessage().empty()
      && ( wm.time().cycle() - say_cycle <= 20 ) )
    //メッセージを受け取ります
    //但し、メッセジを送ってから２０サイクル以内であることを条件に追加
    {
      std::string msg;		//メッセジ用バッファ
      if( get_msg_cycle 
	  != wm.audioMemory().freeMessageTime().cycle() )
	{
	  for(unsigned int i = 0; 
	      i < wm.audioMemory().freeMessage().size();
	      i++)
	    {
	      msg = wm.audioMemory().freeMessage()[i].message_;
	      if(msg.size() != 4)
		//パーサーからの答えの場合受けとります
		continue;
	      
	    }
	  if( !wm.audioMemory().freeMessage().empty() )
	    {
	      get_msg_cycle 
		= wm.audioMemory().freeMessageTime().cycle();
	      if( msg.substr( 2, 2 ) == "OK" )
		//返事かを確認
		{
		  int to_pass_unum 
		    = boost::lexical_cast< int >( msg.substr( 0, 2 ) );
		  if( to_pass_unum != wm.self().unum() )
		    {
		      //自身へのパスでなければ別の場所に移動します
		      SettingPoint_flag = false;
		      //学習を中断する
		    }
		  else if( SettingPoint_flag )
		    {
		      //debug(tostring(current_cycle)+":learn");
		      //自身へのパスの場合
		      if( !learn_flag )
			//学習前の場合
			{
			  learn_flag =  true ;	//学習フラグをオンにします
			  first_action_cycle = agent->world().time().cycle();
			  //最初の行動の時間を調べる
			  passer_unum 
			    =wm.teammatesFromBall().front()->unum();
			  if( wm.teammate( passer_unum ) )
			    before_kicker_pos 
			      = wm.teammate( passer_unum )->pos();
			  else
			    before_kicker_pos = rcsc::Vector2D( 0.0, 0.0 );
			  before_offside = wm.offsideLineX();
			  before_ball_pos = wm.ball().pos();
			  //エピソード開始時の自身の位置を保持
			  before_score = wm.gameMode().scoreLeft();
			  //学習前のスコアを記憶します
			}
		      passer_unum 
			= wm.teammatesFromBall().front()->unum();
		      passing_flag = true;	//パス中フラグをONにする
		      SettingPoint_flag = false;//レシーブフラグをオフにします
		      last_action_cycle = wm.time().cycle();
		      episode++;		//エピソード長を増やします
		      for(int i=0; i<MAX_ACTION; i++)
			{
			  //現エピソード中の方策の値を保存
			  boltz_Policy[i].push_back(tmp_boltz_Policy[i]);
			  for(int j=0; j<MAX_HEURISTICS; j++)
			    //現エピソード中のヒューリスティクスを追加
			    boltz_Heuristics[i][j].push_back( tmp_boltz_Heuristics[i][j] );
			}
		      //実証データ確保
		      receive_count++;
		      debug("receive_count:"
			    +tostring(receive_count)
			    +" --- start ---");
		    }
		}
	    }else{
	    SettingPoint_flag = false;
	  }
	}
    }
  return BasicLearn::init( agent );
};

bool
PassReceiver::set_learn_param( rcsc::PlayerAgent * agent )
{
  for(int i=0; i<MAX_ACTION; i++)
    {
      for(int j=0; j<MAX_HEURISTICS; j++)
	{
	  tmp_boltz_Heuristics[i][j] = getHeuristics(agent, i, j);
	  //ヒューリスティックの一時的保存
	}
      tmp_boltz_Policy[i] = policy[i];	//方策（確率)の一時的保存
    }
  return true;
};

bool
PassReceiver::isEndEpisode( rcsc::PlayerAgent * agent )
{
  if( catch_ball_flag )
    //自身がパスを受け取った場合学習を継続
    {
      if( ( agent->world().gameMode().scoreLeft() - before_score != 0 ) )
	//&& agent->world().gameMode().type() == rcsc::GameMode::KickOff_ )
	//シュートが入れば学習を行う
	{
	  SettingPoint_flag = false;
	  learn_flag = false;
	  passing_flag = false;
	  catch_ball_flag = false;
	  return true;
	}
      if( agent->world().existKickableOpponent() 
	  || agent->world().existKickableOpponent() )
	//敵のプレイヤーがボールを保持している場合、学習を終了します
	{
	  if( agent->world().teammatesFromBall().front()->distFromBall()
	      > agent->world().opponentsFromBall().front()->distFromBall() )
	    {
	      SettingPoint_flag = false;
	      learn_flag = false;
	      passing_flag = false;
	      catch_ball_flag = false;
	      return true;
	    }
	}
    }
  
  else
    {
      bool isKickable = agent->world().self().isKickable();
      bool existKickableTeammate = false;		//どちらのチームが蹴れるか判断
      bool existKickableOpp = false;
      if ( agent->world().existKickableTeammate()
	   && agent->world().teammatesFromBall().front()->distFromBall()
	   < agent->world().ball().distFromSelf() )
	{
	  isKickable = false;
	}
      //調査中：エピソードはパスをしてから誰かがボールをキャッチするまでにするか?
      //調査追加：自身がとった場合は他プレイヤーにとられるまでに延長？
      if( agent->world().existKickableTeammate()
	  || agent->world().existKickableOpponent() )
	{
	  if( agent->world().teammatesFromBall().front()->distFromBall()
	      <= agent->world().opponentsFromBall().front()->distFromBall() )
	    {
	      existKickableTeammate = true;
	    }
	  else
	    {
	      existKickableOpp = true;
	    }
	}
      if( isKickable )
	{
	  debug(tostring(agent->world().time().cycle())+":get ball");
	  debug("x:"+tostring(goToPoint.x)+",y:"+tostring(goToPoint.y));
	  after_ball_pos = agent->world().ball().pos();
	  after_offside = agent->world().offsideLineX();
	  catch_ball_flag = true;
	  passing_flag = false;
	  //実証データ確保
	  if( after_ball_pos.x - after_offside >= -1.0 )
	    {
	      catch_count++;
	      debug("catch_count:"+tostring(catch_count));
	    }else
	    {
	      debug("no catch");
	      debug("ball_x:"+tostring(after_ball_pos.x));
	      debug("offside:"+tostring(after_offside));
	    }
	  return false;
	}else if(( existKickableTeammate
		   && agent->world().teammatesFromBall().front()->unum() 
		   != passer_unum )
		 || existKickableOpp )
	//他の選手がとったら中断？
	{
	  finish_learn( agent );
	  SettingPoint_flag = false;
	  learn_flag = false;
	  passing_flag = false;
	  catch_ball_flag = false;
	  return false;
	}
      
    }
  if( agent->world().gameMode().type() == rcsc::GameMode::FirstHalfOver
      || agent->world().gameMode().isPenaltyKickMode()
      || agent->world().gameMode().isGameEndMode()
      )
    //ハーフタイム終了時に学習中であれば中断する
    //ペナルティ、試合終了自も追加
    {
      finish_learn( agent );
      SettingPoint_flag = false;
      learn_flag = false;
      passing_flag = false;
      catch_ball_flag = false;
      return false;
    }
  
  return false;
};

bool
PassReceiver::isTeamNearestBall( rcsc::PlayerAgent * agent )
{
  double min_dist = 1000.0;;		//ボールからの最小距離を記録
  for(int i=1; i<=MAX_TEAMMATES; i++)
    //味方チームでのボールへの最短距離を計算
    {
      const rcsc::AbstractPlayerObject * team = getTeam( agent, i );
      if( !team )
	//情報が取得できなければ飛ばす
	continue;
      if( team->distFromBall() < min_dist )
	min_dist = team->distFromBall();
    }
  
  for(int i=1; i<=MAX_OPPONENTS; i++)
    //敵が見かたよりもボールに近いかを判断
    {
      const rcsc::AbstractPlayerObject * opp = getOpp( agent, i );
      if( !opp )
	//情報が取得できなければ飛ばす
	continue;
      if( opp->distFromBall() < min_dist )
	return false;
    }
  /*
    if( agent->world().existKickableOpponent() )
    return false;
  */
  return true;
};

bool
PassReceiver::isTeamHoldingBall( rcsc::PlayerAgent * agent )
{
  if( !isTeamNearestBall( agent ) )
    //敵チームがボールを保持と判断
    last_opp_hold_cycle = agent->world().time().cycle();
  
  if( agent->world().time().cycle() - last_opp_hold_cycle >= 5  )
    //敵が保持してから5サイクル以上経った=味方が5サイクル以上保持している
    return true;
  return false;
};

bool
PassReceiver::setReward( rcsc::PlayerAgent * agent, double base )
{
  double value_score;
  double diffBallAndOffsideLine;
  double value_diff_offsideLine;
  double distFromGoal;
  double value_dist_from_goal;
  //条件としてスルーパスの学習なのでスルーパスをすることを報酬として優先する
  
  //パスを受け取った位置と最終ディフェンスラインとの距離差を測ります
  diffBallAndOffsideLine = after_ball_pos.x - after_offside;
  if( diffBallAndOffsideLine < 0.0 )
    if( std::abs(diffBallAndOffsideLine) > 1 )
      value_diff_offsideLine = 1 / std::abs(diffBallAndOffsideLine); //< -1
    else
      value_diff_offsideLine = 1.0;  //-1 < , < 1
  else
    value_diff_offsideLine = diffBallAndOffsideLine + 10.0;  //1 <
  //パスを受け取った位置とゴールとの距離を測る
  //条件変更
  //スルーパス成功時のみ報酬を与える
  if( diffBallAndOffsideLine >= -1.0 )
    {
      distFromGoal = after_ball_pos.dist( rcsc::Vector2D( 50.0, 0.0 ) );
      if( distFromGoal < 1.0 )
	value_dist_from_goal = 20.0;
      else
	value_dist_from_goal = 20.0 / distFromGoal;
    }
  else
    {
      distFromGoal = 1000.0;
      value_dist_from_goal = 0.0;
    }
  //シュートが入ったら報酬を与える
  value_score = agent->world().gameMode().scoreLeft() - before_score;
  value_score *= 30;
  
  double value =
    value_score
    + value_dist_from_goal
    + value_diff_offsideLine;
  
  r = value;
  
  //ファイル出力
  std::string buff;
  std::ofstream out_r("./csv/r.csv", std::ios::out | std::ios::app );
  buff += tostring(game_count) + ",";
  buff += tostring(value_score) + ",";
  buff += tostring(distFromGoal) + ",";
  buff += tostring(diffBallAndOffsideLine) + ",";
  buff += tostring(r);
  out_r << buff << std::endl;
  out_r.close();
  
  return true;
};

//実験実装
void
PassReceiver::sayMessage( rcsc::PlayerAgent * agent, const int action )
{
  //msg = aahh
  //@param a=移動先番号
  //@param h=評価値合計
  //length = 4
  char buff[254];	//messageをstringに書き込むために使う
  sprintf(buff, "%02d", action );
  std::string msg(buff);
  //debug(tostring(current_cycle)+":receiver:say:"+msg);
  agent->addSayMessage( new rcsc::FreeMessage<2>( msg ) );
  return;
};

//ゲッター、セッター
bool
PassReceiver::isPassing()
{
  return passing_flag;
};
