/*** パスレシーバー学習クラス
 * 味方がボールを持っているときにスペースには知ることを学習する ***/

#ifndef PASS_RECEIVER_H
#define PASS_RECEIVER_H

#include "heuristics_method.h"

class PassReceiver : 
public HeuristicsMethod
{
  /*** 変数 ***/
  //execute用
  static const int LIMIT_CYCLE = 10;		//移動にかける最大サイクル数
  //これ以上かかるのであれば別の場所に移動
  bool SettingPoint_flag;			//移動先を決定したかを判断
  int say_cycle;	       //レシーバーが最後にメッセジを送ったサイクル
  int passer_unum;
  //init用
  int get_msg_cycle;	       //最後にメッセジを受け取ったサイクル
  bool passing_flag;	       //現在パス中かのフラグ
  bool catch_ball_flag;	       //自身がパスをとれたかの判断
  
  rcsc::Vector2D goToPoint;    //移動予定の座標
  //レシーブの特性上一時保存し、受け取ったら入れてやる
  double tmp_boltz_Heuristics[MAX_ACTION][MAX_HEURISTICS]; 
  //ヒューリスティックの一時的保存
  double tmp_boltz_Policy[MAX_ACTION];	      //方策（確率)の一時的保存
  
  //報酬計算用
  int before_score;		        //学習前の自チームの得点
  rcsc::Vector2D before_kicker_pos;	//パス前のキッカーの座標
  rcsc::Vector2D after_kicker_pos;	//パス後のキッカーの座標
  rcsc::Vector2D before_ball_pos;	//パス前のボールの座標
  rcsc::Vector2D after_ball_pos;	//パス後のボールの座標
  double before_offside;		//パス前のオフサイドライン
  double after_offside;			//パス後のオフサイドライン
  
  
  //チームのボール保持確認用変数
  int last_opp_hold_cycle;		//敵チームが最後にボールに触れたサイクル
  
  //実証用変数
  int receive_count;	//レシーブ回数
  int catch_count;      //キャッチ回数
  
 public:
  PassReceiver();
  ~PassReceiver();
  rcsc::Vector2D execute( rcsc::PlayerAgent * agent );
  //実行メソッド
  
  bool init( rcsc::PlayerAgent * agent );
  //移動完了フラグ,及び移動後のサイクル数を記録するためにpublicに定義
  //毎サイクル実行
  //@return BasicLearnの返り値を返す

  bool player_init( rcsc::PlayerAgent * agent );
  //プレイヤー行動の初期化

  bool set_learn_param( rcsc::PlayerAgent * agent );
  //BasicLearnの物と同じ役割
  //一時的にヒューリスティクスの値を保存する必要があったのでoverrideした
  
 private:
  bool isEndEpisode( rcsc::PlayerAgent * agent );
  //終了条件を記述

  bool isTeamNearestBall( rcsc::PlayerAgent * agent );
  //味方がもっともボールに近いかを判断
  //@return 味方が近ければtrue 敵が近ければfalse

  bool isTeamHoldingBall( rcsc::PlayerAgent * agent );
  //敵がボールを保持しているかを判断
  //5カウント以上isTeamNearestBallがfalseならばボールを保持と考える

  bool setReward( rcsc::PlayerAgent * agent, double base = 1.0 );
  //報酬を定義

    
  
  //実験動作
  void sayMessage( rcsc::PlayerAgent * agent, const int action );
  //@param agent 必須
  //@param aciton 移動座標番号
  //@param heu_value 評価値の値
  
 public:
  //ゲッター、セッター
  bool isPassing();
};

#endif /* PASS_RECEIVER_H */
