// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_attacker.h"

#include "strategy.h"

#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_move.h"
#include "bhv_basic_tackle.h"

#include <rcsc/action/body_intercept.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_goalie_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/geom/sector_2d.h>
#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_ball.h>

#include <rcsc/action/body_smart_kick.h>
#include <rcsc/math_util.h>

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleAttacker::execute( rcsc::PlayerAgent * agent )
{
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().teammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }
	
    if ( kickable )
    {
        doKick( agent );
    }
    else
    {
        doMove( agent );
    }

}

/*-------------------------------------------------------------------*/
/*!

*/

void
RoleAttacker::doKick( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerObject *opp = wm.getOpponentNearestToSelf(5);
    const rcsc::PlayerObject *goalie = wm.getOpponentGoalie();
    rcsc::Vector2D pass_point;

    switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
      case Strategy::BA_ShootChance:
      {	//こいつはここだとパスしないんだね。まぁいいか

	if(FSShoot(agent))
	  break;

	rcsc::Vector2D dribble_target(52.5
				      ,std::min(wm.self().pos().absY()
						,rcsc::ServerParam::i().goalHalfWidth() - 0.5 - rand()%3));
	if(wm.self().pos().y < 0)
	  dribble_target.y *= -1;
	rcsc::Body_Dribble2008(dribble_target,3.0,
			       rcsc::ServerParam::i().maxDashPower(),2,false
			       ).execute( agent );
	//printf("Drible!\n");
	agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );//t-suzuki
	break;
      }
      Bhv_BasicOffensiveKick().execute( agent );
      break;
    case Strategy::BA_Cross:
      //-----クロス時-----
      if(wm.self().pos().x > 30
	 && wm.self().pos().absY() < 20)
	if(wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(50,0)
						 ,20)
				  ,5,false) > 1
	   &&FSShoot(agent))
	  break;
      if(countKickableCycleAtSamePoint(agent)<20){
	rcsc::Body_Dribble2008(rcsc::Vector2D(52.0,wm.self().pos().y*(rand()%5)/4),
			   5.0,rcsc::ServerParam::i().maxDashPower(),
			   8,false).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
	break;
      }else{
	if(rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL)){
	  rcsc::Body_Pass().execute(agent);
	  agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
	  break;
	}
      }
      Bhv_BasicOffensiveKick().execute( agent );
      break;
      //--------------------------------------------------------
    case Strategy::BA_OffMidField:
    case Strategy::BA_DefMidField:
      if(wm.self().pos().x > 30
	 && wm.self().pos().absY() < 20)
	if(wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(50,0)
						 ,20)
				  ,5,false) > 1
	   &&FSShoot(agent))
	  break;
      if(opp!=NULL && (wm.self().pos().dist(opp->pos())<2.5) &&
	 (wm.self().pos().x < opp->pos().x))
	{
	  //---Pass---
	  if(rcsc::Body_Pass::get_best_pass(wm,&pass_point,NULL,NULL))
	    {
	      if(pass_point.x - 3 >wm.self().pos().x
		 && (rcsc::ServerParam::i().useOffside() 
		     && pass_point.x<wm.offsideLineX()
		     || !rcsc::ServerParam::i().useOffside())
		 && wm.countOpponentsIn(rcsc::Circle2D(pass_point,2.5)
					,7,true) <=1)
		{
		  rcsc::Body_Pass().execute(agent);
		  agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
		  break;
		}
	    }
	  if(wm.self().pos().x<28.0)
	    {//---適当キック（for Corner）---
	      rcsc::Vector2D target_point(45.0,31.0);
	      if(wm.self().pos().y<0)
		target_point*=-1;
	      double first_speed = rcsc::ServerParam::i().ballSpeedMax();
	      double thr = 0.96;
	      int step = 2;
	      double end_speed = 1.5;
	      double target_dist = target_point.dist(wm.ball().pos());
	      
	      do{
		first_speed
		  = rcsc::calc_first_term_geom_series_last
		  ( end_speed,
		    target_dist,
		    rcsc::ServerParam::i().ballDecay() );
		if ( first_speed < rcsc::ServerParam::i().ballSpeedMax() )
		  {
		    break;
		  }
		end_speed -= 0.1;
	      }while ( end_speed > 0.8 );
	      
	      rcsc::Body_SmartKick(target_point,
				   first_speed,
				   first_speed * thr,
				   step).execute(agent);
	    }
	  else if(wm.self().pos().x > 32 && rand()%2)
	    {//---適当キック（for Goal） ---
	      double first_speed = rcsc::ServerParam::i().ballSpeedMax();
	      double thr = 0.96;
	      int step = 2;
	      rcsc::Body_SmartKick(rcsc::Vector2D(52.5,
						  6.5 - 13*rand()%2 ),
				   first_speed,
				   first_speed * thr,
				   step).execute(agent);
	    }

	  agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	  break;
	}
      else
	{//通常のドリブル
	  rcsc::Vector2D dribble_point=wm.self().pos();
	  dribble_point.x+=15;
	  //--Y軸的に敵がいない方へ移動--
	  if(opp!=NULL)
	    {
	      if(opp->pos().y<wm.self().pos().y)
		{
		  dribble_point.y+=8.0;
		}
	      else
		{
		  dribble_point.y-=8.0;
		}
	    }
          //--方向決定終了--
	    
	  //ドリブル開始！
	   if(countKickableCycleAtSamePoint(agent)<12){
	     //printf("%2d:kickable\n",countKickableCycleAtSamePoint(agent));
	     rcsc::Body_Dribble2008(rcsc::Vector2D(47.0,
						   wm.self().pos().y
						   * (rand()%4)/2)
				    ,5.0,rcsc::ServerParam::i().maxPower()
				    ,5,false).execute( agent );
	   }else{
	     //ドリブルできないときは、てきとークリア
	     if(FSClearBall(agent))
	       break;
	   }
	   
	   agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
	   break;
	}
      Bhv_BasicOffensiveKick().execute( agent );
      break;
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DribbleAttack:
    default:
      Bhv_BasicOffensiveKick().execute( agent );
      break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleAttacker::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    int ball_reach_step = 0;
    if ( ! wm.existKickableTeammate() && ! wm.existKickableOpponent())
    {
        ball_reach_step
            = std::min( wm.interceptTable()->teammateReachCycle(),
                        wm.interceptTable()->opponentReachCycle() );
    }
    const rcsc::Vector2D base_pos
        = wm.ball().inertiaPoint( ball_reach_step );

    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, wm.offsideLineX() - 1.0 );
    }
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: HOME POSITION=(%.2f, %.2f) base_point(%.1f %.1f)"
                        ,__FILE__,
                        home_pos.x, home_pos.y,
                        base_pos.x, base_pos.y );
    rcsc::Vector2D ball_point 
      = inertia_n_step_point( wm.ball().pos(),
			      wm.ball().vel(),
			      ball_reach_step,//cycle
			      rcsc::ServerParam::i().ballDecay() );

    //各種サイクルの取得
    int self_cycle = wm.interceptTable()->selfReachCycle();
    int mate_cycle = wm.interceptTable()->teammateReachCycle();
    int opp_cycle = wm.interceptTable()->opponentReachCycle();

      //---------------
     // エリアごとの処理
    switch ( Strategy::get_ball_area( base_pos ) ) {
    case Strategy::BA_ShootChance:
    /*	動きがワンパターンなのは避けたいなぁ。
    //-----ゴール前-----
    //必殺！無差別タックル
    //(チームプレー的な事が難しいのだからこの方がいいかと。)
     if(Bhv_BasicTackle(0.5,90.0).execute( agent ) )
       {
	 break;
       }
     rcsc::Body_Intercept2008(false).execute( agent );
     agent->setNeckAction(new rcsc::Neck_TurnToBall());
     break;
    */
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
    case Strategy::BA_Cross:
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
      /*
	if( wm.existKickableTeammate()
	    && wm.self().pos().x < wm.ball().pos().x )
	  {
	    if(rcsc::Body_GoToPoint(rcsc::Vector2D(wm.offsideLineX()-0.5, home_pos.y)
				    ,0.3,rcsc::ServerParam::i().maxDashPower()
				    ).execute(agent))
		{
		  agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());//Goalie?Ball?
		  break;
		}
	}
	else	
*/
      if(!wm.existKickableTeammate())
	{
	  if( self_cycle < mate_cycle 
	      && self_cycle < opp_cycle + 10 )
	    {//-----取れそうなら無条件でインターセプト-----
	      if(Bhv_BasicTackle( 0.5, 90.0 ).execute( agent ) ){break;}
	      rcsc::Body_Intercept2008(false).execute( agent );
	      agent->setNeckAction(new rcsc::Neck_TurnToBall());
	      break;
	    }
	  //オフサイドの裏をとれそうなら、スタミナ無視ダッシュ
	  if(!wm.existKickableOpponent()
	     && ball_point.x>wm.offsideLineX()
	     && wm.getTeammateNearestToBall(5)!=NULL
	     && wm.getTeammateNearestToBall(5)->isSelf())
	    {
	      if(rcsc::Body_GoToPoint(ball_point
				      ,0.5, rcsc::ServerParam::i().maxDashPower()
				      ,5,false,false,12).execute(agent))
		{
		  agent->setNeckAction(new rcsc::Neck_TurnToBall());
		  break;
		}
	    }
	}
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
      

    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
	{
      //自分がパスするとしたら・・・
	rcsc::Vector2D home_pos_tmp = home_pos;
      rcsc::Body_Pass::get_best_pass(wm,&home_pos,NULL,NULL);
	if(wm.self().pos().dist(home_pos) < 7)
		home_pos = home_pos_tmp;
      Bhv_BasicMove(home_pos).execute(agent);
      break;
	}
    default:
        Bhv_BasicMove( home_pos ).execute( agent );
        break;
    }
}


bool
RoleAttacker::FSClearBall( rcsc::PlayerAgent *agent ){

  const rcsc::WorldModel & wm = agent->world();

  int fewest_area_num_first=-1;
  int min_opps_count=7;

  int opps_num=0;
  double opp_dist=20;
  const rcsc::PlayerPtrCont::const_iterator
    end=wm.opponentsFromSelf().end();
  for(rcsc::PlayerPtrCont::const_iterator
	it=wm.opponentsFromSelf().begin();
      it!=end;++it){
    opp_dist = (*it)->pos().dist(wm.self().pos());
    opps_num++;
    if(opps_num>7)
      break;
  }
  //first search
  for(int i=0;i<3;i++){
    rcsc::Sector2D checkArea(wm.self().pos()
			     ,0.0,opp_dist
			     ,i*60-90,(i+1)*60-90);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_first=i;
    } 
  }

  if(fewest_area_num_first==-1)
    return FSClearBall_v1(agent);

  int fewest_area_num_second=-1;
  min_opps_count=5;
  //second search
  for(int i=0;i<3;i++){
    rcsc::Sector2D checkArea(wm.self().pos()
			     ,0.0,opp_dist
			     ,fewest_area_num_first*60 + i*20 -90
			     ,fewest_area_num_first*60 + (i+1)*20 -90);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_second=i;
    } 
  }

  if(fewest_area_num_second==-1){
    rcsc::Vector2D target_pos
      = rcsc::Vector2D().setPolar(opp_dist
				  ,fewest_area_num_first*60 
				  +30-90);
    target_pos += wm.self().pos();
    double first_speed = rcsc::ServerParam::i().maxPower();

    if(rcsc::Body_SmartKick(target_pos
			    ,first_speed
			    ,first_speed * 0.96
			    ,2
			    ).execute(agent)){
      agent->setNeckAction(new rcsc::Neck_TurnToBall());
      //std::cout << "move" << std::endl;
      return true;
    }
  }
  
  int fewest_area_num_third=-1;
  min_opps_count=3;
  //third search
  for(int i=0;i<4;i++){
    rcsc::Sector2D checkArea(wm.self().pos()
			     ,0.0,opp_dist
			     ,fewest_area_num_first*60 
			     + fewest_area_num_second*20
			     + i*5 -90
			     ,fewest_area_num_first*60 
			     + fewest_area_num_second*20
			     + (i+1)*5 -90);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_third=i;
    } 
  }

  if(fewest_area_num_third==-1){
    rcsc::Vector2D target_pos
      = rcsc::Vector2D().setPolar(opp_dist
				  ,fewest_area_num_first*60 
				  +fewest_area_num_second*20
				  +10-90);
    target_pos += wm.self().pos();
    double first_speed = rcsc::ServerParam::i().maxPower();

    if(rcsc::Body_SmartKick(target_pos
			    ,first_speed
			    ,first_speed * 0.96
			    ,2
			    ).execute(agent)){
      agent->setNeckAction(new rcsc::Neck_TurnToBall());
      //std::cout << "move" << std::endl;
      return true;
    }
  }

  rcsc::Vector2D target_pos
    = rcsc::Vector2D().setPolar(opp_dist
				,fewest_area_num_first*60 
				+ fewest_area_num_second*20
				+ fewest_area_num_third*5 + 2.5 -90);
   target_pos += wm.self().pos();
  double first_speed = rcsc::ServerParam::i().maxPower();

  if(rcsc::Body_SmartKick(target_pos
			  ,first_speed
			  ,first_speed * 0.96
			  ,2
			  ).execute(agent)){
    agent->setNeckAction(new rcsc::Neck_TurnToBall());
    //std::cout << "move" << std::endl;
    return true;
  }
  
  return FSClearBall_v1(agent);
}


bool
RoleAttacker::FSClearBall_v1(rcsc::PlayerAgent *agent){
  const rcsc::WorldModel & wm = agent->world();

  rcsc::Vector2D firstOpp(100,100);
  rcsc::Vector2D secondOpp(100,100);
  const rcsc::PlayerPtrCont::const_iterator
    end=wm.opponentsFromSelf().end();
  for(rcsc::PlayerPtrCont::const_iterator
	it=wm.opponentsFromSelf().begin();
      it!=end;++it){
    if((*it)->pos().x>wm.self().pos().x){
      if(firstOpp.x==100&&firstOpp.y==100){
	firstOpp=(*it)->pos();
      }else if(secondOpp.x==100&&secondOpp.y==100){
	secondOpp=(*it)->pos();
	break;
      }
    }
  }
  if(firstOpp.x!=100&&firstOpp.y!=100){
    rcsc::Vector2D target_point=firstOpp+secondOpp;
    target_point/=2;
    if(secondOpp.x!=100&&secondOpp.y!=100
       && firstOpp.dist(secondOpp)>7 
       && firstOpp.dist(wm.self().pos())>1.5
       && wm.self().pos().dist(target_point)!=0
       && rcsc::Line2D(firstOpp,secondOpp).dist(wm.self().pos())
       /wm.self().pos().dist(target_point)>0.7){
      double first_speed = rcsc::ServerParam::i().ballSpeedMax();
      double thr = 0.96;
      int step = 2;
      if(rcsc::Body_SmartKick(target_point,
			      first_speed,
			      first_speed * thr,
			      step).execute(agent)){
	agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());
	return true;
      }
    }else{
      rcsc::Line2D courseR(wm.self().pos(),firstOpp.add(-1.5,1.5));
      rcsc::Line2D courseL(wm.self().pos(),firstOpp.add(-1.5,-1.5));
      double VCourseR=0,VCourseL=0;
      for(rcsc::PlayerPtrCont::const_iterator
	    it=wm.opponentsFromSelf().begin();
	  it!=end && (*it)->distFromSelf()<40;++it){
	if((*it)->pos().x>wm.self().pos().x
	   && !(*it)->goalie()){
	  VCourseR+=courseR.dist2((*it)->pos());
	  VCourseL+=courseL.dist2((*it)->pos());
	}
      }
      double first_speed = rcsc::ServerParam::i().ballSpeedMax();
      double thr = 0.96;
      int step = 2;
      if(VCourseL<VCourseR){
	if(rcsc::Body_SmartKick(firstOpp.add(-1.5,1.5),
				first_speed,
				first_speed * thr,
				step).execute(agent)){
	  agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());
	  return true;
	}
      }else{
	if(rcsc::Body_SmartKick(firstOpp.add(-1.5,-1.5),
				first_speed,
				first_speed * thr,
				step).execute(agent)){
	  agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());
	  return true;
	}
      }
    }
  }
  return false;
}

int
RoleAttacker::countKickableCycleAtSamePoint(rcsc::PlayerAgent * agent){
  static int startCycle = 0;
  static int prevCycle = 0;
  static rcsc::Vector2D point(0,0);

  if(agent->world().self().isKickable()
     && point.dist(agent->world().ball().pos()) < 3.5
     && agent->world().time().cycle() - prevCycle <= 7){
    prevCycle = agent->world().time().cycle();
  }else{
    startCycle = agent->world().time().cycle();
    prevCycle = agent->world().time().cycle();
    point = agent->world().ball().pos();
  }
  //printf("start:%2d,prev:%2d(%.2lf,%.2lf)\n"
  // ,startCycle
  // ,prevCycle,point.x,point.y);
  return agent->world().time().cycle()-startCycle;
}


bool
RoleAttacker::FSShoot( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel & wm = agent->world();
  const rcsc::PlayerObject *goalie = wm.getOpponentGoalie();

  rcsc::Vector2D opp_goalie_pos(50,0);
  if(goalie!=NULL)
    opp_goalie_pos = goalie->pos();

  double first_speed = rcsc::ServerParam::i().ballSpeedMax();
  double thr = 0.96;
  int step = 2;
  rcsc::Vector2D 
    shoot_point(52.5, rcsc::ServerParam::i().goalHalfWidth() - 0.5);

 if(wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(52.5,0)
					,wm.self().pos().dist(rcsc::Vector2D(52.5,0))+1)
			 ,5,true)<3){
    if( wm.self().pos().dist(opp_goalie_pos) > 8 ){
      return false;
    }else{ 
      shoot_point = rcsc::Vector2D(52.5, rcsc::ServerParam::i().goalHalfWidth() - 1);
      if(opp_goalie_pos.y>0){
	shoot_point.y *= -1;
      }
      rcsc::Body_SmartKick( shoot_point
			    ,first_speed
			    ,first_speed * thr
			    ,step).execute(agent);
      //printf("Shoot! (%f,%f)\n",shoot_point.x,shoot_point.y);
      agent->setNeckAction(new rcsc::Neck_TurnToGoalieOrScan());
      return true;
    }
  }

  double offset = 1;
  int end_num = (rcsc::ServerParam::i().goalHalfWidth() - 0.5)*2;

  if(rcsc::Sector2D(rcsc::Vector2D(52.5,0.0)
		    ,0.0,25.0
		    ,135,-135).contains(wm.self().pos()))
    {
      if(wm.self().pos().y<0)
	shoot_point.y *= -1;

      rcsc::Body_SmartKick( shoot_point
			    ,first_speed
			    ,first_speed * thr
			    ,step).execute(agent);
      //printf("Shoot! (%f,%f)\n",shoot_point.x,shoot_point.y);
      agent->setNeckAction(new rcsc::Neck_TurnToGoalieOrScan());
      return true;      
    }
  else if( opp_goalie_pos.y > 0 )
    {
      shoot_point.y *= -1;
      offset=-1;
    }
  
  //シュートに関してちょっと実験（扇形で敵の有無を調べてからシュートしたい（敵がいても平然と蹴っちゃうので。））	
  bool shoot_flag = false;
   offset /= 2;
  for ( int i = 0;
	i < end_num*2;
	i++ )	//tune this!!
    {
      rcsc::AngleDeg shoot_angle = (shoot_point - wm.self().pos()).th();
      rcsc::Sector2D checkSector( wm.self().pos()
				  ,0.0//min_r
				  ,wm.self().pos().dist( shoot_point )//max_r
				  ,shoot_angle - 3//startAngleDeg	tuneThis!　ここすごく重要
				  ,shoot_angle + 3//endAngleDeg		tuneThis!　ここすごく重要。大切なことなので2回ry
				  );
      //printf("(%f,%f) countOpp=%d \n",shoot_point.x,shoot_point.y,wm.countOpponentsIn(checkSector,5,true));		
      
      if ( wm.countOpponentsIn(checkSector,3,true) == 0 )
	{
	  shoot_flag = true;
	  break;
	}

      /*
      //手前側のゴール端も調べてみる
      if ( i == 0 )
	{
	  shoot_point.y *= -1;
	  continue;
	}
      */
	    
      //------- 改めてシュートポイントを模索する ---------

      /*
      //方法１　逆サイド端しか見ない
      double rand_tmp = rand()%50;	//50種類のシュートコース、まさにFifty-Shoot.
      shoot_point.y = rcsc::ServerParam::i().goalHalfWidth() - 0.1 - rand_tmp/11;	//6.91~2.37くらい
      */
      /*
      if( goalie->pos().y > 0 )
	{
	  shoot_point.y *= -1;
	}
	*/
      //方法２　全部から探す
      //shoot_point.y = rcsc::ServerParam::i().goalHalfWidth() - rand()%15;
      
      //方法３　デフォルト（両端）がダメなら全体から探す
      //shoot_point.y = 

      //線形探索
      shoot_point.y-=offset;
    }
  
  //セクター内に敵がいるか判定
  if( shoot_flag )
    {//shoot
      rcsc::Body_SmartKick( shoot_point
			    ,first_speed
			    ,first_speed * thr
			    ,step).execute(agent);
      //printf("Shoot! (%f,%f)\n",shoot_point.x,shoot_point.y);
      agent->setNeckAction(new rcsc::Neck_TurnToGoalieOrScan());
      return true;
    }
  if(wm.self().pos().dist(rcsc::Vector2D(52.5,0))<12
     || wm.self().pos().dist(rcsc::Vector2D(52.5
					    ,rcsc::ServerParam::i().goalHalfWidth() - 0.5))<12
     || wm.self().pos().dist(rcsc::Vector2D(52.5
					    ,-rcsc::ServerParam::i().goalHalfWidth() - 0.5))<12){
    
    shoot_point 
      = rcsc::Vector2D(52.5
		       , rcsc::ServerParam::i().goalHalfWidth() - 0.5);
    if( wm.self().pos().absY() > 5
	&& wm.self().pos().x>45){ 
      if(wm.self().pos().y < 0)
	shoot_point.y *= -1;
    }else{
      if(opp_goalie_pos.y>0)
	shoot_point.y *= -1;
    }

    rcsc::Body_SmartKick( shoot_point
			  ,first_speed
			  ,first_speed * thr
			  ,step).execute(agent);
    //printf("Shoot! (%f,%f)\n",shoot_point.x,shoot_point.y);
    agent->setNeckAction(new rcsc::Neck_TurnToGoalieOrScan());
    return true;
  }
  
  return false; 
}
