// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_center_forward.h"

#include "strategy.h"

#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_move.h"
#include "bhv_basic_tackle.h"

#include <rcsc/action/body_intercept.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_goalie_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_smart_kick.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_ball.h>

#include <rcsc/geom/sector_2d.h>

#include "bhv_pass_receiver_move.h"
//messageを送るのに必要
//#include <rcsc/player/free_message.h>

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterForward::execute( rcsc::PlayerAgent * agent )
{
  //init処理
  //追加：メッセジを放ってから２０サイクル以内のメッセジのみ学習に用いる
  rcsc::Vector2D home_pos
    = formation().getPosition( agent->config().playerNumber(),
			       agent->world().ball().pos() );
  //Bhv_PassReceiverMove(home_pos).init( agent );
  bool kickable = agent->world().self().isKickable();
  if ( agent->world().existKickableTeammate()
       && agent->world().teammatesFromBall().front()->distFromBall()
       < agent->world().ball().distFromSelf() )
    {
      kickable = false;
    }
  
  if ( kickable )
    {
      doKick( agent );
    }
  else
    {
      doMove( agent );
    }
  
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterForward::doKick( rcsc::PlayerAgent * agent )
{ 
    const rcsc::WorldModel & wm = agent->world();
      rcsc::Vector2D pass_point;

      //-----OpponentgoalieData-----
      const rcsc::PlayerObject *goalie = wm.getOpponentGoalie();
      rcsc::Vector2D ShootPoint = rcsc::Vector2D(52.5,6.0-rand()%12);
      //----------------------------

      const rcsc::PlayerObject *opp =wm.getOpponentNearestToSelf(5);
      double opp_dist=wm.getDistOpponentNearestToSelf(5);
      //double rposY=0;

      bool force =false;
      if(opp_dist<5
	&& wm.self().pos().x- - 1 < opp->pos().x)//条件追加（鈴木）
	force=true;//パラメータ変更(真崎)

     switch ( Strategy::get_ball_area( agent->world().ball().pos()))
       {
	//----------------シュートチャンス----------------------------
       case Strategy::BA_ShootChance:
       {
	 if(FSShoot(agent))
	   break;

	 if ( rcsc::Body_Pass::get_best_pass(wm,&pass_point,NULL,NULL)
	      && pass_point.dist(rcsc::Vector2D(52.5,0)) < wm.self().pos().dist(rcsc::Vector2D(52.5,0))
	      && ( rcsc::ServerParam::i().useOffside() 
		   && pass_point.x<wm.offsideLineX()
		   || !rcsc::ServerParam::i().useOffside() ) )
	   {//pass
	     rcsc::Body_Pass().execute(agent);		
	   }
	 else
	   {//dribble
	     rcsc::Vector2D dribble_target(52.5
					   ,std::min(wm.self().pos().absY()
						     ,rcsc::ServerParam::i().goalHalfWidth() - 0.5 - rand()%3));
	     if(wm.self().pos().y < 0)
	       dribble_target.y *= -1;
	     rcsc::Body_Dribble2008(dribble_target,
				    5.0,rcsc::ServerParam::i().maxDashPower(),
				    2,false).execute( agent );
	     /*			rcsc::Body_Dribble2008(drib_target,3.0,
				rcsc::ServerParam::i().maxDashPower(),2,true
				).execute( agent );
	     */
	   } 
	 agent->setNeckAction(new rcsc::Neck_TurnToGoalieOrScan());
	 break;
       }
	  //-------------オフェンス中央---------------------------------
       case Strategy::BA_OffMidField:
       case Strategy::BA_DefMidField:
       case Strategy::BA_DribbleAttack:
	 if(wm.self().pos().x > 30
	    && wm.self().pos().absY() < 20)
	   if(wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(50,0)
						 ,20)
				  ,5,false) > 1
	      &&FSShoot(agent))
	     break;

	  if(opp!=NULL)
	    {
	      //rposY-=opp->rpos().y;
	    }
	  else//opp==NULL
	    {
	      opp_dist=100;
	    }
	  
	  if(rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL)
	     && opp_dist<=5.0
	     && wm.self().pos().dist(pass_point)>5.0
	     && (wm.self().pos().x-5)<pass_point.x
	     && (rcsc::ServerParam::i().useOffside() 
		 && pass_point.x<wm.offsideLineX()
		 || !rcsc::ServerParam::i().useOffside())
	     && wm.countOpponentsIn(rcsc::Circle2D(pass_point,2.5)
				    ,7,true) <=1
	     && wm.countOpponentsIn(rcsc::Circle2D(wm.self().pos(),5.0)
				    ,7,true) >=2)//敵がいっぱい
	    {
	      rcsc::Body_Pass().execute(agent);
	      //printf("Role center forward : opp.pos(%.1lf, %.1lf)\n",opp->pos().x-wm.self().pos().x,opp->pos().y);
	    }
	  else
	    {
	      if(countKickableCycleAtSamePoint(agent)<12){
		//printf("%2d:kickable\n",countKickableCycleAtSamePoint(agent));
		  rcsc::Vector2D dribble_point(47.0
					       ,std::min(wm.self().pos().absY()
							 * (rand()%4)/2
							 ,30.0));
		  if(wm.self().pos().y<0)
		      dribble_point.y *= -1;

		  if(opp_dist<=5.0){
		     if(wm.countOpponentsIn(rcsc::Circle2D(wm.self().pos(),5.0)
					    ,7,true) >=2){
		       		if(FSClearBall(agent))
				  break;
		     }else{
		       force=true;
		     }
		  }
					       
		  rcsc::Body_Dribble2008(dribble_point
					 ,5.0,rcsc::ServerParam::i().maxPower()
					 ,3,force).execute( agent );
	      }else{
		//ドリブルできないときは、てきとークリア
		if(FSClearBall(agent))
		  break;
	      }
	    }
	  agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
	  break;
	  //----------クロス---------
       case Strategy::BA_Cross:
	 if(wm.self().pos().x > 30
	    && wm.self().pos().absY() < 20)
	   if(wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(50,0)
						 ,20)
				  ,5,false) > 1
	      &&FSShoot(agent))
	     break;
	 if(countKickableCycleAtSamePoint(agent)<15){
	  rcsc::Body_Dribble2008(rcsc::Vector2D(52.0,-4.5+rand()%10),
				 5.0,rcsc::ServerParam::i().maxDashPower(),
				 3,true).execute( agent );
	  agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
	  break;
	 }else{
	   if(rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL)){
	     rcsc::Body_Pass().execute(agent);
	     agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
	     break;
	   }
	 }
	 
       case Strategy::BA_CrossBlock:
	case Strategy::BA_Stopper:
	case Strategy::BA_Danger:
	case Strategy::BA_DribbleBlock:
	default:
	  Bhv_BasicOffensiveKick().execute( agent );
	  break;
	}
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterForward::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    int ball_reach_step = 0;
    if ( ! wm.existKickableTeammate() && ! wm.existKickableOpponent())
    {
        ball_reach_step
            = std::min( wm.interceptTable()->teammateReachCycle(),
                        wm.interceptTable()->opponentReachCycle() );
    }
    const rcsc::Vector2D base_pos
        = wm.ball().inertiaPoint( ball_reach_step );

    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, wm.offsideLineX() - 1.0 );
    }

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: HOME POSITION=(%.2f, %.2f) base_point(%.1f %.1f)"
                        ,__FILE__,
                        home_pos.x, home_pos.y,
                        base_pos.x, base_pos.y );

    //----- Opponents Info -----
    //const rcsc::PlayerObject *goalie = wm.getOpponentGoalie();
    const rcsc::PlayerObject *opp = wm.getOpponentNearestToSelf(10);
    //----- Teammates Info -----
    const rcsc::PlayerObject *team = wm.getTeammateNearestToBall(10);
    //----------------------------------------------
    //佐野さんの学習
    //if(Bhv_PassReceiverMove( home_pos ).execute( agent ))
      //return;
    switch ( Strategy::get_ball_area( base_pos ) ) {
    case Strategy::BA_ShootChance:
      if(wm.self().pos().x>45
	 &&wm.self().pos().absY()<10
	 &&wm.self().pos().x>wm.ball().pos().x){
	//---必殺!ウォールブレイカー---
	if ( Bhv_BasicTackle( 0.5, 90.0 ).execute( agent ) ){
	  break;
	}
      }
      if(wm.existKickableOpponent())
	if ( Bhv_BasicTackle( 0.5, 90.0 ).execute( agent ) ){
	  break;
	}
      if(!wm.existKickableTeammate()
	 &&wm.ball().pos().dist(wm.self().pos())<3)
	{
	  rcsc::Body_Intercept2008(false).execute( agent );
	  agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
	  break;
	}
      /*
      else if(wm.existKickableTeammate()
	      && wm.self().pos().x>wm.offsideLineX())
	{
	  if(rcsc::Body_GoToPoint(rcsc::Vector2D(wm.offsideLineX()-0.5
						 ,home_pos.y)
				  ,0.3,rcsc::ServerParam::i().maxDashPower()*0.75
				  ).execute(agent)){
	    agent->setNeckAction(new rcsc::Neck_TurnToGoalieOrScan());
	    break;
	  }
	}
      */
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
    case Strategy::BA_DribbleAttack:
      if(!wm.existKickableTeammate()
	 &&wm.ball().pos().dist(wm.self().pos())<4)
	{
	  rcsc::Body_Intercept2008(false).execute( agent );
	  agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
	  break;
	}
      //オフサイドの裏をとれそうなら、スタミナ無視ダッシュ
      if(!wm.existKickableOpponent()
	 && wm.ball().pos().x>wm.offsideLineX()
	 && wm.getTeammateNearestToBall(10)!=NULL
	 && wm.getTeammateNearestToBall(10)->isSelf()){
	if(rcsc::Body_GoToPoint(wm.ball().pos()
				,0.5, rcsc::ServerParam::i().maxDashPower()
				,5,false,false,12).execute(agent)){
	  agent->setNeckAction(new rcsc::Neck_TurnToBall());
	  break;
	}
      }
      if(opp!=NULL && wm.existKickableOpponent())
	{
	  home_pos = opp->pos();
	}
      if(wm.self().pos().dist(wm.ball().pos())<4.0)
	{
	  home_pos = wm.ball().pos();
	}
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
    case Strategy::BA_OffMidField:
      //---ボールが近い---
      if(!wm.existKickableTeammate())
	{
	  if(opp != NULL && wm.existKickableOpponent())
	  {
	      home_pos = opp->pos();
	      home_pos.x -= 1.0;
	  }

	  if(wm.self().pos().dist(wm.ball().pos())<9.0)
	    {
	      home_pos = wm.ball().pos();
	    }
	  
	  if(wm.self().pos().dist(wm.ball().pos())<4.5)
	    {//---ボールが取れそう---
	      if ( Bhv_BasicTackle(0.5,90.0).execute(agent))
		{
		  break;
		}
	      rcsc::Body_Intercept2008(false).execute( agent );
	      agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
	      break;
	    }
	}
      else
       {
	 if(wm.ball().pos().dist(wm.self().pos())<1
	    &&wm.ball().vel().x > 0)
	   {
	     rcsc::Body_Intercept2008(false).execute( agent );
	     agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
	     break;
	   }
	 //オフサイドの裏をとれそうなら、スタミナ無視ダッシュ
	 if(!wm.existKickableOpponent()
	    && wm.self().pos().x<wm.offsideLineX()){
	   if(rcsc::Body_GoToPoint(wm.ball().pos()
				   ,0.5, rcsc::ServerParam::i().maxDashPower()
				   ,5,false,false,12).execute(agent)){
	     agent->setNeckAction(new rcsc::Neck_TurnToBall());
	     break;
	   }
	 }
	 home_pos.x = wm.offsideLineX()-0.1;
       }
      //処理の実行
      Bhv_BasicMove( home_pos ).execute( agent );
      break;

    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
      if(!wm.existKickableTeammate()
	 &&wm.ball().pos().dist(wm.self().pos())<3
	 &&!wm.existTeammateIn(rcsc::Circle2D(wm.ball().pos()
					      ,1.5)
			       ,5,false))
	{
	  rcsc::Body_Intercept2008(false).execute( agent );
	  agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
	  break;
	}
      if(moveReceivableArea(agent))
	break;

    case Strategy::BA_Cross:
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:

    default:
      //if(!Bhv_PassReceiverMove( home_pos ).execute( agent ))
	Bhv_BasicMove( home_pos ).execute( agent );
      break;
    }
}

bool
RoleCenterForward::FSClearBall( rcsc::PlayerAgent * agent ){

  const rcsc::WorldModel & wm = agent->world();

  int fewest_area_num_first=-1;
  int min_opps_count=7;

  int opps_num=0;
  double opp_dist=20;
  const rcsc::PlayerPtrCont::const_iterator
    end=wm.opponentsFromSelf().end();
  for(rcsc::PlayerPtrCont::const_iterator
	it=wm.opponentsFromSelf().begin();
      it!=end;++it){
    opp_dist = (*it)->pos().dist(wm.self().pos());
    opps_num++;
    if(opps_num>7)
      break;
  }
  //first search
  for(int i=0;i<3;i++){
    rcsc::Sector2D checkArea(wm.self().pos()
			     ,0.0,opp_dist
			     ,i*60-90,(i+1)*60-90);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_first=i;
    } 
  }

  if(fewest_area_num_first==-1)
    return FSClearBall_v1(agent);

  int fewest_area_num_second=-1;
  min_opps_count=5;
  //second search
  for(int i=0;i<3;i++){
    rcsc::Sector2D checkArea(wm.self().pos()
			     ,0.0,opp_dist
			     ,fewest_area_num_first*60 + i*20 -90
			     ,fewest_area_num_first*60 + (i+1)*20 -90);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_second=i;
    } 
  }

  if(fewest_area_num_second==-1){
    rcsc::Vector2D target_pos
      = rcsc::Vector2D().setPolar(opp_dist
				  ,fewest_area_num_first*60 
				  +30-90);
    target_pos += wm.self().pos();
  
    double first_speed = rcsc::ServerParam::i().maxPower();

    if(rcsc::Body_SmartKick(target_pos
			    ,first_speed
			    ,first_speed * 0.96
			    ,2
			    ).execute(agent)){
      agent->setNeckAction(new rcsc::Neck_TurnToBall());
      return true;
    }
  }
  
  int fewest_area_num_third=-1;
  min_opps_count=3;
  //third search
  for(int i=0;i<4;i++){
    rcsc::Sector2D checkArea(wm.self().pos()
			     ,0.0,opp_dist
			     ,fewest_area_num_first*60 
			     + fewest_area_num_second*20
			     + i*5 -90
			     ,fewest_area_num_first*60 
			     + fewest_area_num_second*20
			     + (i+1)*5 -90);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_third=i;
    } 
  }

  if(fewest_area_num_third==-1){
    rcsc::Vector2D target_pos
      = rcsc::Vector2D().setPolar(opp_dist
				  ,fewest_area_num_first*60 
				  +fewest_area_num_second*20
				  +10-90);
    target_pos += wm.self().pos();
    double first_speed = rcsc::ServerParam::i().maxPower();

    if(rcsc::Body_SmartKick(target_pos
			    ,first_speed
			    ,first_speed * 0.96
			    ,2
			    ).execute(agent)){
      agent->setNeckAction(new rcsc::Neck_TurnToBall());
      return true;
    }
  }

  rcsc::Vector2D target_pos
    = rcsc::Vector2D().setPolar(opp_dist
				,fewest_area_num_first*60 
				+ fewest_area_num_second*20
				+ fewest_area_num_third*5 + 2.5 -90);
  target_pos += wm.self().pos();
  double first_speed = rcsc::ServerParam::i().maxPower();

  if(rcsc::Body_SmartKick(target_pos
			  ,first_speed
			  ,first_speed * 0.96
			  ,2
			  ).execute(agent)){
    agent->setNeckAction(new rcsc::Neck_TurnToBall());
    return true;
  }
  
  return FSClearBall_v1(agent);
}

bool
RoleCenterForward::FSClearBall_v1( rcsc::PlayerAgent *agent ){
  const rcsc::WorldModel & wm = agent->world();

  rcsc::Vector2D firstOpp(100,100);
  rcsc::Vector2D secondOpp(100,100);
  const rcsc::PlayerPtrCont::const_iterator
    end=wm.opponentsFromSelf().end();
  for(rcsc::PlayerPtrCont::const_iterator
	it=wm.opponentsFromSelf().begin();
      it!=end;++it){
    if((*it)->pos().x>wm.self().pos().x){
      if(firstOpp.x==100&&firstOpp.y==100){
	firstOpp=(*it)->pos();
      }else if(secondOpp.x==100&&secondOpp.y==100){
	secondOpp=(*it)->pos();
	break;
      }
    }
  }
  if(firstOpp.x!=100&&firstOpp.y!=100){
    rcsc::Vector2D target_point=firstOpp+secondOpp;
    target_point/=2;
    if(secondOpp.x!=100&&secondOpp.y!=100
       && firstOpp.dist(secondOpp)>7 
       && firstOpp.dist(wm.self().pos())>1.5
       && wm.self().pos().dist(target_point)!=0
       && rcsc::Line2D(firstOpp,secondOpp).dist(wm.self().pos())
       /wm.self().pos().dist(target_point)>0.7){
      double first_speed = rcsc::ServerParam::i().ballSpeedMax();
      double thr = 0.96;
      int step = 2;
      if(rcsc::Body_SmartKick(target_point,
			      first_speed,
			      first_speed * thr,
			      step).execute(agent)){
	agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());
	return true;
      }
    }else{
      rcsc::Line2D courseR(wm.self().pos(),firstOpp.add(-1.5,1.5));
      rcsc::Line2D courseL(wm.self().pos(),firstOpp.add(-1.5,-1.5));
      double VCourseR=0,VCourseL=0;
      for(rcsc::PlayerPtrCont::const_iterator
	    it=wm.opponentsFromSelf().begin();
	  it!=end && (*it)->distFromSelf()<40;++it){
	if((*it)->pos().x>wm.self().pos().x
	   && !(*it)->goalie()){
	  VCourseR+=courseR.dist2((*it)->pos());
	  VCourseL+=courseL.dist2((*it)->pos());
	}
      }
      double first_speed = rcsc::ServerParam::i().ballSpeedMax();
      double thr = 0.96;
      int step = 2;
      if(VCourseL<VCourseR){
	if(rcsc::Body_SmartKick(firstOpp.add(-1.5,1.5),
				first_speed,
				first_speed * thr,
				step).execute(agent)){
	  agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());
	  return true;
	}
      }else{
	if(rcsc::Body_SmartKick(firstOpp.add(-1.5,-1.5),
				first_speed,
				first_speed * thr,
				step).execute(agent)){
	  agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());
	  return true;
	}
      }
    }
  }
  return false;
}

int
RoleCenterForward::countKickableCycleAtSamePoint( rcsc::PlayerAgent * agent ){
  static int startCycle = 0;
  static int prevCycle = 0;
  static rcsc::Vector2D point(0,0);

  if(agent->world().self().isKickable()
     && point.dist(agent->world().ball().pos()) < 3.5
     && agent->world().time().cycle() - prevCycle <= 7){
    prevCycle = agent->world().time().cycle();
  }else{
    startCycle = agent->world().time().cycle();
    prevCycle = agent->world().time().cycle();
    point = agent->world().ball().pos();
  }
  //printf("start:%2d,prev:%2d(%.2lf,%.2lf)\n"
  // ,startCycle
  // ,prevCycle,point.x,point.y);
  return agent->world().time().cycle()-startCycle;
}

bool
RoleCenterForward::moveReceivableArea( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel & wm = agent->world();
  if(!wm.existKickableTeammate())
    return false;

  const rcsc::Vector2D ball_pos = wm.ball().pos();

  int fewest_area_num_first=-1;
  int min_opps_count=7;

  //first search
  for(int i=0;i<4;i++){
    rcsc::Sector2D checkArea(ball_pos
			     ,0.0,ball_pos.dist(wm.self().pos())
			     ,i*90,(i+1)*90);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_first=i;
    } 
  }

  if(fewest_area_num_first==-1)
    return false;

  int fewest_area_num_second=-1;
  min_opps_count=5;
  //second search
  for(int i=0;i<3;i++){
    rcsc::Sector2D checkArea(ball_pos
			     ,0.0,ball_pos.dist(wm.self().pos())
			     ,fewest_area_num_first*90 + i*30
			     ,fewest_area_num_first*90 + (i+1)*30);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_second=i;
    } 
  }

  if(fewest_area_num_second==-1)
    return false;
  
  int fewest_area_num_third=-1;
  min_opps_count=3;
  //third search
  for(int i=0;i<3;i++){
    rcsc::Sector2D checkArea(ball_pos
			     ,0.0,ball_pos.dist(wm.self().pos())
			     ,fewest_area_num_first*90 
			     + fewest_area_num_second*30
			     + i*10
			     ,fewest_area_num_first*90 
			     + fewest_area_num_second*30
			     + (i+1)*10);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_third=i;
    } 
  }

  if(fewest_area_num_third==-1)
    return false;

  rcsc::Vector2D target_pos
    = rcsc::Vector2D().setPolar(ball_pos.dist(wm.self().pos())/2
				,fewest_area_num_first*90 
				+ fewest_area_num_second*30
				+ fewest_area_num_third*10 + 5);
  target_pos += ball_pos;
  if(target_pos.dist(wm.self().pos())>20
     || target_pos.dist(wm.self().pos())>ball_pos.dist(wm.self().pos())*1.5
     || target_pos.x < ball_pos.x-10)
    return false;
			    
  if(rcsc::Body_GoToPoint(target_pos
			  ,0.5, rcsc::ServerParam::i().maxDashPower()
			  ,5,false,false,12).execute(agent)){
    agent->setNeckAction(new rcsc::Neck_TurnToBall());
    return true;
  }
  
  return false;
}

bool
RoleCenterForward::FSShoot( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel & wm = agent->world();
  const rcsc::PlayerObject *goalie = wm.getOpponentGoalie();

  if(wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(52.5,0)
					,wm.self().pos().dist(rcsc::Vector2D(52.5,0))+1)
			 ,5,true)<3){
    
  }

  rcsc::Vector2D opp_goalie_pos(50,0);
  if(goalie!=NULL)
    opp_goalie_pos = goalie->pos();

  double first_speed = rcsc::ServerParam::i().ballSpeedMax();
  double thr = 0.96;
  int step = 2;
  rcsc::Vector2D 
    shoot_point(52.5, rcsc::ServerParam::i().goalHalfWidth() - 0.5);

  if(wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(52.5,0)
					,wm.self().pos().dist(rcsc::Vector2D(52.5,0))+1)
			 ,5,true)<3){
    if( wm.self().pos().dist(opp_goalie_pos) > 8 ){
      return false;
    }else{ 
      shoot_point = rcsc::Vector2D(52.5, rcsc::ServerParam::i().goalHalfWidth() - 1);
      if(opp_goalie_pos.y>0){
	shoot_point.y *= -1;
      }
      rcsc::Body_SmartKick( shoot_point
			    ,first_speed
			    ,first_speed * thr
			    ,step).execute(agent);
      //printf("Shoot! (%f,%f)\n",shoot_point.x,shoot_point.y);
      agent->setNeckAction(new rcsc::Neck_TurnToGoalieOrScan());
      return true;
    }
  }

  double offset = 1;
  int end_num = (rcsc::ServerParam::i().goalHalfWidth() - 0.5)*2;

  if(!rcsc::Sector2D(rcsc::Vector2D(52.5,0.0)
		    ,0.0,25.0
		    ,135,-135).contains(wm.self().pos()))
    {
      if(wm.self().pos().y<0){
	shoot_point.y *= -1;
      }

      rcsc::Body_SmartKick( shoot_point
			    ,first_speed
			    ,first_speed * thr
			    ,step).execute(agent);
      //printf("Shoot! (%f,%f)\n",shoot_point.x,shoot_point.y);
      agent->setNeckAction(new rcsc::Neck_TurnToGoalieOrScan());
      return true;
      
    }
  else if( opp_goalie_pos.y > 0 )
    {
      shoot_point.y *= -1;
      offset=-1;
    }
  
  //シュートに関してちょっと実験（扇形で敵の有無を調べてからシュートしたい（敵がいても平然と蹴っちゃうので。））	

  bool shoot_flag = false;
  offset /= 2;
  for (int i = 0;
	i < end_num*2;
	i++ )	//tune this!!
    {
      rcsc::AngleDeg shoot_angle = (shoot_point - wm.self().pos()).th();
      rcsc::Sector2D checkSector( wm.self().pos()
				  ,0.0//min_r
				  ,wm.self().pos().dist( shoot_point )//max_r
				  ,shoot_angle - 3//startAngleDeg	tuneThis!　ここすごく重要
				  ,shoot_angle + 3//endAngleDeg		tuneThis!　ここすごく重要。大切なことなので2回ry
				  );
      //printf("(%f,%f) countOpp=%d \n",shoot_point.x,shoot_point.y,wm.countOpponentsIn(checkSector,5,true));		
      
      if ( wm.countOpponentsIn(checkSector,3,true) == 0 )
	{
	  shoot_flag = true;
	  break;
	}

      /*
      //手前側のゴール端も調べてみる
      if ( i == 0 )
	{
	  shoot_point.y *= -1;
	  continue;
	}
      */
	    
      //------- 改めてシュートポイントを模索する ---------

      /*
      //方法１　逆サイド端しか見ない
      double rand_tmp = rand()%50;	//50種類のシュートコース、まさにFifty-Shoot.
      shoot_point.y = rcsc::ServerParam::i().goalHalfWidth() - 0.1 - rand_tmp/11;	//6.91~2.37くらい
      */
      /*
      if( goalie->pos().y > 0 )
	{
	  shoot_point.y *= -1;
	}
	*/
      //方法２　全部から探す
      //shoot_point.y = rcsc::ServerParam::i().goalHalfWidth() - rand()%15;
      
      //方法３　デフォルト（両端）がダメなら全体から探す
      //shoot_point.y = 

      //線形探索
      shoot_point.y-=offset;
    }
  
  //セクター内に敵がいるか判定
  if( shoot_flag )
    {//shoot
      rcsc::Body_SmartKick( shoot_point
			    ,first_speed
			    ,first_speed * thr
			    ,step).execute(agent);
      //printf("Shoot! (%f,%f)\n",shoot_point.x,shoot_point.y);
      agent->setNeckAction(new rcsc::Neck_TurnToGoalieOrScan());
      return true;
    }
  if(wm.self().pos().dist(rcsc::Vector2D(52.5,0))<12
     || wm.self().pos().dist(rcsc::Vector2D(52.5
					   ,rcsc::ServerParam::i().goalHalfWidth() - 0.5))<12
     || wm.self().pos().dist(rcsc::Vector2D(52.5
					   ,-rcsc::ServerParam::i().goalHalfWidth() - 0.5))<12)
    {
    
      shoot_point 
	= rcsc::Vector2D(52.5
			 , rcsc::ServerParam::i().goalHalfWidth() - 0.5);
      if( wm.self().pos().absY() > 5
	  && wm.self().pos().x>45){ 
	if(wm.self().pos().y < 0)
	  shoot_point.y *= -1;
      }else{
	if( opp_goalie_pos.y>0)
	  shoot_point.y *= -1;
      }
      
      rcsc::Body_SmartKick( shoot_point
			    ,first_speed
			    ,first_speed * thr
			    ,step).execute(agent);
      //printf("Shoot! (%f,%f)\n",shoot_point.x,shoot_point.y);
      agent->setNeckAction(new rcsc::Neck_TurnToGoalieOrScan());
      return true;
    }
  
  return false; 
}
