// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_defensive_half.h"

#include "strategy.h"

#include "bhv_basic_tackle.h"
#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_move.h"

#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc/action/basic_actions.h>

#include <rcsc/geom/sector_2d.h>

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleDefensiveHalf::execute( rcsc::PlayerAgent * agent )
{
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().teammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }

    if ( kickable )
    {
        doKick( agent );
    }
    else
    {
        doMove( agent );
    }

}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleDefensiveHalf::doKick( rcsc::PlayerAgent * agent )
{
  //通常の処理
  switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
  case Strategy::BA_CrossBlock:
  case Strategy::BA_Stopper:
  case Strategy::BA_Danger:
  case Strategy::BA_DribbleBlock:
  case Strategy::BA_DefMidField:
  case Strategy::BA_DribbleAttack:
  case Strategy::BA_OffMidField:
  case Strategy::BA_Cross:
  case Strategy::BA_ShootChance:
  default:
    Bhv_BasicOffensiveKick().execute( agent );
    break;
  }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleDefensiveHalf::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    //const rcsc::PlayerObject *opp 
    //= wm.getOpponentNearestTo(rcsc::Vector2D(-50,0),5,NULL);
    //.getOpponentNearestToSelf(5);

    int ball_reach_step = 0;
    if ( ! wm.existKickableTeammate()
         && ! wm.existKickableOpponent() )
    {
        ball_reach_step
            = std::min( wm.interceptTable()->teammateReachCycle(),
                        wm.interceptTable()->opponentReachCycle() );
    }
    const rcsc::Vector2D base_pos
        = wm.ball().inertiaPoint( ball_reach_step );

    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, wm.offsideLineX() - 1.0 );
    }

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: HOME POSITION=(%.2f, %.2f) base_point(%.1f %.1f)"
                        ,__FILE__,
                        home_pos.x, home_pos.y,
                        base_pos.x, base_pos.y);
    //strategyの適用
    switch ( Strategy::get_ball_area( base_pos ) ) {
    case Strategy::BA_OffMidField:
    case Strategy::BA_DefMidField:
      /*
	if((wm.self().pos().x<0)&&(wm.ball().distFromSelf()<15.0))
	{
	if (!wm.existKickableTeammate())
	{
	home_pos = rcsc::Vector2D(wm.ball().pos().x,
	wm.ball().pos().y);
	}
	}
      */

      if(wm.ball().pos().dist(wm.self().pos())<4)
	{
	  rcsc::Body_Intercept2008(false).execute( agent );
	  agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());
	  break;
	}
      if(moveReceivableArea(agent))
	break;
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
    case Strategy::BA_Stopper:
/*	if(wm.existKickableOpponent())
	{//Stopper内で敵ボールがあればタックル
	  if ( Bhv_BasicTackle( 0.5, 90.0 ).execute( agent ) )
	  {
	   break;
	  }
	  if(wm.ball().pos().y*wm.self().pos().y>=0)
	  {
	   rcsc::Body_Intercept2008(false).execute( agent );
           agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
	   break;
	  }
	}
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
*/
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Danger:
    {
      if(wm.existKickableOpponent())
	{
	  if( Bhv_BasicTackle(0.5,90.0).execute(agent) )
	    break;
	  
	  int self_min = wm.interceptTable()->selfReachCycle();
	  int mate_min = wm.interceptTable()->teammateReachCycle();
	  int opp_min = wm.interceptTable()->opponentReachCycle();
	  
	  if ( ! wm.existKickableTeammate()
	       && ( self_min <= 7
		    || ( self_min <= mate_min + 1
			 && self_min < opp_min+7
			 )
		    )
	       )
	    {
	      rcsc::dlog.addText( rcsc::Logger::TEAM,
				  "%s:%d: intercept"
				  ,__FILE__, __LINE__ );
	      rcsc::Body_Intercept2008().execute( agent );
	      if ( wm.ball().distFromSelf() < rcsc::ServerParam::i().visibleDistance() )
		agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	      else
		agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	      
	      break;
	    }
	}
      if(markOppAtOurGoal(agent))
	break;
    }
    Bhv_BasicMove( home_pos ).execute( agent );
    break;
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DribbleAttack:
      if(moveReceivableArea(agent))
	break;
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
    default:
        Bhv_BasicMove( home_pos ).execute( agent );
        break;
    }
}


bool
RoleDefensiveHalf::markOppAtOurGoal( rcsc::PlayerAgent * agent ){
  const rcsc::WorldModel & wm = agent->world();

  if(wm.ball().distFromSelf()>10.0
     && wm.ball().pos().x < -30
     && wm.ball().pos().absY()>10){
    rcsc::Vector2D markPoint = wm.ball().pos();
    bool markFlag=false;
    const rcsc::PlayerPtrCont::const_iterator
      opps_end = wm.opponentsFromSelf().end();
    for ( rcsc::PlayerPtrCont::const_iterator
	    opps_it = wm.opponentsFromSelf().begin();
	  opps_it != opps_end; ++opps_it ){
      if ( (*opps_it)->pos().dist(rcsc::Vector2D(-52.5,0))<15 
	   && !(*opps_it)->isKickable()){
	const rcsc::PlayerPtrCont::const_iterator
	  teams_end = wm.teammatesFromSelf().end();
	for(rcsc::PlayerPtrCont::const_iterator
	      teams_it = wm.teammatesFromSelf().begin();
	    teams_it != teams_end; ++teams_it){
	  if( (*opps_it)->pos().dist((*teams_it)->pos())<2 ){
	    markFlag=false;
	    break;
	  }else{
	    markFlag=true;
	  }
	}
	if(markFlag){
	  if ( (*opps_it)->pos().x < wm.ball().pos().x )
	    markPoint.x = (*opps_it)->pos().x + 0.5;
	  else
	    markPoint.x = (*opps_it)->pos().x - 0.5;
	  
	  if ( (*opps_it)->pos().y < wm.ball().pos().y )
	    markPoint.y = (*opps_it)->pos().y + 0.5;
	  else
	    markPoint.y = (*opps_it)->pos().y - 0.5;
	  /*
	  markPoint=
	    rcsc::Segment2D((*opps_it)->pos()
			    ,wm.ball().pos()).nearestPoint(rcsc::Vector2D(-52.5
			    ,0));
	  */
	  break;
	}
      }
    }

    if(!markFlag)
      return false;

    double dash_power = getDashPower(agent);
    int cycle=100;
    bool back_mode=false;
    bool save_recovery=true;
    double dir_thr=12;
    
    if(wm.countTeammatesIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
			   ,5,true)
       <=wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
			     ,5,false)){
      dash_power=std::min(dash_power*1.05,rcsc::ServerParam::i().maxPower());
      save_recovery=false;
    }
    /*
    if(wm.self().pos().x < wm.ball().pos().x
       && wm.self().pos().x < markPoint.x
       && wm.self().pos().dist(rcsc::Vector2D(-52.5,0.0))<10)
      back_mode=true;
    */

    for(int i=0;i<5;i++){
      if(wm.ball().inertiaPoint(i).dist(markPoint)<1){
	
	save_recovery=false;
	if(! rcsc::Body_GoToPoint( wm.ball().pos(),
				   0.25,dash_power
				   ,cycle,back_mode
				   ,save_recovery,dir_thr
				   ).execute( agent )
	   ){
	  rcsc::Body_TurnToPoint(markPoint).execute(agent);
	}
	agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	return true;
      }
    }

    if(! rcsc::Body_GoToPoint( markPoint,
			       0.25,dash_power
			       ,cycle,back_mode
			       ,save_recovery,dir_thr
			       ).execute( agent )
       ){
      if(wm.self().posPrev().dist(markPoint)>1)
	rcsc::Body_TurnToPoint(markPoint).execute(agent);
      else
	rcsc::Body_TurnToBall().execute(agent);
    }
    if ( wm.ball().distFromSelf() < 18.0 ){
      agent->setNeckAction( new rcsc::Neck_TurnToBall() );
    }else{
      agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }
    return true;
  }else{
    return false;
  }
}

bool
RoleDefensiveHalf::moveReceivableArea( rcsc::PlayerAgent * agent )
{

  const rcsc::WorldModel & wm = agent->world();
  if(!wm.existKickableTeammate())
    return false;

  const rcsc::Vector2D ball_pos = wm.ball().pos();

  int fewest_area_num_first=-1;
  int min_opps_count=7;

  //first search
  for(int i=0;i<4;i++){
    rcsc::Sector2D checkArea(ball_pos
			     ,0.0,ball_pos.dist(wm.self().pos())
			     ,i*90,(i+1)*90);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_first=i;
    } 
  }

  if(fewest_area_num_first==-1)
    return false;

  int fewest_area_num_second=-1;
  min_opps_count=5;
  //second search
  for(int i=0;i<3;i++){
    rcsc::Sector2D checkArea(ball_pos
			     ,0.0,ball_pos.dist(wm.self().pos())
			     ,fewest_area_num_first*90 + i*30
			     ,fewest_area_num_first*90 + (i+1)*30);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_second=i;
    } 
  }

  if(fewest_area_num_second==-1)
    return false;
  
  int fewest_area_num_third=-1;
  min_opps_count=3;
  //third search
  for(int i=0;i<3;i++){
    rcsc::Sector2D checkArea(ball_pos
			     ,0.0,ball_pos.dist(wm.self().pos())
			     ,fewest_area_num_first*90 
			     + fewest_area_num_second*30
			     + i*10
			     ,fewest_area_num_first*90 
			     + fewest_area_num_second*30
			     + (i+1)*10);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_third=i;
    } 
  }

  if(fewest_area_num_third==-1)
    return false;

  if(rcsc::Sector2D(ball_pos
		    ,0.0,ball_pos.dist(wm.self().pos())
		    , fewest_area_num_first*90 
		    + fewest_area_num_second*30
		    + fewest_area_num_third*10
		    , fewest_area_num_first*90 
		    + fewest_area_num_second*30
		    + fewest_area_num_third*10 + 5).contains(wm.self().pos()))
    {
      if(rcsc::Body_GoToPoint(ball_pos
			      ,0.5, rcsc::ServerParam::i().maxDashPower()
			      ,5,false,false,12).execute(agent)){
      }else{
	rcsc::Body_TurnToBall();
      }
      agent->setNeckAction(new rcsc::Neck_TurnToBall());
      //std::cout << "move" << std::endl;
      return true;
    }

  rcsc::Vector2D target_pos
    = rcsc::Vector2D().setPolar(ball_pos.dist(wm.self().pos())/2
				,fewest_area_num_first*90 
				+ fewest_area_num_second*30
				+ fewest_area_num_third*10 + 5);
  target_pos += ball_pos;
  if(target_pos.dist(wm.self().pos())>20
     || target_pos.dist(wm.self().pos())>ball_pos.dist(wm.self().pos())*1.5
     || target_pos.x < ball_pos.x-10)
    return false;
  
  if(rcsc::Body_GoToPoint(target_pos
			  ,0.5, rcsc::ServerParam::i().maxDashPower()
			  ,5,false,false,12).execute(agent)){
    agent->setNeckAction(new rcsc::Neck_TurnToBall());
    //std::cout << "move" << std::endl;
    return true;
  }
  
  return false;

}


double
RoleDefensiveHalf::getDashPower( const rcsc::PlayerAgent * agent)
{
    static bool s_recover_mode = false;

    const rcsc::WorldModel & wm = agent->world();
    /*--------------------------------------------------------*/
    // check recover

    if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.5 )
    {
        s_recover_mode = true;
    }
    else if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.7 )
    {
        s_recover_mode = false;
    }

    /*--------------------------------------------------------*/
    double dash_power = 100.0;
    const double my_inc
        = wm.self().playerType().staminaIncMax()
        * wm.self().recovery();

    if ( wm.defenseLineX() > wm.self().pos().x
         && wm.ball().pos().x < wm.defenseLineX() + 20.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: correct DF line. keep max power"
                            ,__FILE__, __LINE__ );
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
    else if ( s_recover_mode )
    {
        dash_power = my_inc - 25.0; // preffered recover value
        if ( dash_power < 0.0 ) dash_power = 0.0;

        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: recovering"
                            ,__FILE__, __LINE__ );
    }
    // exist kickable teammate
    else if ( wm.existKickableTeammate()
              && wm.ball().distFromSelf() < 20.0 )
    {
        dash_power = std::min( my_inc * 1.1,
                               rcsc::ServerParam::i().maxPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: exist kickable teammate. dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }
    // in offside area
    else if ( wm.self().pos().x > wm.offsideLineX() )
    {
        dash_power = rcsc::ServerParam::i().maxPower();
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: in offside area. dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }
    // normal
    else
    {
        dash_power = std::min( my_inc * 1.7,
                               rcsc::ServerParam::i().maxPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: normal mode dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }

    return dash_power;
}
