// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA

 *EndCopyright
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_offensive_half.h"

#include "strategy.h"

#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_move.h"
#include "bhv_basic_tackle.h"

#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_ball.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/body_pass.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/action/body_smart_kick.h>
#include <rcsc/geom/sector_2d.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>

//#include "bhv_mf_kick.h"

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::execute( rcsc::PlayerAgent * agent )
{
  //mf_off.init( agent );
  bool kickable = agent->world().self().isKickable();
  if ( agent->world().existKickableTeammate()
       && agent->world().teammatesFromBall().front()->distFromBall()
       < agent->world().ball().distFromSelf() )
    {
      kickable = false;
    }
  
  if ( kickable )
    {
      doKick( agent );
    }
  else
    {
      doMove( agent );
    }
  
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doKick( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel &wm=agent->world();
    switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
    case Strategy::BA_Cross:{
      rcsc::Vector2D pass_point;
      if(rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL)
	 && Strategy::get_ball_area(pass_point) 
	 == Strategy::BA_ShootChance
	 &&(rcsc::ServerParam::i().useOffside() 
	    && pass_point.x<wm.offsideLineX()
	    || !rcsc::ServerParam::i().useOffside()))
	{//PassPoint is in ShootChanceArea
	  rcsc::Body_Pass().execute(agent);
	}
      else
	{
	  rcsc::Body_Dribble2008(rcsc::Vector2D(46.0,0),5.0,
				 rcsc::ServerParam::i().maxDashPower()*0.8,
				 3,false).execute( agent );
	}
      agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
      break;
    }
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
    case Strategy::BA_ShootChance:
    default:
      if(countKickableCycleAtSamePoint(agent)>20)
	if(FSClearBall(agent))
	  break;

	//mf_off.execute( agent );
	Bhv_BasicOffensiveKick().execute(agent);
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerObject 
      *opp_NearestFromSelf = wm.getOpponentNearestToSelf(10);

    int ball_reach_step = 0;
    if ( ! wm.existKickableTeammate()
         && ! wm.existKickableOpponent() )
    {
        ball_reach_step
            = std::min( wm.interceptTable()->teammateReachCycle(),
                        wm.interceptTable()->opponentReachCycle() );
    }
    const rcsc::Vector2D base_pos
        = wm.ball().inertiaPoint( ball_reach_step );

    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, wm.offsideLineX() - 1.0 );
    }

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: HOME POSITION=(%.2f, %.2f) base_point(%.1f %.1f)"
                        ,__FILE__,
                        home_pos.x, home_pos.y,
                        base_pos.x, base_pos.y );

    switch ( Strategy::get_ball_area( base_pos ) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
      if(wm.existKickableOpponent())
       {
	 if ( Bhv_BasicTackle(0.5,90.0).execute(agent)){break;}
    	 if ( wm.ball().distFromSelf() < 8)
	   {
	     rcsc::Body_Intercept2008().execute( agent );
	     agent->setNeckAction( new rcsc::Neck_TurnToBall );
	     break;
	   }
    	 else
	   {
	     if(opp_NearestFromSelf!=NULL)
	       {
		 home_pos = opp_NearestFromSelf->pos();
	       }
	   }
       }
      else if(!wm.existKickableOpponent() &&
	      !wm.existKickableTeammate())
	{ 
	  //Tonegawa:
	  //自軍向きのこぼれ玉はとりあえず取っておく
	  if(wm.ball().vel().x < 0 && wm.ball().distFromSelf() < 12)
	    //ボールのベクトルが自軍向き
	    {
	      rcsc::Body_Intercept2008(false).execute( agent );
	      agent->setNeckAction( new rcsc::Neck_TurnToBall );
	      break;
	    }
	}
      Bhv_BasicMove( home_pos ).execute( agent );
      break;


    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
      if(wm.existKickableOpponent()&&
	 wm.self().pos().x<wm.ball().pos().x)
	{
	  if ( Bhv_BasicTackle(0.5,90.0).execute(agent)){break;}
	  if ( wm.ball().distFromSelf() < 5.0 )
	    {
	      rcsc::Body_Intercept2008().execute( agent );
	      agent->setNeckAction( new rcsc::Neck_TurnToBall );
	      break;
	    }
	  else
	    {
		if(opp_NearestFromSelf!=NULL)
		{
		  home_pos = opp_NearestFromSelf->pos();
		}
	    }
	}
      else if(!wm.existKickableOpponent() &&
	      !wm.existKickableTeammate())
	{ 
	  //Tonegawa:
	  //自軍向きのこぼれ玉はとりあえず取っておく
	  if(wm.ball().vel().x < 0 && wm.ball().distFromSelf() < 12)
	    //ボールのベクトルが自軍向き
	    {
	      rcsc::Body_Intercept2008(false).execute( agent );
	      agent->setNeckAction( new rcsc::Neck_TurnToBall );
	      break;
	    }
	}//else if-End
 
      Bhv_BasicMove( home_pos ).execute( agent );
      break;


    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
      //Tonegawa:
      //自軍向きのこぼれ玉はとりあえず取っておく
      if(!wm.existKickableOpponent() && 
	 !wm.existKickableTeammate())
	{
	  if(wm.ball().vel().x < 0 && wm.ball().distFromSelf() < 12)
	    //ボールのベクトルが自軍向き
	    {
	      rcsc::Body_Intercept2008(false).execute( agent );
	      agent->setNeckAction( new rcsc::Neck_TurnToBall );
	      break;
	    }
	}
      Bhv_BasicMove( home_pos ).execute( agent );
      break;


    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
    default:
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
    }
}


bool
RoleOffensiveHalf::FSClearBall( rcsc::PlayerAgent * agent ){

  const rcsc::WorldModel & wm = agent->world();

  int fewest_area_num_first=-1;
  int min_opps_count=7;

  int opps_num=0;
  double opp_dist=20;
  const rcsc::PlayerPtrCont::const_iterator
    end=wm.opponentsFromSelf().end();
  for(rcsc::PlayerPtrCont::const_iterator
	it=wm.opponentsFromSelf().begin();
      it!=end;++it){
    opp_dist = (*it)->pos().dist(wm.self().pos());
    opps_num++;
    if(opps_num>7)
      break;
  }
  //first search
  for(int i=0;i<3;i++){
    rcsc::Sector2D checkArea(wm.self().pos()
			     ,0.0,opp_dist
			     ,i*60-90,(i+1)*60-90);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_first=i;
    } 
  }

  if(fewest_area_num_first==-1)
    return FSClearBall_v1(agent);

  int fewest_area_num_second=-1;
  min_opps_count=5;
  //second search
  for(int i=0;i<3;i++){
    rcsc::Sector2D checkArea(wm.self().pos()
			     ,0.0,opp_dist
			     ,fewest_area_num_first*60 + i*20 -90
			     ,fewest_area_num_first*60 + (i+1)*20 -90);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_second=i;
    } 
  }

  if(fewest_area_num_second==-1){
    rcsc::Vector2D target_pos
      = rcsc::Vector2D(wm.self().pos()).setPolar(opp_dist
						 ,fewest_area_num_first*60 
						 +30-90);
  
    double first_speed = rcsc::ServerParam::i().maxPower();
  
    if(rcsc::Body_SmartKick(target_pos
			    ,first_speed
			    ,first_speed * 0.96
			    ,2
			    ).execute(agent)){
      agent->setNeckAction(new rcsc::Neck_TurnToBall());
      //std::cout << "move" << std::endl;
      return true;
    }
  }
  
  int fewest_area_num_third=-1;
  min_opps_count=3;
  //third search
  for(int i=0;i<4;i++){
    rcsc::Sector2D checkArea(wm.self().pos()
			     ,0.0,opp_dist
			     ,fewest_area_num_first*60 
			     + fewest_area_num_second*20
			     + i*5 -90
			     ,fewest_area_num_first*60 
			     + fewest_area_num_second*20
			     + (i+1)*5 -90);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_third=i;
    } 
  }

  if(fewest_area_num_third==-1){
    rcsc::Vector2D target_pos
      = rcsc::Vector2D(wm.self().pos()).setPolar(opp_dist
						 ,fewest_area_num_first*60 
						 +fewest_area_num_second*20
						 +10-90);
    
    double first_speed = rcsc::ServerParam::i().maxPower();
    
    if(rcsc::Body_SmartKick(target_pos
			    ,first_speed
			    ,first_speed * 0.96
			    ,2
			    ).execute(agent)){
      agent->setNeckAction(new rcsc::Neck_TurnToBall());
      //std::cout << "move" << std::endl;
      return true;
    }
  }

  rcsc::Vector2D target_pos
    = rcsc::Vector2D(wm.self().pos()).setPolar(opp_dist
				,fewest_area_num_first*60 
					      + fewest_area_num_second*20
				+ fewest_area_num_third*5 + 2.5 -90);
  
  double first_speed = rcsc::ServerParam::i().maxPower();
  
  if(rcsc::Body_SmartKick(target_pos
			  ,first_speed
			  ,first_speed * 0.96
			  ,2
			  ).execute(agent)){
    agent->setNeckAction(new rcsc::Neck_TurnToBall());
    //std::cout << "move" << std::endl;
    return true;
  }
  
  return FSClearBall_v1(agent);
}

bool
RoleOffensiveHalf::FSClearBall_v1( rcsc::PlayerAgent * agent ){
  
  const rcsc::WorldModel & wm = agent->world();

  rcsc::Vector2D firstOpp(100,100);
  rcsc::Vector2D secondOpp(100,100);
  const rcsc::PlayerPtrCont::const_iterator
    end=wm.opponentsFromSelf().end();
  for(rcsc::PlayerPtrCont::const_iterator
	it=wm.opponentsFromSelf().begin();
      it!=end;++it){
    if((*it)->pos().x>wm.self().pos().x){
      if(firstOpp.x==100&&firstOpp.y==100){
	firstOpp=(*it)->pos();
      }else if(secondOpp.x==100&&secondOpp.y==100){
	secondOpp=(*it)->pos();
	break;
      }
    }
  }
  if(firstOpp.x!=100&&firstOpp.y!=100){
    rcsc::Vector2D target_point=firstOpp+secondOpp;
    target_point/=2;
    if(secondOpp.x!=100&&secondOpp.y!=100
       && firstOpp.dist(secondOpp)>7 
       && firstOpp.dist(wm.self().pos())>1.5
       && wm.self().pos().dist(target_point)!=0
       && rcsc::Line2D(firstOpp,secondOpp).dist(wm.self().pos())
       /wm.self().pos().dist(target_point)>0.7){
      double first_speed = rcsc::ServerParam::i().ballSpeedMax();
      double thr = 0.96;
      int step = 2;
      if(rcsc::Body_SmartKick(target_point,
			      first_speed,
			      first_speed * thr,
			      step).execute(agent)){
	agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());
	return true;
      }
    }else{
      rcsc::Line2D courseR(wm.self().pos(),firstOpp.add(-1.5,1.5));
      rcsc::Line2D courseL(wm.self().pos(),firstOpp.add(-1.5,-1.5));
      double VCourseR=0,VCourseL=0;
      for(rcsc::PlayerPtrCont::const_iterator
	    it=wm.opponentsFromSelf().begin();
	  it!=end && (*it)->distFromSelf()<40;++it){
	if((*it)->pos().x>wm.self().pos().x
	   && !(*it)->goalie()){
	  VCourseR+=courseR.dist2((*it)->pos());
	  VCourseL+=courseL.dist2((*it)->pos());
	}
      }
      double first_speed = rcsc::ServerParam::i().ballSpeedMax();
      double thr = 0.96;
      int step = 2;
      if(VCourseL<VCourseR){
	if(rcsc::Body_SmartKick(firstOpp.add(-1.5,1.5),
				first_speed,
				first_speed * thr,
				step).execute(agent)){
	  agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());
	  return true;
	}
      }else{
	if(rcsc::Body_SmartKick(firstOpp.add(-1.5,-1.5),
				first_speed,
				first_speed * thr,
				step).execute(agent)){
	  agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());
	  return true;
	}
      }
    }
  }
  return false;
}

int
RoleOffensiveHalf::countKickableCycleAtSamePoint(rcsc::PlayerAgent * agent){
  static int startCycle = 0;
  static int prevCycle = 0;
  static rcsc::Vector2D point(0,0);

  if(agent->world().self().isKickable()
     && point.dist(agent->world().ball().pos()) < 3.5
     && agent->world().time().cycle() - prevCycle <= 7){
    prevCycle = agent->world().time().cycle();
  }else{
    startCycle = agent->world().time().cycle();
    prevCycle = agent->world().time().cycle();
    point = agent->world().ball().pos();
  }
  //printf("start:%2d,prev:%2d(%.2lf,%.2lf)\n"
  // ,startCycle
  // ,prevCycle,point.x,point.y);
  return agent->world().time().cycle()-startCycle;
}
