// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_side_forward.h"

#include "strategy.h"

#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_move.h"
#include "bhv_basic_tackle.h"

#include <rcsc/formation/formation.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>


#include <rcsc/action/neck_turn_to_goalie_or_scan.h>
#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_go_to_point.h>

#include <rcsc/action/body_smart_kick.h>

#include <rcsc/geom/sector_2d.h>

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::execute( rcsc::PlayerAgent * agent )
{
  
  bool kickable = agent->world().self().isKickable();
  if ( agent->world().existKickableTeammate()
       && agent->world().teammatesFromBall().front()->distFromBall()
       < agent->world().ball().distFromSelf() )
    {
      kickable = false;
    }
  
    if ( kickable )
      {
        doKick( agent );
      }
    else
      {
        doMove( agent );
      }
    
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::doKick( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    rcsc::Vector2D pass_point;
    rcsc::Vector2D Dribble_point=rcsc::Vector2D( 52.5
						,wm.self().pos().y);

    //nearest opponent
    const rcsc::PlayerObject *Opp = wm.getOpponentNearestToSelf(10);
    double opp_dist = wm.getDistOpponentNearestToSelf(10);
    //RaidAttack
    bool force = false;

    switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
  //-----------------------オフェンスフィールド--------------------------
    case Strategy::BA_OffMidField:
    case Strategy::BA_DribbleAttack:
      //オフサイドラインを越えてたら、ゴールまで1直線
      if(wm.self().pos().x>wm.offsideLineX()){
	if(!rcsc::Body_Dribble2008(Dribble_point,4.0,
				   rcsc::ServerParam::i().maxDashPower(),5,
				   force).execute( agent )){
	}else{
	  agent->setNeckAction(new rcsc::Neck_TurnToGoalieOrScan());
	  break;
	}
      }
      //---パスが出きるのであればパス---
      if(rcsc::Body_Pass::get_best_pass(wm, &pass_point, NULL, NULL)
	 &&opp_dist<=5.0)
	{ 
	  if( wm.self().pos().dist(pass_point) > 6.0
	      && (wm.self().pos().x - 4.0) < pass_point.x
	      && (rcsc::ServerParam::i().useOffside() 
		  && pass_point.x<wm.offsideLineX()
		  || !rcsc::ServerParam::i().useOffside())
	      && wm.countOpponentsIn(rcsc::Circle2D(pass_point,3.0)
				     ,7,true)<=1 )
	    {
	      rcsc::Body_Pass().execute(agent);
	      agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
	      break;
	    }
	}
      //---そうでなければ強行突破---
      
      if(opp_dist<=10.0)force=true;//回避モード //パラメータ変更(真崎)
      if(wm.self().pos().x<25.0)
	{//---SideLine Dribble---
	  if(wm.self().pos().y> 0)Dribble_point.y= 32.0;
	  if(wm.self().pos().y< 0)Dribble_point.y=-32.0;
	}
      else
	{
	  Dribble_point.x = 52.5;
	  Dribble_point.y = (wm.self().pos().y*2/3+2.0-rand()%5);
	}
      if(countKickableCycleAtSamePoint(agent)<12){
	//printf("%2d:kickable\n",countKickableCycleAtSamePoint(agent));
	rcsc::Body_Dribble2008(Dribble_point
			       ,4.0,rcsc::ServerParam::i().maxDashPower()
			       ,5,force).execute( agent );
      }else{
	//ドリブルできないときは、てきとークリア
	if(FSClearBall(agent))
	  break;
      }
      
      agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
      break;
  //------------------------クロスフィールド-----------------------------
    case Strategy::BA_Cross:
      if(rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL)
	 && Strategy::get_ball_area(pass_point) 
	 == Strategy::BA_ShootChance
	 &&(rcsc::ServerParam::i().useOffside() 
	    && pass_point.x<wm.offsideLineX()
	    || !rcsc::ServerParam::i().useOffside()))
	{//PassPoint is in ShootChanceArea
	  rcsc::Body_Pass().execute(agent);
	}
      else
	{
	  rcsc::Body_Dribble2008(rcsc::Vector2D(46.0,0),5.0,
				 rcsc::ServerParam::i().maxDashPower()*0.8,
				 3,false).execute( agent );
	}
      agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
      break;
 //---------------------------ゴール間際--------------------------------
    case Strategy::BA_ShootChance:
      if(rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL))
	{
	  if(pass_point.dist(rcsc::Vector2D(52.5,0))<
	     wm.self().pos().dist(rcsc::Vector2D(52.5,0))-5)//パラメータ追加(真崎)無駄パスにならない？
	    {//--ゴールに近い味方がいたらパス--
	      if(rcsc::ServerParam::i().useOffside() 
		 && pass_point.x<wm.offsideLineX()
		 || !rcsc::ServerParam::i().useOffside()){
		rcsc::Body_Pass().execute(agent);
		agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
		break;
	      }
	    }
	}
      if(wm.self().pos().dist(rcsc::Vector2D(52.5,0))<16.0)
        {
         //--そうでなければシュート--
	  
	  rcsc::Vector2D 
	    target_pos(52.5
		       ,rcsc::ServerParam::i().goalHalfWidth() - 0.5);
	  const rcsc::PlayerObject *goalie = wm.getOpponentGoalie();
	  rcsc::Vector2D opp_goalie_pos(50,0);
	  if(goalie!=NULL)
	    opp_goalie_pos = goalie->pos();
	  if( wm.self().pos().absY() > 5){ 
	    if(wm.self().pos().y < 0)
	      target_pos.y *= -1;
	  }else{
	    if( opp_goalie_pos.y>0)
	      target_pos.y *= -1;
	  }
	  double first_speed = rcsc::ServerParam::i().ballSpeedMax();
	  double thr = 0.96;
	  int step = 2;
	  rcsc::Body_SmartKick(target_pos,
			       first_speed,
			       first_speed * thr,
			       step).execute(agent);
	  agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );   
	  break;
	}
	else
	{
	  if(opp_dist<=1.0)force=true;
	  rcsc::Body_Dribble2008(rcsc::Vector2D(50.0,0),5.0,
			     rcsc::ServerParam::i().maxDashPower(),
			     2,force).execute( agent );
	  agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );   
	  break;
	}
      Bhv_BasicOffensiveKick().execute( agent );
      break;
      //---------------------------------------------------------------
    case Strategy::BA_DefMidField:
    case Strategy::BA_DribbleBlock:
	//---Deffense region ---
	 //---パスが出きるのであればパス---
	 if(rcsc::Body_Pass::get_best_pass(wm, &pass_point, NULL, NULL)
	    &&opp_dist<=8.0)
	    { 
		if(wm.self().pos().dist(pass_point)>7.0 && 
		   (wm.self().pos().x - 5.0)<pass_point.x)
		{
		    rcsc::Body_Pass().execute(agent);
		    agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
		    break;
		}
	    }
         //---そうでなければ強行突破---
	 if(opp_dist<=18.0)
	   force=true;//回避モード
	 //---SideLine Dribble---
	 Dribble_point.x=10.0;
	 if(wm.self().pos().y> 0)Dribble_point.y= 22.0;
	 if(wm.self().pos().y< 0)Dribble_point.y=-22.0;
	  rcsc::Body_Dribble2008(Dribble_point,4.0,
			     rcsc::ServerParam::i().maxDashPower(),2,
			     force).execute( agent );
	 agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
	 break;
  //--------------------------その他の領域------------------------------
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    default:
      Bhv_BasicOffensiveKick().execute( agent );
      break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    int ball_reach_step = 0;
	int self_min = wm.interceptTable()->selfReachCycle();//add by t-suzuki
    if ( ! wm.existKickableTeammate()
         && ! wm.existKickableOpponent() )
    {
        ball_reach_step
            = std::min( wm.interceptTable()->teammateReachCycle(),
                        wm.interceptTable()->opponentReachCycle() );
    }
    const rcsc::Vector2D base_pos
        = wm.ball().inertiaPoint( ball_reach_step );

    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, wm.offsideLineX() - 1.0 );
    }

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: HOME POSITION=(%.2f, %.2f) base_point(%.1f %.1f)"
                        ,__FILE__,
                        home_pos.x, home_pos.y,
                        base_pos.x, base_pos.y );

    const rcsc::PlayerObject *Opp = wm.getOpponentNearestToSelf(10);
    //Strategyの部分
    switch ( Strategy::get_ball_area( base_pos ) ) {
    case Strategy::BA_Cross:
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
    case Strategy::BA_DefMidField:
      if(!wm.existKickableTeammate()
	 &&wm.ball().pos().dist(wm.self().pos())<4
	 && rcsc::Body_Intercept2008(false).execute( agent ) )
	{
	  agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
	  break;
	}
      //オフサイドの裏をとれそうなら、スタミナ無視ダッシュ
      if(!wm.existKickableOpponent()
	 && wm.ball().pos().x>wm.offsideLineX()
	 && wm.getTeammateNearestToBall(5)!=NULL
	 && wm.getTeammateNearestToBall(5)->isSelf()){
	if(rcsc::Body_GoToPoint(wm.ball().pos()
				,0.5, rcsc::ServerParam::i().maxDashPower()
				,5,false,false,12).execute(agent)){
	  agent->setNeckAction(new rcsc::Neck_TurnToBall());
	  break;
	}
      }
      if(!wm.existKickableTeammate())
	{
	  if(wm.existKickableOpponent() && Opp!=NULL)
	    {
	      home_pos = Opp->pos();
	      home_pos.y -= 5.0;
	    }
	  //if(wm.self().pos().dist(wm.ball().pos())<4.0)
	  if( self_min <= ball_reach_step - 2 )//changed by t-suzuki    
	    {//---ボールが取れそう---
	      if ( Bhv_BasicTackle(0.5,90.0).execute(agent)){break;}
	      if( rcsc::Body_Intercept2008(false).execute( agent )){
		//agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
		agent->setNeckAction(new rcsc::Neck_TurnToBall());//ボール追うときはボール見た方がいいんじゃね? by t-suzuki
		break;
	      }
            }
	}
	
	//-- カウンターのために設置　自分より前(-α)をドリブルしてる仲間がいるなら良いパスがもらえる位置までダッシュ --
	if( ( wm.existKickableTeammate()
		|| wm.interceptTable()->teammateReachCycle() < wm.interceptTable()->opponentReachCycle() )
		&& wm.ball().pos().x > wm.self().pos().x - 10
		&& wm.ball().pos().y * wm.self().pos().y > 0 )
	{
		rcsc::Body_GoToPoint(rcsc::Vector2D(wm.offsideLineX()-1.0, home_pos.y)
				  ,5.0,rcsc::ServerParam::i().maxPower()
				  ).execute(agent);
		agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());
		break;
	}
	
	Bhv_BasicMove( home_pos ).execute( agent );
	break;

    case Strategy::BA_ShootChance:
      if(wm.existKickableTeammate()
	 && wm.self().pos().x>wm.offsideLineX())
	{
	  if(rcsc::Body_GoToPoint(rcsc::Vector2D(wm.offsideLineX()-0.5
						 ,home_pos.y)
				  ,0.3,rcsc::ServerParam::i().maxDashPower()*0.5
				  ).execute(agent)){
	    agent->setNeckAction(new rcsc::Neck_TurnToGoalieOrScan());
	    break;
	  }
	}
      	Bhv_BasicMove( home_pos ).execute( agent );
	break;
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_Danger:

    default:
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
    }
}

bool
RoleSideForward::FSClearBall( rcsc::PlayerAgent * agent ){

  const rcsc::WorldModel & wm = agent->world();
  
  int fewest_area_num_first=-1;
  int min_opps_count=7;

  int opps_num=0;
  double opp_dist=20;
  const rcsc::PlayerPtrCont::const_iterator
    end=wm.opponentsFromSelf().end();
  for(rcsc::PlayerPtrCont::const_iterator
	it=wm.opponentsFromSelf().begin();
      it!=end;++it){
    opp_dist = (*it)->pos().dist(wm.self().pos());
    opps_num++;
    if(opps_num>7)
      break;
  }
  //first search
  for(int i=0;i<3;i++){
    rcsc::Sector2D checkArea(wm.self().pos()
			     ,0.0,opp_dist
			     ,i*60-90,(i+1)*60-90);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_first=i;
    } 
  }

  if(fewest_area_num_first==-1)
    return FSClearBall_v1(agent);

  int fewest_area_num_second=-1;
  min_opps_count=5;
  //second search
  for(int i=0;i<3;i++){
    rcsc::Sector2D checkArea(wm.self().pos()
			     ,0.0,opp_dist
			     ,fewest_area_num_first*60 + i*20 -90
			     ,fewest_area_num_first*60 + (i+1)*20 -90);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_second=i;
    } 
  }

  if(fewest_area_num_second==-1){
    rcsc::Vector2D target_pos
      = rcsc::Vector2D().setPolar(opp_dist
				  ,fewest_area_num_first*60 
				  +30-90);
    target_pos += wm.self().pos();
    double first_speed = rcsc::ServerParam::i().maxPower();

    if(rcsc::Body_SmartKick(target_pos
			    ,first_speed
			    ,first_speed * 0.96
			    ,2
			    ).execute(agent)){
      agent->setNeckAction(new rcsc::Neck_TurnToBall());
      //std::cout << "move" << std::endl;
      return true;
    }
  }
  
  int fewest_area_num_third=-1;
  min_opps_count=3;
  //third search
  for(int i=0;i<4;i++){
    rcsc::Sector2D checkArea(wm.self().pos()
			     ,0.0,opp_dist
			     ,fewest_area_num_first*60 
			     + fewest_area_num_second*20
			     + i*5 -90
			     ,fewest_area_num_first*60 
			     + fewest_area_num_second*20
			     + (i+1)*5 -90);
    int opps_count 
      = wm.countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_third=i;
    } 
  }

  if(fewest_area_num_third==-1){
    rcsc::Vector2D target_pos
      = rcsc::Vector2D().setPolar(opp_dist
				  ,fewest_area_num_first*60 
				  +fewest_area_num_second*20
				  +10-90);
    target_pos += wm.self().pos();
    double first_speed = rcsc::ServerParam::i().maxPower();

    if(rcsc::Body_SmartKick(target_pos
			    ,first_speed
			    ,first_speed * 0.96
			    ,2
			    ).execute(agent)){
      agent->setNeckAction(new rcsc::Neck_TurnToBall());
      //std::cout << "move" << std::endl;
      return true;
    }
  }

  rcsc::Vector2D target_pos
    = rcsc::Vector2D().setPolar(opp_dist
				,fewest_area_num_first*60 
				+ fewest_area_num_second*20
				+ fewest_area_num_third*5 + 2.5 -90);
   target_pos += wm.self().pos();
  double first_speed = rcsc::ServerParam::i().maxPower();

  if(rcsc::Body_SmartKick(target_pos
			  ,first_speed
			  ,first_speed * 0.96
			  ,2
			  ).execute(agent)){
    agent->setNeckAction(new rcsc::Neck_TurnToBall());
    //std::cout << "move" << std::endl;
    return true;
  }
  
  return FSClearBall_v1(agent);
}

bool
RoleSideForward::FSClearBall_v1( rcsc::PlayerAgent * agent ){
  
  const rcsc::WorldModel & wm = agent->world();

  rcsc::Vector2D firstOpp(100,100);
  rcsc::Vector2D secondOpp(100,100);
  const rcsc::PlayerPtrCont::const_iterator
    end=wm.opponentsFromSelf().end();
  for(rcsc::PlayerPtrCont::const_iterator
	it=wm.opponentsFromSelf().begin();
      it!=end;++it){
    if((*it)->pos().x>wm.self().pos().x){
      if(firstOpp.x==100&&firstOpp.y==100){
	firstOpp=(*it)->pos();
      }else if(secondOpp.x==100&&secondOpp.y==100){
	secondOpp=(*it)->pos();
	break;
      }
    }
  }
  if(firstOpp.x!=100&&firstOpp.y!=100){
    rcsc::Vector2D target_point=firstOpp+secondOpp;
    target_point/=2;
    if(secondOpp.x!=100&&secondOpp.y!=100
       && firstOpp.dist(secondOpp)>7 
       && firstOpp.dist(wm.self().pos())>1.5
       && wm.self().pos().dist(target_point)!=0
       && rcsc::Line2D(firstOpp,secondOpp).dist(wm.self().pos())
       /wm.self().pos().dist(target_point)>0.7){
      double first_speed = rcsc::ServerParam::i().ballSpeedMax();
      double thr = 0.96;
      int step = 2;
      if(rcsc::Body_SmartKick(target_point,
			      first_speed,
			      first_speed * thr,
			      step).execute(agent)){
	agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());
	return true;
      }
    }else{
      rcsc::Line2D courseR(wm.self().pos(),firstOpp.add(-1.5,1.5));
      rcsc::Line2D courseL(wm.self().pos(),firstOpp.add(-1.5,-1.5));
      double VCourseR=0,VCourseL=0;
      for(rcsc::PlayerPtrCont::const_iterator
	    it=wm.opponentsFromSelf().begin();
	  it!=end && (*it)->distFromSelf()<40;++it){
	if((*it)->pos().x>wm.self().pos().x
	   && !(*it)->goalie()){
	  VCourseR+=courseR.dist2((*it)->pos());
	  VCourseL+=courseL.dist2((*it)->pos());
	}
      }
      double first_speed = rcsc::ServerParam::i().ballSpeedMax();
      double thr = 0.96;
      int step = 2;
      if(VCourseL<VCourseR){
	if(rcsc::Body_SmartKick(firstOpp.add(-1.5,1.5),
				first_speed,
				first_speed * thr,
				step).execute(agent)){
	  agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());
	  return true;
	}
      }else{
	if(rcsc::Body_SmartKick(firstOpp.add(-1.5,-1.5),
				first_speed,
				first_speed * thr,
				step).execute(agent)){
	  agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());
	  return true;
	}
      }
    }
  }
  return false;
}

int
RoleSideForward::countKickableCycleAtSamePoint(rcsc::PlayerAgent * agent){
  static int startCycle = 0;
  static int prevCycle = 0;
  static rcsc::Vector2D point(0,0);

  if(agent->world().self().isKickable()
     && point.dist(agent->world().ball().pos()) < 3.5
     && agent->world().time().cycle() - prevCycle <= 7){
    prevCycle = agent->world().time().cycle();
  }else{
    startCycle = agent->world().time().cycle();
    prevCycle = agent->world().time().cycle();
    point = agent->world().ball().pos();
  }
  //printf("start:%2d,prev:%2d(%.2lf,%.2lf)\n"
  // ,startCycle
  // ,prevCycle,point.x,point.y);
  return agent->world().time().cycle()-startCycle;
}
