#include "selectpass.h"
#include <string>
#include <vector>
#include <list>
#include <fstream>
#include <iostream>
#include <rcsc/player/intercept_table.h>
#include <rcsc/math_util.h>
/* re_kuma */
#include <boost/tokenizer.hpp>		// for tokenizer
#include <boost/lexical_cast.hpp>	// for lexical_cast
#include <boost/cast.hpp>
#include <boost/random.hpp>			// for random
#include <rcsc/action/body_kick_multi_step.h>
/* end re_kuma */

/*
 * body_passのヘッダ
 */
#include <rcsc/action/intention_kick.h>
#include <rcsc/action/body_stop_ball.h>
#include <rcsc/action/body_hold_ball.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/audio_sensor.h>
#include <rcsc/player/say_message_builder.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/rect_2d.h>
#include <rcsc/geom/sector_2d.h>
#include <rcsc/soccer_math.h>
#include <rcsc/math_util.h>
/* body_passのヘッダ 終了 */

/* bhv_basic_offensive_kickのヘッダ */
#include "body_kick_to_corner.h"

#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
/* bhv_basic_offensive_kickのヘッダ 終了 */

/*** ACTION ***/
void SelectPass::execute(rcsc::PlayerAgent* agent)
{
	/*** Boltzmann select execute ***/
	int target_num = 0;
	agentp = agent;		  //Boltzmann用にPlayerAgentを保存
	target_num = Boltzmann();	 //パス先を決定

	//targetの座標をとってくる
	const AbstractPlayerObject * target = getteam(agent, target_num);
	if( !target )
	{
		std::cout << "Don'get teammate:" << target_num << std::endl;
		return;
	};
	Vector2D target_vec = target->pos();


	/*** body_passから流用 ***/
	/*** body_passのパス先の位置をselectpassで選択 ***/
    dlog.addText( Logger::ACTION,
                  "%s:%d: Body_Pass. execute()"
                  ,__FILE__, __LINE__ );

    // キックできない場合
    if ( ! agent->world().self().isKickable() )
    {
        std::cerr << __FILE__ << ": " << __LINE__
                  << " not ball kickable!"
                  << std::endl;
        dlog.addText( Logger::ACTION,
                      "%s:%d:  not kickable"
                      ,__FILE__, __LINE__ );
        return;
    }

	//第二行動カウントを保存
	setLastActionTime( agent->world().time().cycle() );
	a = target_num;
	//第一行動のカウントを保存
	if( !learnFlag )
	{
		setFirstActionTime( agent->world().time().cycle() );
		a_F = target_num;
	}
	setLearnFlag( true ); //学習フラグをオンにする

    /*** initialize end ****/

	if( target_num != agent->world().self().unum() )
    	//他のチームメートにパスできると
    {
    	if( pass( agent , target_num , target_vec ) );
    		return;
    }else
    //dribble command
    {
    	if( dribble( agent ) )
    		return;
    }
    return;
};

const rcsc::AbstractPlayerObject *  SelectPass::getteam( rcsc::PlayerAgent * agent , int num )
{
	AbstractPlayerCont team = agent->world().allTeammates();
	for(	AbstractPlayerCont::const_iterator target_it = team.begin();
			target_it != team.end();
			target_it++
			)
		if( (*target_it)->unum() == num ) return (*target_it);
	return ( *team.begin() );

};

const rcsc::AbstractPlayerObject * SelectPass::getopp( rcsc::PlayerAgent * agent , int num )
{
	AbstractPlayerCont opp = agent->world().allOpponents();
	for(	AbstractPlayerCont::const_iterator target_it = opp.begin();
			target_it != opp.end();
			target_it++
			)
		if( (*target_it)->unum() == num ) return (*target_it);
	return ( *opp.begin() );
};


bool SelectPass::pass(rcsc::PlayerAgent * agent , int num , rcsc::Vector2D target_vec)
{
	/*** body_passから流用 ***/
	/*** body_passのパス先の位置をselectpassで選択 ***/
	//std::cout << "cycle:" << agent->world().time().cycle() << " unum:" << agent->world().self().unum() << ": can pass to " << num << std::endl;

	//速度生成
	double end_speed = 1.5;
	double first_speed = 100.0;
	const Vector2D receiver_rel = target_vec - agent->world().ball().pos();
	const double receiver_dist = receiver_rel.r();
	const AngleDeg receiver_angle = receiver_rel.th();
	do
	{
		first_speed
		= calc_first_term_geom_series_last
			( end_speed,
					receiver_dist,
					ServerParam::i().ballDecay() );
		//= required_first_speed( receiver_dist,
		//ServerParam::i().ballDecay(),
		//end_speed );
		//printcrash(agent->world().self().unum(), "キック速度を調整しました");
		if ( first_speed < ServerParam::i().ballSpeedMax() )
		{
			//printcrash(agent->world().self().unum(), "キック限界スピードを越えました");
			break;
		}
		end_speed -= 0.1;
	}
	while ( end_speed > 0.8 );

//    if ( first_speed > ServerParam::i().ballSpeedMax() )
//    {
//		printcrash(agent->world().self().unum(), "キック限界スピードを越えました");
//    	agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
//    	return false;
//    }
	agent->debugClient().addMessage( "pass" );
	agent->debugClient().setTarget( target_vec );

	if(
			Body_KickMultiStep( target_vec,
					first_speed,
					false
			).execute( agent )
	)
	{
		//printcrash(agent->world().self().unum(), "キックしました");
		agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
		return true;
	}

	dlog.addText( Logger::ACTION,
			"%s:%d: execute() register pass intention"
			,__FILE__, __LINE__ );


	if ( agent->world().gameMode().type() != GameMode::PlayOn )
	{
		agent->setIntention
		( new IntentionKick( target_vec,
				first_speed,
				3, // max kick step
				false, // not enforce
				agent->world().time() ) );
	}

	if ( agent->config().useCommunication()
			&& num != Unum_Unknown )
	{
		dlog.addText( Logger::ACTION,
				"%s:%d: execute() set pass communication."
				,__FILE__, __LINE__ );
		Vector2D target_buf = target_vec - agent->world().self().pos();
		target_buf.setLength( 1.0 );

		agent->addSayMessage( new PassMessage( num,
				target_vec + target_buf,
				agent->effector().queuedNextBallPos(),
				agent->effector().queuedNextBallVel() ) );
	}
	//printcrash(agent->world().self().unum(), "行動しませんでした");
	agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
	return true;
};

bool SelectPass::dribble(rcsc::PlayerAgent * agent)
{
	/*** Bhv_BasicOffensiveKickの流用 ***/
	/*** パスの部分を除いてありますが、基本そのままです ***/
	//std::cout << "cycle:" << agent->world().time().cycle() << " unum:" << agent->world().self().unum() << ": dribble " << std::endl;
	const rcsc::WorldModel & wm = agent->world();

	const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();
	const rcsc::PlayerObject * nearest_opp
	= ( opps.empty()
			? static_cast< rcsc::PlayerObject * >( 0 )
					: opps.front() );
    const double nearest_opp_dist = ( nearest_opp
    		? nearest_opp->distFromSelf()
    				: 1000.0 );
    const rcsc::Vector2D nearest_opp_pos = ( nearest_opp
    		? nearest_opp->pos()
    				: rcsc::Vector2D( -1000.0, 0.0 ) );

    if ( nearest_opp_dist < 5.0
    		&& nearest_opp_dist > ( rcsc::ServerParam::i().tackleDist()
    				+ rcsc::ServerParam::i().defaultPlayerSpeedMax() * 1.5 )
    				&& wm.self().body().abs() < 70.0 )
    {
    	const rcsc::Vector2D body_dir_drib_target
    	= wm.self().pos()
    	+ rcsc::Vector2D::polar2vector(5.0, wm.self().body());

    	int max_dir_count = 0;
    	wm.dirRangeCount( wm.self().body(), 20.0, &max_dir_count, NULL, NULL );

    	if ( body_dir_drib_target.x < rcsc::ServerParam::i().pitchHalfLength() - 1.0
    			&& body_dir_drib_target.absY() < rcsc::ServerParam::i().pitchHalfWidth() - 1.0
    			&& max_dir_count < 3
    	)
    	{
    		// check opponents
    		// 10m, +-30 degree
    		const rcsc::Sector2D sector( wm.self().pos(),
    				0.5, 10.0,
    				wm.self().body() - 30.0,
    				wm.self().body() + 30.0 );
    		// opponent check with goalie
    		if ( ! wm.existOpponentIn( sector, 10, true ) )
    		{
    			rcsc::dlog.addText( rcsc::Logger::TEAM,
    					"%s:%d: dribble to my body dir"
    					,__FILE__, __LINE__ );
    			agent->debugClient().addMessage( "OffKickDrib(1)" );
    			rcsc::Body_Dribble( body_dir_drib_target,
    					1.0,
    					rcsc::ServerParam::i().maxPower(),
    					2
    			).execute( agent );
    			agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
    			return true;
            	}
    	}
    }

    rcsc::Vector2D drib_target( 50.0, wm.self().pos().absY() );
    if ( drib_target.y < 20.0 ) drib_target.y = 20.0;
    if ( drib_target.y > 29.0 ) drib_target.y = 27.0;
    if ( wm.self().pos().y < 0.0 ) drib_target.y *= -1.0;
    const rcsc::AngleDeg drib_angle = ( drib_target - wm.self().pos() ).th();

    // opponent is behind of me
    if ( nearest_opp_pos.x < wm.self().pos().x + 1.0 )
    {
    	// check opponents
    	// 15m, +-30 degree
    	const rcsc::Sector2D sector( wm.self().pos(),
    			0.5, 15.0,
    			drib_angle - 30.0,
    			drib_angle + 30.0 );
    	// opponent check with goalie
    	if ( ! wm.existOpponentIn( sector, 10, true ) )
    	{
    		const int max_dash_step
    		= wm.self().playerType()
    		.cyclesToReachDistance( wm.self().pos().dist( drib_target ) );
    		if ( wm.self().pos().x > 35.0 )
            {
    			drib_target.y *= ( 10.0 / drib_target.absY() );
            }

    		rcsc::dlog.addText( rcsc::Logger::TEAM,
    				"%s:%d: fast dribble to (%.1f, %.1f) max_step=%d"
    				,__FILE__, __LINE__,
    				drib_target.x, drib_target.y,
    				max_dash_step );
    		agent->debugClient().addMessage( "OffKickDrib(2)" );
    		rcsc::Body_Dribble( drib_target,
    				1.0,
    				rcsc::ServerParam::i().maxPower(),
    				std::min( 5, max_dash_step )
    		).execute( agent );
    	}
    	else
    	{
    		rcsc::dlog.addText( rcsc::Logger::TEAM,
    				"%s:%d: slow dribble to (%f, %f)"
    				,__FILE__, __LINE__,
    				drib_target.x, drib_target.y );
    		agent->debugClient().addMessage( "OffKickDrib(3)" );
    		rcsc::Body_Dribble( drib_target,
    				1.0,
    				rcsc::ServerParam::i().maxPower(),
    				2
    		).execute( agent );
    	}
    	agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
    	return true;
    }

    // opp is far from me
    if ( nearest_opp_dist > 5.0 )
    {
    	rcsc::dlog.addText( rcsc::Logger::TEAM,
    			"%s:%d: opp far. dribble(%f, %f)"
    			,__FILE__, __LINE__,
    			drib_target.x, drib_target.y );
    	agent->debugClient().addMessage( "OffKickDrib(4)" );
    	rcsc::Body_Dribble( drib_target,
    			1.0,
    			rcsc::ServerParam::i().maxPower() * 0.4,
    			1
    	).execute( agent );
    	agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
    	return true;
    }

    // opp is near

    // opp is far from me
    if ( nearest_opp_dist > 3.0 )
    {
    	rcsc::dlog.addText( rcsc::Logger::TEAM,
    			"%s:%d: opp far. dribble(%f, %f)"
    			,__FILE__, __LINE__,
    			drib_target.x, drib_target.y );
    	agent->debugClient().addMessage( "OffKickDrib(5)" );
    	rcsc::Body_Dribble( drib_target,
    			1.0,
    			rcsc::ServerParam::i().maxPower() * 0.2,
    			1
    	).execute( agent );
    	agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
    	return true;
    }

    if ( nearest_opp_dist > 2.5 )
    {
    	rcsc::dlog.addText( rcsc::Logger::TEAM,
    			"%s:%d: hold"
    			,__FILE__, __LINE__ );
    	agent->debugClient().addMessage( "OffKickHold" );
    	rcsc::Body_HoldBall().execute( agent );
    	agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
    	return true;
    }

    if ( wm.self().pos().x > wm.offsideLineX() - 10.0 )
    {
    	rcsc::dlog.addText( rcsc::Logger::TEAM,
    			"%s:%d: kick near side"
    			,__FILE__, __LINE__ );
    	agent->debugClient().addMessage( "OffKickToCorner" );
    	Body_KickToCorner( (wm.self().pos().y < 0.0) ).execute( agent );
    	agent->setNeckAction( new rcsc::Neck_ScanField() );
    }
    else
    {
    	rcsc::dlog.addText( rcsc::Logger::TEAM,
    			"%s:%d: advance"
    			,__FILE__, __LINE__ );
    	agent->debugClient().addMessage( "OffKickAdvance" );
    	rcsc::Body_AdvanceBall().execute( agent );
    	agent->setNeckAction( new rcsc::Neck_ScanField() );
    }
    return false;
};

void SelectPass::init(rcsc::PlayerAgent * agent)
{
	agentp = agent;
	cntPossCycle();
};

void SelectPass::learn_init(PlayerAgent * agent)
{
	agentp = agent;

	if(!initFlag){
		isMFLearning(); //学習するかどうか初期化
		initLearning();
		if(MFreaddata == true)
			ReadData();
		initFlag = true;
	}
	if(cntPossCycle() && learnFlag && !isBallInOurPossesion( agent )){
		setReward(-1.0);
		Updating(a_F, a);
		if(MFsavedata == true)
		{
			std::cout << "saveflag on" << std::endl;
			SaveData(a_F, a);
		}
		clearVar();
		setLearnFlag ( false );
	}
};

void SelectPass::isMFLearning( ){

	char LearningParam[128];
	strcpy(LearningParam, getLearningFileName().c_str());
	std::string file_mes;
	std::ifstream filein(LearningParam, std::ios::in);
	if(!filein){
		std::cout << "not found:" << LearningParam << std::endl;
	}
	else{
		typedef boost::tokenizer<boost::char_separator<char> > tokenizer;
		while(getline(filein, file_mes)){
			// 行の先頭が# または 空行なら読み飛ばす
			if(file_mes.substr(0) == "#" || file_mes.empty())
				continue;
			tokenizer tok(file_mes);
			boost::char_separator<char> sep("=");
			tokenizer tokens(file_mes, sep);
			tokenizer::iterator first = tokens.begin();

			if(*first == "isMFReaddata"){
				if(*(++first) == "true")
				{
					std::cout << "重みを読み出します" << std::endl;
					MFreaddata = true;
				}
			}
			else if(*first == "isMFSavedata"){
				if(*(++first) == "true")
				{
					std::cout << "重みを保存します" << std::endl;
					MFsavedata = true;
				}
			}
			else if(*first == "isMFLearn"){
				if(*(++first) == "true")
				{
					std::cout << "学習します" << std::endl;
					MFLearn = true;
				}
			}
		}
	}

	filein.close();
	/*
	  cout << "MFreaddata = " << readdata << endl;
	  cout << "savedata = " << savedata << endl;
	  cout << "MFLearn = " << MFLearn << endl;
	 */
}

void SelectPass::initLearning()
{
	//std::cout << "initLearning OK" << std::endl;
	srand((unsigned)time(NULL) % RAND_MAX);

	for(int i=0; i<MAX_HEURISTICS; i++){
		//w[i] = 10.0 + ((double(rand())/RAND_MAX) * 20.01);
		w[i] = 0.0;
		for(int j=1; j<=MAX_PLAYER; j++)
			tU[j][i] = 0.0;
	}
	//cout << "my Type : " << getAgentObjectType() << endl;
	//cout << "munumber : " << getPlayerNumber() << endl;
	std::string num;
	int obj = agentp->world().self().unum();
	switch(obj){
	case 1:
		num = "1";
		break;
	case 2:
		num = "2";
		break;
	case 3:
		num = "3";
		break;
	case 4:
		num = "4";
		break;
	case 5:
		num = "5";
		break;
	case 6:
		num = "6";
		break;
	case 7:
		num = "7";
		break;
	case 8:
		num = "8";
		break;
	case 9:
		num = "9";
		break;
	case 10:
		num = "10";
		break;
	case 11:
		num = "11";
		break;
	default:
		num = "0";
		break;
	}
	FileName = "learning_" + num + ".csv";
	PlayerCycleFile = "playercyle_" + num + ".csv";
	/*
  	epsilon = 0.1;
  	r = 0.0;
  	T = 0.1;
  	label = 1;
  	cyclelabel = 1;
  	LastActionTime = 0;
  	FirstActionTime = 0;
	 */

	DirPath = dir_0;
	std::string file_mes;
	std::ifstream filein(Learning_ini.c_str(), std::ios::in);
	if(!filein){
		std::cout << "not found:" << Learning_ini << std::endl;
	}
	else{
		typedef boost::tokenizer<boost::char_separator<char> > tokenizer;
		while(getline(filein, file_mes)){
			// 行の先頭が# または 空行なら読み飛ばす
			if(file_mes.substr(0) == "#" || file_mes.empty())
				continue;
			tokenizer tok(file_mes);
			boost::char_separator<char> sep("=");
			tokenizer tokens(file_mes, sep);
			tokenizer::iterator first = tokens.begin();
			if(*first == "isDFLearn"){
				if(*(++first) == "true")
					DirPath = dir_1;   //DF
			}
			else if(*first == "isMFLearn"){
				if(*(++first) == "true"){
					if(DirPath == dir_1)
						DirPath = dir_4;   //DF-MF
					else
						DirPath = dir_2;   //MF
				}
			}
			else if(*first == "isFWLearn"){
				if(*(++first) == "true"){
					if(DirPath == dir_1)
						DirPath = dir_5;   //DF-FW
    		  else if(DirPath == dir_2)
    			  DirPath = dir_6;   //MF-FW
    		  else if(DirPath == dir_4)
    			  DirPath = dir_7;   //DF-MF-FW
    		  else
    			  DirPath = dir_3;   //FW
    	  }
      }
      else if(*first == "epsilon"){
    	  epsilon = boost::lexical_cast<double>(*(++first));
      }
      else if(*first == "temperature"){
    	  T = boost::lexical_cast<double>(*(++first));
    	  //std::cout << "T " << T << std::endl;
      }
      else if(*first == "reward"){
    	  r = boost::lexical_cast<double>(*(++first));
      }
    }
  }
  FileName = DirPath + FileName;
  //std::cout << "FileName: " << FileName << std::endl;
  filein.close();
  //std::cout << "Learning_ini = " << Learning_ini << std::endl;
  /*
  cout << "T = " << T << endl;
  cout << "epsilon = " << epsilon << endl;
  cout << " r = " << r << endl;
  cout << "DirPath = " << DirPath << endl;
  */
};

void SelectPass::SaveData(int ak_F, int ak)
{
	if(Ccnt - FirstActionTime == 0)
	{
		return;
	}
	if(Ccnt - LastActionTime <= 0)
	{
		return;
	}
	//std::cout << "save start" << std::endl;
	// ファイルを追記モードで開く
	std::ofstream fout(FileName.c_str(), std::ios::out | std::ios::app);
	if( !fout )
		std::cout << FileName << "が開くことができません" << std::endl;

	fout << label << ",";		// 学習回数を行頭へ追加
	fout << r2 << ",";
	fout << r << ",";
	fout << ak_F << ",";
	fout << ak << ",";
	// 重み配列を保存

	std::cout << agentp->world().self().unum() << ":保存します" << std::endl;

	for(int i=0; i<MAX_HEURISTICS; i++){
		if(i != MAX_HEURISTICS-1)	// まだヒューリスティックがある
			fout << w[i] << ",";
		else				// 最後のヒューリスティック
			fout << w[i] << std::endl;
	}

	/*
	for(int i=0; i<MAX_HEURISTICS; i++){
		if(i != MAX_HEURISTICS-1)	// まだヒューリスティックがある
			fout << w[i] << ",";
		else				// 最後のヒューリスティック
			fout << w[i] << ",";
	}

	fout << getCurrentCycle() << "," << LastActionTime << endl;
	 */
	//cout <<"mynumber , CurrentCycle , FirstActionTime , LastActiontime , Kickcount" << endl;
	//cout << getPlayerNumber() << "," << getCurrentCycle() << "," << FirstActionTime << "," << LastActionTime << "," << KickNum << endl;
	label++;	// 学習回数を１増やす
	fout.close();
};

void SelectPass::SavePlayerCycle( ){
  if(agentp->world().time().cycle() - LastActionTime == 0) return;
	// ファイルを追記モードで開く
	std::ofstream fout(PlayerCycleFile.c_str(), std::ios::app | std::ios::out );
	fout << cyclelabel << ",";		// 学習回数を行頭へ追加
	fout << KickNum << ",";
	fout << FirstActionTime << ",";
	fout << LastActionTime << ",";
	fout << agentp->world().time().cycle() << std::endl;
	cyclelabel++;
	fout.close();
};

void SelectPass::ReadData(){
	//std::cout << "ReadData OK" << std::endl;
	std::string temp;
	// ファイルを読み取り専用で開く
	std::ifstream fin(FileName.c_str(), std::ios::in);
	if(!fin)
		std::cout << FileName << "は存在しません" << std::endl;

	typedef boost::tokenizer<boost::char_separator<char> > tokenizer;
	boost::char_separator<char> sep(",");
	while(getline(fin, temp)){
		//tokenizer tok(temp);
		tokenizer tokens(temp, sep);
		tokenizer::iterator first = tokens.begin();
		first++; //skip r2
		first++; //skip r
		first++; //skip ak_F
		first++; //skip ak

		//first++; //skip cycle
		//first++;
		//std::cout << "number:" << agentp->world().self().unum() << std::endl;
		//std::cout << "start w[]" << std::endl;
		for(int i=0; i<MAX_HEURISTICS; i++){
			first++;
			w[i] = boost::lexical_cast<double>(*first);
			//std::cout << "w[" << i << "] : " << w[i] << std::endl;
		}
		//std::cout << "end w[]" << std::endl;
		//std::cout << "fin end" << std::endl;
	}
	fin.close();
};

void SelectPass::Updating(int tmp_a_F, int tmp_a){
  //if(getCurrentCycle() - LastActionTime <= 0) return;
  if(Ccnt - FirstActionTime <= 0) return;
  if(Ccnt - LastActionTime <= 0) return;

  //std::cout << "Updata start" << std::endl;
  //最後に行動したほうの更新
  if(1 <= tmp_a && tmp_a <= MAX_PLAYER){
	  double sum[MAX_HEURISTICS];             // 全エピソードのsum
	  for(int j=0; j<MAX_HEURISTICS; j++)
		  sum[j] = 0.0;                         // 初期化
	  for(int j=0; j<Policy_Vec.size(); j++){ // エピソードを調べていく
		  double tmp[MAX_HEURISTICS];           // ヒューリスティクスごとのsum_a
		  for(int m=0; m<MAX_HEURISTICS; m++){
			  tmp[m] = 0.0;
			  for(int n=1; n<=MAX_PLAYER; n++){
				  tmp[m] += tmp_U[j][n][m] * Policy_Vec[j][n];
			  }
			  sum[m] += tmp_U[j][tmp_a][m] - tmp[m];
		  }
	  }
	  for(int i=0; i<MAX_HEURISTICS; i++){
		  w[i] -= epsilon * r * sum[i] / T;
		  if(w[i] < 0) w[i] = 0.0;
	  }
  }
};

bool SelectPass::cntPossCycle(){
  if( isBallInOurPossesion( agentp ) ) Ccnt = agentp->world().time().cycle();
  if( agentp->world().time().cycle() - Ccnt > 5 ) return true; //敵ボールが5サイクル以上続いたら
  return false;
};

bool SelectPass::isBallInOurPossesion( rcsc::PlayerAgent * agent )
{
	if( TeamNearestBall( agent ) < OppNearestBall( agent ) )
	{
		//std::cout << "味方ボール" << std::endl;
		return true;
	}
	else
	{
		//std::cout << "敵ボール" << std::endl;
		return false;
	}
};

double SelectPass::getreliability(const rcsc::AbstractPlayerObject * target)
{
	return 1 - target->posCount() / 100;
};

void SelectPass::setLastActionTime(int cycle)
{
  LastActionTime = cycle;
  if(LastActionTime < 0)
  {
    LastActionTime = 0;
  }
};

void SelectPass::setFirstActionTime(int cycle)
{
  FirstActionTime = cycle;
  if(FirstActionTime < 0)
  {
    FirstActionTime = 0;
  }
};


void SelectPass::cntKickNum(int pnum)
{
  KickNum[pnum]++;
};

void SelectPass::initKickNum()
{
  for(int i=0; i<=MAX_PLAYER+1; i++){
    KickNum[i] = 0;
  }
};

void SelectPass::setOurPossesionCycle(long cycle)
{
  startOurPossesionCycle = cycle;
};

void SelectPass::setReward(double base)
{
  //int tmp = getCurrentCycle() - LastActionTime;

  int tmp = Ccnt - LastActionTime;
  int tmp2 = Ccnt - FirstActionTime;

  //if(tmp == 0) tmp = 1;

  if(tmp <= 0) return;
  if(tmp2 <= 0 ) return;

  //r = base / (double)tmp;

  r = base / (double)tmp;
  r2 = base / (double)tmp2;
  if(fabs(base) == 5.0){ //ゴールしたときの報酬
    r = base;
    r2 = base;
  }
  //r = base;
  //isInfOrNan(r, "reward at setR");
};

void SelectPass::setLearnFlag(bool learnf)
{
	learnFlag = learnf;
};

std::string SelectPass::getLearningFileName()
{
	return Learning_ini;
};

bool SelectPass::getreaddata()
{
	return MFreaddata;
};

bool SelectPass::getsavedata()
{
	return MFsavedata;
};

bool SelectPass::getLearn()
{
	return MFLearn;
};

void SelectPass::clearVar()
{
  //r = 0.0;
  //LastActionTime = getCurrentCycle();
  Policy_Vec.clear();
  tmp_U.clear();
}

int SelectPass::Boltzmann(){
	double P[MAX_PLAYER+1];		// 行動の確率
	//P[]はBoltzmannの計算結果を保存
	double temp;            	// 計算に使用
	double temp2 = 0.0;			// 計算に使用
	double tmpEnergy[MAX_PLAYER+1];

	/*
	for(int i=2; i<=MAX_PLAYER; i++)
		std::cout << i << ":" << getteam(agentp, i)->posCount() << std::endl;
	*/

	for(int m=2; m<=MAX_PLAYER; m++){
		tmpEnergy[m] = calcEnergyFunc(m);
	}
	//cout << "T : " << T << endl;
	// 各セルへの確率を計算
	for(int i=2; i<=MAX_PLAYER; i++){
		for(int j=2; j<=MAX_PLAYER; j++){
		  //temp = (tmpEnergy[i] - tmpEnergy[j]) / T;
		  //↓Σの計算でマイナスを付けたver(論文と逆のためUも変更)test@fuku
		  temp = (tmpEnergy[j] - tmpEnergy[i]) / T;
		  //std::cout << "temp:" << temp << std::endl;
		  temp2 += pow(M_E, temp);
		}
		//std::cout << "temp2:" << temp2 << std::endl;
		P[i] = 1 / temp2;
		temp2 = 0.0;
	}

	double RANDOM = (double(rand())/RAND_MAX); // 乱数生成
	double temp3 = 0.0;		// 計算に使用

	// 方策とヒューリスティクスをメンバ変数にコピー
	for(int l=2; l<=MAX_PLAYER; l++){
	  Policy[l] = P[l];
	  for(int o=0; o<MAX_HEURISTICS; o++)
	  {
	    tU[l][o] = U(o, l);
	    //std::cout << "U[" << l << "][" << o << "]" << "=" << tU[l][o] << std::endl;
	  }
	}
	pushPolicyData(Policy);
	//Boltzmannの計算結果をPolicy_Vecに格納

	std::vector< std::vector<double> > tmp;
	tmp.resize(MAX_PLAYER+1);
	for(int i=0; i<(int)tmp.size(); i++)
	  tmp[i].resize(MAX_HEURISTICS, 0.0);
	for(int i=2; i<=MAX_PLAYER; i++){
	  for(int j=0; j<MAX_HEURISTICS; j++)
	    tmp[i][j] = tU[i][j];
	}
	tmp_U.push_back(tmp);

	// 乱数を基に方策を決定
	//std::cout << "rand select : " << RANDOM << std::endl;
	for(int k=1; k<=MAX_PLAYER; k++){
		temp3 += P[k];		// 確率を足していく
		//std::cout << "temp3:" << temp3 << std::endl;
		if(RANDOM <= temp3)	// 乱数を超えたら
			return k;	// k:方策
	}
	//printcrash(agentp->world().self().unum(), "諦めました");
	return MAX_PLAYER;
};

double SelectPass::calcEnergyFunc(int tmp_a){
	E[tmp_a] = 0.0;		// 初期化

	// 重みとヒューリスティックを掛け合わせて足す
	for(int i=0; i<MAX_HEURISTICS; i++){
		//std::cout << "w[" << i << "]" << w[i] << std::endl;
		//std::cout << "U(" << i << "," << tmp_a << ")" << U(i, tmp_a) << std::endl;
		E[tmp_a] += w[i] * U(i, tmp_a);
	}
	return E[tmp_a];
};

void SelectPass::pushPolicyData(double *pol)
{
  std::vector<double> tmp(MAX_PLAYER+1);
  tmp.push_back(0.0); // ダミーデータ挿入
  for(int i=1; i<=MAX_PLAYER; i++)
    tmp.push_back(pol[i]);
  Policy_Vec.push_back(tmp);
};

double SelectPass::U(int tmp_i, int target_num){
	//ボルツマン選択の部分の目的関数の計算にマイナスを付けたver:test@fuku
	double u_value;

	U_EnemyCont = 1.0;		//敵信頼性を初期化(１＝最大評価)

	switch(tmp_i){
		case 0:	 // パスコースに敵はいるか
			u_value = ( 10 - isPassCourse(target_num) / 5.0 );
			break;
		case 1:	 // パス先とゴールとの距離
			u_value = ( 10 - DistanceToGoal(target_num) / 5.0 );
			break;
		//cut
		
		case 2:	 // パス先と最も近い敵との距離
			u_value = (10 - MostNearEnemy(target_num) / 5.0);
			break;
		/*
		case 2:
			u_value = ( 10 - CheckTarget(agentp, target_num) );
			break;
		*/
		case 3: // パス先の敵の数
			u_value = ( 10 - EnemyNum(target_num)  );
			break;
		case 4: // パス先の信頼性
			u_value = ( 10 - AgentConf(target_num) / 5.0 );
			break;
		case 5: // パス先との距離
			u_value = ( 10 - PassDistance(target_num) / 5.0 );
			break;
		default:        // エラーコード
			std::cout << "U error" << std::endl;
			exit(1);
	}
	//キーパーが相手の場合バックパスを想定して-50を返す
	//if(target_num == 1)
		//return u_value - 5;
	//	return 0;
	//else

	return u_value * U_EnemyCont;
};

// パスコースに敵はいるか
double SelectPass::isPassCourse(int target_num)
{
	double tmp_U_EnemyCont = 1.0;		//一時的に信頼性を保存
	const AbstractPlayerObject * target = getteam( agentp, target_num );

	/*
	if( PLAYER_RELIA <= target->posCount())
		return 50.0;
	*/

	if(target_num == agentp->world().self().unum())
		return 0.0;
	const Vector2D target_pos = target->pos();
	double agentDist = agentp->world().self().pos().dist( target->pos() );
	double agentAngle = ( target_pos - agentp->world().self().pos() ).dir().degree();

	double Point = -1.0;
	const AbstractPlayerCont opp_cont = agentp->world().allOpponents();
	for(AbstractPlayerCont::const_iterator opp_it = opp_cont.begin();
		opp_it != opp_cont.end();
  		opp_it++)
	{
		if( !(*opp_it) ) continue;		//いなければ飛ばす

		const double oppDist = agentp->world().self().pos().dist( (*opp_it)->pos() );
		const double oppAngle = ( (*opp_it)->pos() - agentp->world().self().pos() ).dir().degree();

		double diff = abs(agentAngle - oppAngle);
		double tmpPoint = 0.0;

		if(diff >= 30.0 || oppDist > agentDist)
			tmpPoint = 0.0;
		if(diff < 30.0 && oppDist <= agentDist)
			tmpPoint = 5.0;
		if(diff < 20.0 && oppDist <= agentDist)
			tmpPoint = 15.0;
		if(diff < 10.0 && oppDist <= agentDist)
			tmpPoint = 30.0;

		if(Point < tmpPoint)
		{
			Point = tmpPoint;
			tmp_U_EnemyCont = EnemyCont( agentp, (*opp_it)->unum() );	//信頼性を渡します
		}
	}
	if(Point < 0.0)
		return 40.0;
	U_EnemyCont = tmp_U_EnemyCont;					//信頼性を渡します
	return Point;
};

//HELIOS
double SelectPass::DistanceToGoal(int target_num)
{
	const AbstractPlayerObject * target = getteam( agentp, target_num );

	/*
	if( PLAYER_RELIA <= target->posCount())
		return 50.0;
	*/

	const Vector2D goal_pos(52.5, 0.0);
	const Vector2D target_pos = goal_pos - target->pos();

	double dist = target_pos.r()/2.0;
	if(dist > 40.0)
		dist = 40.0;
	return dist;
};

// レシーバと最近接の敵との距離0から
double SelectPass::MostNearEnemy(int target_num)
{
	double tmp_U_EnemyCont = 1.0;		//一時的に信頼性を保存

	const AbstractPlayerObject * target = getteam(agentp, target_num);

	/*
	if( getreliability( target ) < PLAYER_RELIA)
		return 50.0;
	*/

	if( PLAYER_RELIA <= target->posCount())
		return 50.0;
	double min_dist = 100.0;
	const AbstractPlayerCont opp_cont = agentp->world().allOpponents();
	for(AbstractPlayerCont::const_iterator opp_it = opp_cont.begin();
		opp_it != opp_cont.end();
		opp_it++ )
	{
		if( !(*opp_it) ) continue;		//いなければ飛ばす

		const double tmp = target->pos().dist( (*opp_it)->pos() );
		if(min_dist > tmp)
		{
			min_dist = tmp;
			tmp_U_EnemyCont = EnemyCont( agentp, (*opp_it)->unum() );	//信頼性を渡します
		}
	}
	if(min_dist == 100.0) return 40.0;
	if(min_dist > 30.0) min_dist = 30.0;
	min_dist = -1 * min_dist + 30.0;
	U_EnemyCont = tmp_U_EnemyCont;					//信頼性を渡します
	return min_dist;
};

//ヒューリスティクスのテスト@080501
//パス先に敵が何人いるか
int SelectPass::EnemyNum(int target_num)
{
	double limit_dist = 10.0;
	int enemnum = 0;
	int tmp_U_num = 0;					//信頼性を確認した敵の数を保存
	double tmp_U_EnemyCont = 1.0;		//一時的に信頼性を保存
	const AbstractPlayerObject * target = getteam(agentp, target_num);

	/*
	if( PLAYER_RELIA <= target->posCount())
		return 50.0;
	*/

	const AbstractPlayerCont opp_cont = agentp->world().allOpponents();
	for(AbstractPlayerCont::const_iterator opp_it = opp_cont.begin();
		opp_it != opp_cont.end();
		opp_it++)
	{
		if( !(*opp_it) ) continue;		//いなければ飛ばす

		double tmp = target->pos().dist( (*opp_it)->pos() );
		if(limit_dist < tmp)
		{
			enemnum++;
			tmp_U_EnemyCont = EnemyCont( agentp, (*opp_it)->unum() );	//信頼性を渡します
			tmp_U_num++;							//信頼性を取得したプレイヤー数を増やします
		}
	}
	//cout << "enemynum (around " << TargetPlayer << "): " << enemnum << endl;

	if(enemnum > 10)
		enemnum = 10;
	if( tmp_U_num != 0 )
		U_EnemyCont = tmp_U_EnemyCont / tmp_U_num;		//信頼性の平均値を返します
	return enemnum;
};

//HELIOS
double SelectPass::AgentConf(int target_num)
{

	if( target_num == agentp->world().self().unum() )
		return 0.0;
	//Confはカウント数
	//HELIOSは周りを常に確認しているので
	//10カウント以内ならsafeとする(暫定)
	double Conf = getreliability( getteam(agentp, target_num) );
	if( Conf < RELIA )
		return 0;
	//10カウント以上は正確性にかける
	//return -50 * 1/Conf + 50.0;
	return -50 * Conf + 50.0;
};

double SelectPass::PassDistance(int target_num)
{
	if(target_num == agentp->world().self().unum())
		return 0.0;
	double agentDist = agentp->world().self().pos().dist( getteam(agentp, target_num)->pos() );
	if(agentDist > 50.0)
		agentDist = 50.0;
	return agentDist;
};

//パス先のポジションによって判断
double SelectPass::CheckTarget( rcsc::PlayerAgent * agent , int target_num )
{
	double return_hue;	//返す評価値

	if( agent->world().self().unum() == target_num )
	//自分自身へのパス(ドリブル)の時は0を返す
	{
		return_hue = 0;
	}
	else
	{
		switch( target_num )
		{
		case 9:
		case 10:
		case 11:
			//FWにパス
		{
			return_hue = 0;
			break;
		}
		case 2:
		case 3:
		case 4:
		case 5:
			//DFにパス
		{
			return_hue = 10;
			break;
		}
		default:
			//それ以外(MFにパス）
		{
			return_hue = 0;
			break;
		}

		}
	}
	return return_hue;
};

/*** 追加 ***/
double SelectPass::TeamNearestBall( rcsc::PlayerAgent * agent )
{
	double dist;
	double min = MAX_DIST;
	rcsc::Vector2D ballpos = agent->world().ball().pos();
	AbstractPlayerCont team = agent->world().allTeammates();
	for(	AbstractPlayerCont::const_iterator target_it = team.begin();
			target_it != team.end();
			target_it++
			)
	{
		dist = (*target_it)->pos().dist( ballpos );
		if( dist < min )
			min = dist;
	}
	//std::cout << min << std::endl;
	return min;
};

double SelectPass::OppNearestBall( rcsc::PlayerAgent * agent )
{
	double dist;
	double min = MAX_DIST;
	rcsc::Vector2D ballpos = agent->world().ball().pos();
	AbstractPlayerCont team = agent->world().allOpponents();
	for(	AbstractPlayerCont::const_iterator target_it = team.begin();
			target_it != team.end();
			target_it++
			)
	{
		dist = (*target_it)->pos().dist( ballpos );
		if( dist < min )
			min = dist;
	}
	//std::cout << min << std::endl;
	return min;
};

double SelectPass::EnemyCont( rcsc::PlayerAgent * agent, int enemy_num)
{
	//自作敵位置評価値
	double Conf = 1 - getopp( agent, enemy_num )->posCount() / 10;
	if( Conf < RELIA )
		return 0;
	//Conf[0,1]
	return Conf;
};
