#ifndef SELECTPASS
#define SELECTPASS

#include <rcsc/player/player_agent.h>
#include <string>
#include <vector>

using namespace rcsc;


//クラス化して見やすくしてみる
class SelectPass
{
public:
	SelectPass() :
		dir_0("csv/"),
		dir_1("csv/DF_Learning/"),
		dir_2("csv/MF_Learning/"),
		dir_3("csv/FW_Learning/"),
		dir_4("csv/DF-MF_Learning/"),
		dir_5("csv/DF-FW_Learning/"),
		dir_6("csv/MF-FW_Learning/"),
		dir_7("csv/DF-MF-FW_Learning/"),
		Learning_ini("learn.ini"),
		initFlag(false),
		learnFlag(false),
		epsilon(0.1),
		r(0.0),
		T(0.1),
		Ccnt(0),
		FirstActionTime(0),
		LastActionTime(0),
		label(1),
		cyclelabel(1)
	{};
private:
	// @param

	//static
	static const int MAX_HEURISTICS = 5;
  	static const int MAX_PLAYER = 10;
  	static const double MAX_DIST = 100.0;
  	//信頼性
  	//カウント数で表現
  	//未確定
  	//10カウント辺りが限界だとおもわれる(暫定）
   //static const double PLAYER_CONF_THR = 4;
  	//static const double PLAYER_CONF_THR = 0.90;
  	static const int PLAYER_RELIA = 10;
  	static const double RELIA = 0.9;

	//AgentPlayerPoint
	PlayerAgent* agentp;

	//Path
	const std::string dir_0;
	const std::string dir_1;
	const std::string dir_2;
	const std::string dir_3;
	const std::string dir_4;
	const std::string dir_5;
	const std::string dir_6;
	const std::string dir_7;


   	//学習フラグ
    bool MFreaddata;
    bool MFsavedata;
    bool MFLearn;
    bool initFlag;
    bool learnFlag;

   	// 学習関係
	double w[MAX_HEURISTICS];				 //!< 重み配列
	double epsilon;							 //!< 学習係数
	double r,r2;							 //!< 最初の行動に対する報酬、最後の行動に対する報酬
	double T;								 //!< 温度
	double E[MAX_PLAYER+1];					 //!< セルのエネルギー
	double tU[MAX_PLAYER+1][MAX_HEURISTICS]; //!< ヒューリスティックの一時的保存
	double Policy[MAX_PLAYER+1];	    	 //!< 方策（確率）
	int Ccnt;                          		 //!< サイクルのカウント
	int LastActionTime;
	int FirstActionTime;
	int sd, a, a_F;	      		    		 //!< 次状態、行動、最初の行動
	std::string FileName;					 //!< ファイル名
	std::string CoachCycleFile;     		 //!< 保持時完の記録ファイル
	std::string PlayerCycleFile;       		 //!< 保持時完の記録ファイル
	int label;						    	 //!< 学習回数（ファイルの保存用）
	int cyclelabel;
	std::string DirPath; 	       	    	 //!< 読み込むディレクトリ
	std::string Learning_ini;     	    	 //!< 学習設定ファイル
	int startOurPossesionCycle;
	int KickNum[MAX_PLAYER+1];
	std::vector< std::vector<double> > Policy_Vec;
	std::vector< std::vector< std::vector<double> > > tmp_U;
	bool OCsavecycle;
	double U_EnemyCont;						//ヒューリスティクスごとの敵

public:
	//!実行メソッド
	//! @param PlayerAgent
	//! @return なし
	void execute(rcsc::PlayerAgent * agent);

	//! SelectPassの初期化
	//! @param PlayerAgent
	//r @return なし
	void init(rcsc::PlayerAgent * agent);

	//! SelectPass学習部分の初期化
	//! @param PlayerAgent
	//! @return なし
	void learn_init(rcsc::PlayerAgent * agent);

private:

	/***
	 *
	 *  学習部分
	 *
	***/
	/*** 初期化 ***/
	//!
	//! 学習するかどうかの初期化
	//! @param なし
	//! @return なし
	//!
	void isMFLearning();

	void initLearning();

	/*** 保存部分 ***/
	//!
	//! 重み配列をファイル(csv形式)に保存
	//! @param なし
	//! @return なし
	//!
	void SaveData(int ak_F, int ak);

	//!
	//! 保持時間をファイルに保存
	//! @param なし
	//! @return なし
	//!
	void SavePlayerCycle( );

	//学習結果のファイル名を取得
	//!
	//! ファイル(csv形式)を読み込んで重み配列へ格納
	//! @param なし
	//! @return なし
	//!
	void ReadData();

	//!
	//! 重みを更新
	//! @param tmp_a 実際に行動したセル番号
	//! @param tmp_a_F 最初の行動
	//! @return なし
	//!
	void Updating(int tmp_a_F, int tmp_a);

	//敵がボールを５サイクル以上保持しているか判定
	bool cntPossCycle();
	//ボールに一番近いのが見方であるかを確認
	bool isBallInOurPossesion( rcsc::PlayerAgent * agent );
	//信頼性を計算
	double getreliability(const rcsc::AbstractPlayerObject * target);
	//味方の内もっともボールに近い距離
	double TeamNearestBall( rcsc::PlayerAgent * agent );
	//味方の内もっともボールに近い距離
	double OppNearestBall( rcsc::PlayerAgent * agent );

	//Setter
	void setLastActionTime(int cycle);
	void setFirstActionTime(int cycle);
	void cntKickNum(int pnum);
	void initKickNum();
	void setOurPossesionCycle(long cycle);
	void setReward(double base);
	void setLearnFlag(bool learnf);
	//getter
	std::string getLearningFileName();
	bool getreaddata();
	bool getsavedata();
	bool getLearn();

	void clearVar();


	/*** 行動選択部分 ***/

	//!
	//! ボルツマン分布関数による行動選択
	//! @param なし
	//! @return 実際に行う行動(パス先の背番号)
	//!
	int Boltzmann();
	//!
	//! エネルギー関数を計算
	//! @param tmp_a セル番号
	//! @return セルの価値（エネルギー）
	//!
	double calcEnergyFunc(int tmp_a);

	void pushPolicyData(double *pol);

	//!
	//! それぞれのヒューリスティクス
	//! @param tmp_i ヒューリスティクスの番号
	//! @param tmp_sd 次状態のセル番号
	//! @return 計算結果
	//!
	double U(int tmp_i, int target_num);

	/**** ヒューリスティクス ****/
	// パスコースに敵はいるか
	double isPassCourse(int target_num);
	// ゴールとの距離を測る
	double DistanceToGoal(int target_num);
	// レシーバと最近接の敵との距離0から
	double MostNearEnemy(int target_num);
	// ヒューリスティクスのテスト@080501
	// パス先に敵が何人いるか
	int EnemyNum(int target_num);
	// パス先の信頼性を返す
	double AgentConf(int target_num);
	// パス先の距離を返す
	double PassDistance(int target_num);
	//パス先のポジションによって判断
	double CheckTarget( rcsc::PlayerAgent * agent , int target_num );
	//@param 敵番号enemy_numの信頼性を返す
	double EnemyCont( rcsc::PlayerAgent * agent, int enemy_num);

	//自作
	//引数のAbstractPlayerObjectを渡す
	//@param int num : 背番号
	//@retrun AbstractPlayerObject : プレイヤーオブジェクト
	const rcsc::AbstractPlayerObject * getteam( rcsc::PlayerAgent * agent , int num );
	const rcsc::AbstractPlayerObject * getopp( rcsc::PlayerAgent * agent , int num );
	//行動部分
	//pass
	//@param PlayerAgnet agent : そのまま渡す
	//@param int num 			   : Bolzmannで得られた値を渡す
	bool pass( rcsc::PlayerAgent * agent , int num , rcsc::Vector2D target_vec);
	//dribble
	//@param PlayerAgnet agent : そのまま渡す
	bool dribble( rcsc::PlayerAgent * agent );

	/*** Debug tool ***/
	void printcrash(long unum, std::string point)
	{
		std::cout << unum << " : " << point << ": error crash!" << std::endl;
	};
public:
};

#endif
